/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _BFIR_H_
#define _BFIR_H_

#include <inttypes.h>

#include "defs.h"
#include "almusvcu.h"
#include "apui_defs.h"
#include "bfirmem.h"

/*
 * FIXME: comments are not fully up to date
 * FIXME: revprg reverbprogram naming...
 */

struct _bfirfilter {
    int block_size;   /* block size in samples (NB: frequency domain block size
			 is twice the size) */
    int n_blocks;     /* how many blocks in filter */
    struct memseg memseg;
};

struct bfirfilter {
#define BFIR_FILTER_TYPE_REVERB 0
#define BFIR_FILTER_TYPE_EQ     1
#define BFIR_FILTER_TYPE_XTC    2
#define BFIR_FILTER_TYPE_UHJ    3
#define BFIR_FILTER_TYPE_COUNT  4    
    int type;
    int n_rates;
    int rates[VCU_VALID_RATES_COUNT];
    struct _bfirfilter f[VCU_VALID_RATES_COUNT];
};

struct bfirdyneq {
    int eq_id; /* not used externally */
    int blocks; /* not used externally */
    int n_bands;
    double band[VCU_EQPRG_MAX_BANDS];
    double amp[VCU_EQPRG_MAX_BANDS];
};

extern bool_t bfir_debug;

/*
 * This function must be called first to initiliaise this module. It cannot be
 * called more than once. Returns true on success, false on failure.
 *
 * 'alloc_bytes' is how many shared memory bytes to allocate for filters used
 * by the reverb, equalisation and crosstalk cancellation programs. This should
 * typically be more than 100 megabytes.
 *
 * 'brutefir_path' is the path to the BruteFIR binary.
 *
 * 'brutefir_local_socket' is the name of the local socket that will be created
 * by BruteFIR when run. Typically it should point to a name in /tmp
 *
 * 'brutefir_input_device' is the name of the BruteFIR input device. See
 * BruteFir documentation for its syntax. Typically it is "hw:0,0,0".
 *
 * 'brutefir_input_device_sample_format' is one of the ALSA constants
 * SND_PCM_FORMAT_S32_LE and SND_PCM_FORMAT_S16_LE, and describes the sample
 * format of the input device.
 *
 * 'brutefir_input_device_open_channels' is the amount of channels that should
 * be open on the input device. The idea is that it should be possible to open
 * more channels than the two that are used for input. This is necessary for
 * many sound cards.
 *
 * 'brutefir_input_device_channel_index' is the channel indexes of left and
 * right input stereo channels.
 *
 * 'brutefir_output*' works the same as 'brutefir_input*' described above.
 *
 * In 'brutefir_output_device_channel_index' places 0 and 1 points out the
 * ambiopole speakers left and right, and the following points out reverb
 * speakers.
 */
bool_t
bfir_init(int alloc_bytes,
          const char log_path[],
	  const char brutefir_path[],
          const char brutefir_modules_path[],
          const char brutefir_convolver_config[],
          int brutefir_input_device_count,
	  char brutefir_input_device[CONF_DEVICES_MAX][PATH_MAX],
	  char brutefir_input_device_params[CONF_DEVICES_MAX][PATH_MAX],
	  const int brutefir_input_open_channels[CONF_DEVICES_MAX],
	  int brutefir_input_channel_map[CONF_DEVICES_MAX][VCU_MAX_CHANNELS],
          int brutefir_output_device_count,
	  char brutefir_output_device[CONF_DEVICES_MAX][PATH_MAX],
	  char brutefir_output_device_params[CONF_DEVICES_MAX][PATH_MAX],
	  const int brutefir_output_open_channels[CONF_DEVICES_MAX],
	  int brutefir_output_channel_map[CONF_DEVICES_MAX][VCU_MAX_CHANNELS],
          int min_blocksize,
          bool_t allow_poll_mode);


void
bfir_flush_logs(void);

/*
 * This function sets up the filter engine for a specific configuration. This
 * can be changed at all times when BruteFIR is not running. Note that all
 * previously allocated filters are dropped when this is called. Returns true
 * on success, false on failure.
 *
 * 'n_rates' is how many sample rates there should be support for. 'rates'
 * contains the rates in Hz, typically one or more of 44100, 48000, 88200,
 * 96000.
 *
 * 'block_sizes' contains the block sizes for each rate. The block size decides
 * how large the I/O-delay will be. It gets twice the number of samples as the
 * block size is. However, a too short block size will take too much processor
 * time. Use 'bfir_benchmark' to find out suitable settings.
 *
 * The 'max_*' arrays contain the maximum lengths in samples for each rate for
 * the filters in their names. The lengths must be divisable with the matching
 * block size. Note that there is only one equalisation filter, although it may
 * consist of several convolved filters in reality. Thus, the equalisation
 * length should be chosen to be as long so it fits the convolved filters.
 *
 * 'n_reverb_channels' is the total amount of reverberation channels. Typically
 * 6 - 8.
 */
bool_t
bfir_hard_settings(int n_rates,
		   const int rates[],
		   const int block_sizes[],
		   int n_reverb_input_channels,
		   int n_reverb_output_channels,
                   int ambisonic_input_format,
                   int realsize);

struct bfir_chmap {
    uint32_t channels[VCU_CH_COUNT][VCU_MAX_CHANNELS/32];
};

int
bfir_verify_phys_map(const struct bfir_chmap *maps[2],
                     int ambisonic_input_format,
                     int channel_count[2][VCU_CH_COUNT]);

bool_t
bfir_soft_settings(int input_sample_format,
		   int output_sample_format,
                   bool_t apply_dither,
		   uint32_t apply_eq[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS/32],
		   const bool_t apply_xtc[],
                   const double ambisonic_gains[],
		   const struct bfir_chmap *maps[2]);

/*
 * Allocates a filter set for a reverb program and puts it into 'filters'. The
 * desired length in samples for each rate is specified by 'lengths'. The
 * lengths must be divisable with the corresponding block size given in
 * 'bfir_settings'. The length of the 'filters' array must be
 * 2 * 'n_reverb_channels' given in 'bfir_settings'. The filters are ordered in
 * left and right for each reverb channel.
 *
 * On success, a filter id which is a number >= 0 is returned, on error -1 is
 * returned.
 *
 * BruteFIR must not be running when calling this function.
 */
int
bfir_new_filters(int type,
                 struct bfirfilter filters[],
                 const int lengths[]);

int
bfir_shorten_filters(int id,
                     const int lengths[]);

void
bfir_delete_filters(int id);

double
bfir_filter_capacity(int type,
                     double *max_capacity);

/*
 * Set the reverb program that should be used. This can be done when BruteFIR
 * is running, and then it will change the reverb program in runtime. If done
 * when BruteFIR is not running, 'id' will identify the reverb program id that
 * will be the first used when BruteFIR starts. 'id' is a id number previously
 * received from 'bfir_new_reverbpgr_filters'. It is the user's responsibility
 * that the 'id' exists and is allocated.
 */
void
bfir_set_reverbprg(int id);

void
bfir_set_equalisation(const uint32_t mask[],
		      int type,
		      int id, /* may be -1 */
                      int io);

void
bfir_set_xtalkcancel(const uint32_t mask[],
                     int id);

void
bfir_set_ambisonic_filters(int uhj_id);

void
bfir_set_dyneq(int id, /* may be -1 (to clear all) */
               const struct bfirdyneq *dyneq); /* may be NULL */

void
bfir_set_volume(int io,
                int type,
                const uint32_t mask[],
                int start,
                int end,
                double volume);

void
bfir_set_delay(int io,
               int type,
               const uint32_t mask[],
               int start,
               int end,
               int delay);

void
bfir_set_active(int io,
                int type,
                const uint32_t mask[],
                int start,
                int end);

/*
 * Reset the decibel peak meter. Makes only sense to call it in runtime. See
 * 'bfir_start' for further information.
 */
void
bfir_reset_peak_meter(void);

void
bfir_start_buffering_commands(void);

void
bfir_commit_buffered_commands(void);

/*
 * Stop BruteFIR if it is running.
 */
void
bfir_stop(void);

/*
 * Start BruteFIR. Before doing this, a few things must be prepared. Apart from
 * the initialisation function 'bfir_settings' and filter allocation and
 * initialisation that must be done, the following functions should also be be
 * used to set up things: 'bfir_set_delay', 'bfir_input_volume',
 * 'bfir_output_volume', 'bfir_set_active', 'bfir_set_reverb_program'. All
 * these can later be called during runtime to change settings.
 *
 * The 'bfir_benchmark' functions can be used as help to get suitable values of
 * some of the settings.
 *
 * 'sample_rate' specifies which sample rate to start with, which must be one of
 * the ones defined in the last 'brutefir_settings' call.
 *
 * 'maxdelay' is the maximum amount of samples which a channel can be delayed in
 * runtime with the 'bfir_set_delay' function. Naturally, it must be set to
 * equal to or higher than the delays given to 'bfir_set_delay' earlier, but
 * should not be set higher than necessary, since BruteFIR must preallocate
 * memory to handle delay changes.
 *
 * 'apply_crosstalk_cancellation' specifies if the crosstalk cancellation
 * filters should be used for the ambiopole, or if only equalisation should be
 * applied.
 *
 * 'peak_meter' is a callback function which will receive the currently highest
 * peak in decibel achieved. Negative values should be the normal one, any
 * positive numbers mean overflow. The peak meter of BruteFIR can be reset in
 * runtime by calling 'bfir_reset_peak_meter', and then the callback function
 * will receive updates accordingly.
 *
 * 'process_terminated' is a callback function which will be called when
 * BruteFIR terminates, the 'reason' argument being the exit code of BruteFIR.
 *
 * The return value is -1 on failure, and a file descriptor on success. The file
 * descriptor should be put in the read set of a select call, and when active,
 * the 'bfir_process' function should be called, which in its turn may invoke
 * one or both of the callback functions.
 */
int
bfir_start(int sample_rate,
	   int maxdelay[2],
           bool_t powersave,
           bool_t watch_clock,
	   void (*peak_meter)(double peak),
	   void (*process_terminated)(int reason),
           int sample_format,
           int channel_count,
           bool_t loop,
           const char path[]);

/*
 * To be called when the 'bfir_start' file descriptor becomes active. From this
 * function calls to the 'bfir_start' callback functions may be generated.
 */
void
bfir_process(void);

/*
 * A benchmark function used to find out how long filters and how short delay
 * the current hardware can handle. Note that this function normally takes a
 * few minutes to execute. The result is a series of suggestions, the amount
 * stored in 'n_alternatives', which is BFIR_BENCHMARK_N_BLOCK_SIZES or less.
 * If there was too little memory or too little processor power,
 * 'n_alternatives' will receive 0, meaning that the hardware cannot run
 * BruteFIR at all with the given settings. The arrays 'block_sizes' will
 * receive the suggested block sizes, matched with the longest possible reverb
 * length in samples stored in 'max_reverb_lengths'. The arrays are sorted with
 * the shortest lengths and the shortest block sizes first. The I/O delay of
 * BruteFIR is twice the amount of samples in the block size.
 *
 * 'sample_rate' is the sample rate BruteFIR should be run
 * 'n_reverb_channels' is the amount of reverberation speakers.
 * 'crosstalk_cancellation_length' is the length in samples of the crosstalk
 * cancellation filters. If not divisable with the currently used block size,
 * the length that is closest larger and divisable will be used. The same counts
 * for 'equalisation_length'.
 *
 * 'max_load' is the maximum allowable processor load to accept a benchmark run
 * as practical. Since the benchmarking is not very precise and there must be
 * some processor time left to handle the task in realtime, this value should
 * not be set to higher than 0.9 (1.0 means 100%).
 *
 * 'max_seconds' is a memory limiter, the maximum amount of seconds a reverb
 * program should be. This should be set to a reasonable value, there is no
 * idea to have say 20 seconds of reverb, if there is only memory for one or
 * no such reverb program. A lower value will also reduce the time for
 * benchmarking
 *
 * 'max_megabytes' is also a memory limiter, and is the maximum amount of
 * memory BruteFIR is allowed to take. This can typically be 40 megabytes. It is
 * perfectly ok to use only the 'max_seconds' limiter, and set this one to a
 * very large value, but at least one of the memory limiters must be set to a
 * reasonable value.
 *
 * 'progress_callback' is a callback function which will get the results of
 * each test run. This has no use other than to give the user feedback during
 * the few minutes the benchmark run will take. 'n_blocks' is the number of
 * reverb blocks tested, 'block_size' is the block size tested, and load is the
 * processor load needed to achieve real-time (which could be higher than 1.0).
 *
 * The return value is one of BFIR_BENCHMARK_STATUS_*, indicating failure, or
 * success with the extra information saying if memory or processor was the
 * limiting factor.
 */
#define BFIR_BENCHMARK_STATUS_INVALID_VALUE -3
#define BFIR_BENCHMARK_STATUS_OUT_OF_RANGE -2 /* Too demanding setup */
#define BFIR_BENCHMARK_STATUS_RUN_FAILED -1 /* Failed to run BruteFIR */
#define BFIR_BENCHMARK_STATUS_MEMORY_LIMITED 1
#define BFIR_BENCHMARK_STATUS_PROCESSOR_LIMITED 2
#define BFIR_BENCHMARK_STATUS_HARD_LIMITED 3
int
bfir_benchmark(int sample_rate,
               bool_t test_low_latency,
               bool_t apply_dither,
               int ambisonic_input_format,
               const struct bfir_chmap *maps[2],
	       double max_load,
	       double max_seconds,
	       int *n_alternatives,
	       int block_sizes[],
	       int n_blocks[],
               void (*filter_lengths)(int block_size,
                                      int sample_rate,
                                      int xtc_length[],
                                      int *eq_length[2][VCU_CH_WP_COUNT]),
               void (*progress_callback)(int n_blocks,
					 int block_size,
					 double load));
#endif
