/*
 * (c) Copyright 2002 - 2005 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdarg.h>
#include <dirent.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>

#include "elog.h"

#define MAX_ASCTIME 25

static char elog_filename[PATH_MAX] = "";
int global_loglevel = LOGINFO;
FILE *elog_stream = NULL;

static char *
level_str(int level)
{
    switch (level) {
    case LOGDEBUG:
        return "   debug";
    case LOGINFO:
        return "    info";
    case LOGERROR:
        return "critical";
    default:
        return " UNKNOWN";
    }
}

static void
check_elog_stream(void)
{
    if (elog_stream == NULL && elog_filename[0] != '\0') {
	if ((elog_stream = fopen(elog_filename, "at")) == NULL) {
	    elog_stream = stderr;
	    elog(LOGERROR, "Failed to open \"%s\" for writing (%s), "
                 "using stderr instead.", elog_filename, strerror(errno));
	    /* continue anyway */
	}
    } else if (elog_stream == NULL) {
        elog_stream = stderr;
    }
}

void
elog_init(const char log_path[])
{
    strncpy(elog_filename, log_path, PATH_MAX);
    elog_filename[PATH_MAX-1] = '\0';
}

void
elog_concatenate_file(const char path[])
{
    struct stat statbuf;
    uint8_t buf[1024];
    FILE *source;
    size_t bytes;
    char *p;

    check_elog_stream();
    
    if (stat(path, &statbuf) == -1) {
        return;
    }
    p = asctime(localtime(&statbuf.st_mtime));
    p[MAX_ASCTIME - 1] = '\0';
    fprintf(elog_stream,
            "---- START FILE \"%s\" (size %llu bytes, date %s) ----\n",
            path, (uint64_t)statbuf.st_size, p);
    if ((source = fopen(path, "rb")) == NULL) {
	return;
    }
    while (1) {
        if ((bytes = fread(buf, 1, sizeof(buf), source)) < 0) {
            fclose(source);
            return;
        }
        if (bytes == 0) {
            break;
        }
        if (fwrite(buf, 1, bytes, elog_stream) != bytes) {
            fclose(source);
            return;
        }
    }
    fclose(source);
    fprintf(elog_stream,
            "---- END FILE \"%s\" (size %llu bytes, date %s) ----\n",
            path, (uint64_t)statbuf.st_size, p);

    if (elog_filename[0] == '\0') {
        elog_stream = NULL;
    }
}

void
elog_truncate(void)
{
    if (elog_stream != NULL) {
        rewind(elog_stream);
        ftruncate(fileno(elog_stream), 0);
    }
}

int
elog_loglevel_get(void)
{
    return global_loglevel;
}

void
elog_loglevel_set(int loglevel)
{
    global_loglevel = loglevel;
}

void
elog_close(void)
{
    if (elog_stream != NULL) {
        fclose(elog_stream);
        elog_stream = NULL;
    }
}

void
_elog(int level,
      const char _file[],
      const char _function[],
      int _line,
      const char format[],
      ...)
{
    char file[PATH_MAX], *p, prelog[PATH_MAX + 200];
    struct timeval tv;
    va_list arg;

    if (level > global_loglevel) {
        return;
    }

    check_elog_stream();
    
    strcpy(file, _file);
    if ((p = strchr(file, '.')) != NULL) {
	*p = '\0';
    }
    gettimeofday(&tv, NULL);
    p = asctime(localtime(&tv.tv_sec));
    p[MAX_ASCTIME - 6] = '\0';
    p[MAX_ASCTIME - 1] = '\0';
    sprintf(prelog, "[ %s.%06lu %s %s/%s/%d ] %s: ", &p[4], tv.tv_usec,
            &p[MAX_ASCTIME - 5], file, _function, _line,
            level_str(level));
    fprintf(elog_stream, "%-50s", prelog);
    if (level <= LOGERROR && elog_stream != stderr) {
        fprintf(stderr, "%-50s", prelog);
    }
    va_start(arg, format);
    vfprintf(elog_stream, format, arg);
    va_end(arg);
    fprintf(elog_stream, "\n");
    fflush(elog_stream);
    if (level <= LOGERROR && elog_stream != stderr) {
        va_start(arg, format);
        vfprintf(stderr, format, arg);
        va_end(arg);
        fprintf(stderr, "\n");
    }
    if (elog_filename[0] == '\0') {
        elog_stream = NULL;
    }
}
