/*
 * (c) Copyright 2002 - 2004, 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "resample.h"
#include "firwindow.h"

#define TAP_MULTIPLIER 200
#define MAX_SPACING 5000

struct _resample_t_ {
    void *filter;
    void *dbuf[2];
    int inspacing;
    int outspacing;
    int n_taps;
    int realsize;
};

resample_t *
resample_new(int inrate,
	     int outrate,
             int realsize,
	     int *_error)
{
    int inspacing, outspacing, n_taps, minrate, gcd, filter_srate, n, *error;
    int __error;
    double cutoff, *dpos;
    float *fpos;
    resample_t *rs;

    error = (_error == NULL) ? &__error : _error;
    
    *error = 0;
    
    /* sanity check input */
    if (inrate < 44100 || inrate > 384000 ||
	outrate < 44100 || outrate > 384000)
    {
	*error = RESAMPLE_ERR_INVALID_RATE;
	return NULL;
    }
    if (realsize != 4 && realsize != 8) {
        *error = RESAMPLE_ERR_INVALID_VALUE;
        return NULL;
    }

    /* find greatest common divisor */
    minrate = (inrate > outrate) ? outrate : inrate;
    gcd = minrate;
    while (inrate / gcd * gcd != inrate || outrate / gcd * gcd != outrate) {
	gcd--;
    }
    inspacing = outrate / gcd;
    outspacing = inrate / gcd;
    if (inspacing > MAX_SPACING || outspacing > MAX_SPACING) {
	*error = RESAMPLE_ERR_GCD_TOO_SMALL;
	return NULL;
    }

    /* allocate memory */
    if ((rs = malloc(sizeof(resample_t))) == NULL) {
	*error = RESAMPLE_ERR_INTERNAL_ERROR;
	return NULL;
    }
    n_taps = inspacing * TAP_MULTIPLIER;
    rs->filter = malloc(n_taps * realsize);
    rs->dbuf[0] = malloc((TAP_MULTIPLIER + 10) * realsize);
    rs->dbuf[1] = malloc((TAP_MULTIPLIER + 10) * realsize);
    if (rs->filter == NULL || rs->dbuf[0] == NULL || rs->dbuf[1] == NULL) {
	*error = RESAMPLE_ERR_INTERNAL_ERROR;
	resample_delete(rs);
	return NULL;
    }
    rs->inspacing = inspacing;
    rs->outspacing = outspacing;
    rs->n_taps = n_taps;
    rs->realsize = realsize;

    /* calculate lowpass filter */
    filter_srate = inspacing * inrate;
    cutoff = 0.90 * (double)minrate / (double)filter_srate * M_PI;
    if (rs->realsize == 4) {
        fpos = &((float *)rs->filter)[(n_taps - 1) >> 1];
        for (n = -((n_taps - 1) >> 1); n != 0; n++) {
            fpos[n] = sin((double)n * cutoff) / ((double)n * M_PI);
        }
        fpos[0] = cutoff / M_PI;
        for (n = 1; n <= n_taps >> 1; n++) {
            fpos[n] = sin((double)n * cutoff) / ((double)n * M_PI);
        }
    } else {
        dpos = &((double *)rs->filter)[(n_taps - 1) >> 1];
        for (n = -((n_taps - 1) >> 1); n != 0; n++) {
            dpos[n] = sin((double)n * cutoff) / ((double)n * M_PI);
        }
        dpos[0] = cutoff / M_PI;
        for (n = 1; n <= n_taps >> 1; n++) {
            dpos[n] = sin((double)n * cutoff) / ((double)n * M_PI);
        }
    }
    firwindow_kaiser(rs->filter, n_taps,
                     (n_taps & 1) != 0 ? 0 : -1.0,
                     9, realsize);
    return rs;
}

void
resample_delete(resample_t *rs)
{
    if (rs == NULL) {
	return;
    }
    free(rs->filter);
    free(rs->dbuf[0]);
    free(rs->dbuf[1]);
    free(rs);
}

int
resample(resample_t *rs,
	 void *input,
	 void *output,
	 int insize,
	 int outsize,
	 int (*read_block)(void *inbuf, int n_samples, int realsize),
	 int (*write_block)(void *outbuf, int n_samples, int realsize))
{
    int outlim, inread, n, i, k, x, y, z, cc;
    int curbuf, n_samples, outblocks, base;
    void *outbuf, *inbuf, *buf;
    bool_t islastblock = false;

    /*
     * FIXME: Stopped coding when it started working, some cleanups and
     * optimisations should be possible to do.
     */
    
    memset(rs->dbuf[0], 0, (TAP_MULTIPLIER + 10) * rs->realsize);
    memset(rs->dbuf[1], 0, (TAP_MULTIPLIER + 10) * rs->realsize);
    if (read_block != NULL) {
	if ((inbuf = malloc(rs->n_taps * rs->realsize)) == NULL) {
	    return -1;
	}
    } else {
	inbuf = input;
    }
    if (write_block != NULL) {
	if ((outbuf = malloc(rs->n_taps * rs->realsize)) == NULL) {
	    if (read_block != NULL) {
		free(inbuf);
	    }
	    return -1;
	}
    } else {
	outbuf = output;
    }
    i = k = x = y = z = cc = curbuf = inread = outblocks = 0;
    while (true) {
	if (read_block != NULL) {
	    if ((n_samples = read_block(inbuf, rs->n_taps, rs->realsize))
                <= 0)
            {
		break;
	    }
	} else {
	    inbuf = &((uint8_t *)input)[inread * rs->realsize];
	    if (inread + rs->n_taps > insize) {
		if ((n_samples = insize - inread) == 0) {
		    break;
		}
	    } else {
		n_samples = rs->n_taps;
	    }
	}
	inread += n_samples;
	for (n = 0; n < n_samples; n++) {
	    /*
	     * NB: got noise problems when rs->outspacing multiplication was
	     * done in the filter directly.
	     *
	     * Overall problems with -O2 flag... gcc is doing wrong -> noisy
	     * outputs at some occations (test 48000 -> 44100 Hz).
	     */
            if (rs->realsize == 4) {
                ((float *)rs->dbuf[curbuf])[z++] =
                    ((float *)inbuf)[n] * (float)rs->outspacing;
            } else {
                ((double *)rs->dbuf[curbuf])[z++] =
                    ((double *)inbuf)[n] * (double)rs->outspacing;
            }
	    if (i + rs->inspacing >= rs->n_taps) {
		if (write_block == NULL &&
		    outsize < (outblocks + 1) * rs->n_taps)
		{
		    outlim = outsize % rs->n_taps;
		    islastblock = true;
		} else {
		    outlim = rs->n_taps;
		}
		for (x %= rs->n_taps;
		     x < rs->n_taps;
		     x += rs->outspacing, cc++)
		{
		    base = cc * rs->outspacing % rs->inspacing;
		    z = (rs->n_taps - base - 1) / rs->inspacing;

                    if (rs->realsize == 4) {
                        ((float *)outbuf)[k] = 0;
                        buf = rs->dbuf[!curbuf];
                        for (y = base + z * rs->inspacing,
                                 z = (rs->n_taps + x - y) / rs->inspacing;
                             y > x;
                             y -= rs->inspacing, z++)
                        {
                            ((float *)outbuf)[k] +=
                                ((float *)buf)[z] * ((float *)rs->filter)[y];
                        }
                        buf = rs->dbuf[curbuf];
                        for (z = (x - y) / rs->inspacing;
                             y >= base;
                             y -= rs->inspacing, z++)
                        {
                            ((float *)outbuf)[k] +=
                                ((float *)buf)[z] * ((float *)rs->filter)[y];
                        }
                    } else {
                        ((double *)outbuf)[k] = 0;
                        buf = rs->dbuf[!curbuf];
                        for (y = base + z * rs->inspacing,
                                 z = (rs->n_taps + x - y) / rs->inspacing;
                             y > x;
                             y -= rs->inspacing, z++)
                        {
                            ((double *)outbuf)[k] +=
                                ((double *)buf)[z] * ((double *)rs->filter)[y];
                        }
                        buf = rs->dbuf[curbuf];
                        for (z = (x - y) / rs->inspacing;
                             y >= base;
                             y -= rs->inspacing, z++)
                        {
                            ((double *)outbuf)[k] +=
                                ((double *)buf)[z] * ((double *)rs->filter)[y];
                        }
                    }
                    
		    if (++k == outlim) {
			if (write_block != NULL) {
			    if (write_block(outbuf, k, rs->realsize) != k) {
				free(outbuf);
				if (read_block != NULL) {
				    free(inbuf);
				}
				return -1;
			    }
			} else {
			    outbuf = (uint8_t *)outbuf +
                                rs->n_taps * rs->realsize;
			}
			if (islastblock) {
			    if (read_block != NULL) {
				free(inbuf);
			    }
			    if (write_block != NULL) {
				free(outbuf);
			    }
			    return outblocks * rs->n_taps + k;
			}
			outblocks++;
			k = 0;
		    }
		}
		i += rs->inspacing;
		i %= rs->n_taps;
		curbuf = !curbuf;
		z = 0;
	    } else {
		i += rs->inspacing;
	    }
	}
    }
    if (read_block != NULL) {
	free(inbuf);
    }
    if (write_block != NULL && k != 0) {
	if (write_block(outbuf, k, rs->realsize) != k) {
	    free(outbuf);
	    return -1;
	}
    }
    if (write_block != NULL) {
	free(outbuf);
    }    
    return outblocks * rs->n_taps + k;
}
