/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <inttypes.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include <bfmod.h>

#include "textgui.h"
#include "defs.h"
#include "almusvcu.h"
#include "elog.h"
#include "bit.h"
#include "textmenu.h"
#include "textwin.h"
#include "twki.h"

#define STATE_STACK_SIZE 100
#define SCREENSAVER_TIMER 600

#define MIN_XSIZE 20
#define MIN_YSIZE 2
#define ROM_COUNT 8

static int stackpos = -1;
static int screensaver_timer_id = -1;
static struct tw_screen *screensaver_screen = NULL;
static state_func_t state_stack[STATE_STACK_SIZE];
static struct vcu vcu;

struct romdev {
    char *path;
    char *device;
    char *filesys;
    char *name;
    bool_t mount;
    bool_t write;
};

static struct {
    struct romdev rom[ROM_COUNT];
    int n_roms;
    int xsize;
    int ysize;
} conf_static;

static struct {
    double volume_step;
    bool_t autostart;
#define SELECT_OPTION_LOCKED 0
#define SELECT_OPTION_SEL_ON 1
#define SELECT_OPTION_SEL_OFF 2
    int channel_mode_option;
    int eq_mode_option;
    bool_t vol_master;
    int dimmer;
    int vol_groups[VCU_MAX_CHANNEL_GROUPS];
    struct {
        bool_t use;
        bool_t loop;
        int sample_format;
        int channel_count;
        char *path;
    } file;
} conf_user;

static struct {
    struct vcu_eqamp amp[VCU_FREQ_ISO_3OCT_BANDS];
    bool_t mean_mode;
    double peak_amp;
    double delay;
    char name[VCU_EQPRG_NAME_SIZE];
} eq_properties_state;

static struct {
    bool_t modify;
    struct vcu_eqprg *prg;
    struct vcu_eqpart part;
    bool_t did_something;
    bool_t delete;
    int current_eq_index;
} eq_setup_state;

static struct {
    double maxload;
    double memused;
    double sample_rate;
    bool_t callback_active;
} last_sysinfo;

static struct {
    int io;
    int type;
    int idx;
} mix_channels_state;

static struct {
    int io;
    int type;
    int logical_count;
    int physical_count;
    void (*state)(int action);
    char heading[20];
} inout_generic_state;

static struct {
    int n_programs;
    int revidx;
    int curgrp;
    int channel_mode;
    int eq_mode;
    const struct filterprg *fprgs;
    double master_volume;
    bool_t master_mute;
    double group_volume[VCU_MAX_CHANNEL_GROUPS];
    bool_t group_mute[VCU_MAX_CHANNEL_GROUPS];
    double peak;
} runtime_state;

static int access_program_on_rom = -1;
static int current_program_type;
static int current_ambiopole;
static int convolver_last_error = 0;
static int current_channel_group = -1;
static int current_channel_mode = -1;
static int current_eq_mode = -1;
static const struct filterprg *tune_reverb_program_fprg;
static bool_t hard_settings_error_pop = false;
static char *mount_path = NULL;

#define UPDATE_HARD_SETTINGS_CALL                                              \
    if (vcu.ctl.update_hard_settings() != 0) {                                 \
        hard_settings_error_pop = true;                                        \
        textmenu_message("Dropped program due ",                               \
                         "to full memory      ");                              \
        return;                                                                \
    }

#define UPDATE_HARD_SETTINGS_CHILD_POP                                         \
    if (hard_settings_error_pop) {                                             \
        hard_settings_error_pop = false;                                       \
	textmenu_state_delete(mstate);                                         \
	textgui_pop_state();                                                   \
        return;                                                                \
    }

static void
array_bit2bool(const uint32_t bits[],
               bool_t boolean[],
               int len)
{
    int n;
    
    for (n = 0; n < len; n++) {
        boolean[n] = !!bit_isset(bits, n);
    }
}

static void
array_bool2bit(const bool_t boolean[],
               uint32_t bits[],               
               int len)
{
    int n;
    
    for (n = 0; n < len; n++) {
        if (boolean[n]) {
            bit_set(bits, n);
        } else {
            bit_clr(bits, n);
        }
    }
}

static void
set_conf_user_defaults(void)
{
    memset(&conf_user, 0, sizeof(conf_user));
    conf_user.dimmer = 1;
    conf_user.volume_step = 3.0;
    conf_user.vol_master = true;
}

static void
screensaver(void *param)
{
    struct timeval tv;

    if (screensaver_screen == NULL) {
        screensaver_screen = tw_store_screen();
    }
    tw_clear();
    tw_putstring(random() % (tw_xsize() - 8), random() % tw_ysize(),
                 "AlmusVCU");
    tw_refresh();
    tv.tv_usec = 0;
    tv.tv_sec = 3;
    screensaver_timer_id = vcu.ctl.register_timer(NULL, screensaver, &tv);
}

static bool_t
have_channels(void)
{
    int type;

    for (type = 0; type < VCU_CH_COUNT; type++) {
        if (vcu.channel.get_count(VCU_OUT, type) > 0) {
            return true;
        }
    }
    return false;
}

static int
total_channel_count(int io)
{
    int type, ch;
    
    for (type = ch = 0; type < VCU_CH_WP_COUNT; type++) {
        ch += vcu.channel.get_count(io, type);
    }
    return ch;
}

static int
active_group_count(void)
{
    int n, count;

    for (n = 0, count = 0; n < vcu.channel.get_group_count(); n++) {
        if (vcu.runtime.channel.get_group_active(n)) {
            count++;
        }
    }
    return count;
}

static void
state_setup(int action);

void
vcu_progress_init_phase(int phase)
{
    switch (phase) {
    case VCU_INIT_PHASE_CONVOLVER_SUBSYSTEM:
	tw_clear();
	tw_putstring(0, 0, "Convolver subsystem ");
	tw_putstring(0, 1, "is initialising...  ");
	tw_refresh();
	break;
    case VCU_INIT_PHASE_LOAD_REVERB_PROGRAM:
	tw_clear();
	tw_putstring(0, 0, "About to load reverb");
	tw_putstring(0, 1, "programs...         ");
	tw_refresh();
	break;
    case VCU_INIT_PHASE_LOAD_XTC_PROGRAM:
	tw_clear();
	tw_putstring(0, 0, "About to load xtc   ");
	tw_putstring(0, 1, "programs...         ");
	tw_refresh();
	break;
    case VCU_INIT_PHASE_GENERATING_EQUALISERS:
	tw_clear();
	tw_putstring(0, 0, "About to generate   ");
	tw_putstring(0, 1, "equalisers...       ");
	tw_refresh();
	break;
    case VCU_INIT_PHASE_FINISHED:
	textgui_push_state(state_setup);
	break;
    }
}

void
vcu_progress_generate_ambisonic(int filter_index,
                                int n_filters)
{
    elog(LOGERROR, "not implemented");
}

void
vcu_progress_generate_eq(int eq_index,
                         int n_equalisers,
                         int error_code)
{
    int x, y;
    
    /*
      +--------------------+
      |Generating equaliser|
      |1 of 8... failed    |
      |Eq too large        |
      +--------------------+
    */
    if (eq_index == -1) {
        return;
    }
    tw_clear();
    tw_putstring(0, 0, "Generating equaliser");
    if (error_code == 0) {
        tw_printf(0, 1, "%d of %d...", eq_index + 1, n_equalisers);
        tw_refresh();
        usleep(300000);
    } else {
        if (tw_ysize() > 2) {
            tw_printf(0, 1, "%d of %d... failed", eq_index + 1, n_equalisers);
            y = 2;
            x = 0;
        } else {
            y = 1;
            x = 3;
            tw_printf(0, 1, "%d:", eq_index + 1);
        }
        switch (error_code) {
        case VCU_ERR_INVALID_VALUE:
            tw_putstring(x, y, "Invalid eq          ");
            break;
        case VCU_ERR_TRUNCATION_NECESSARY:
            tw_putstring(x, y, "Eq too long         ");
            break;
        case VCU_ERR_OBJECT_TOO_LARGE:
            tw_putstring(x, y, "Eq too large        ");
            break;
        case VCU_ERR_NO_MEMORY:
            tw_putstring(x, y, "Out of memory       ");
            break;
        default:
            tw_putstring(x, y, "Unknown error       ");
            break;
        }
        tw_refresh();
        sleep(5);
    }
}

static char *
make_delaylenstr(char str[],
                 double length,
                 double delay)
{
    if (length == 0.0) {
        sprintf(str, "%.1fms", delay * 1000.0);
    } else {
        if (delay < 0.1) {
            if (length < 0.01) {
                sprintf(str, "%.1fms,%.3fs", delay * 1000.0, length);
            } else if (length < 0.1) {
                sprintf(str, "%.1fms,%.2fs", delay * 1000.0, length);
            } else {
                sprintf(str, "%.1fms,%.1fs", delay * 1000.0, length);
            }
        } else if (delay < 1.0) {
            sprintf(str, "%.0fms,%.1fs", delay * 1000.0, length);
        } else {
            sprintf(str, "%.2fs,%.1fs", delay, length);
        }
    }
    return str;
}

void
vcu_progress_benchmark(int sample_rate,
                       double length,
                       double delay,
                       double load)
{
    char str[100];
    
    tw_clear();
    if (length == 0.0) {
        tw_printf(0, 1, " delay: %s", make_delaylenstr(str, length, delay));
    } else {
        tw_printf(0, 1, " D,Rt: %s", make_delaylenstr(str, length, delay));
    }    
    if (tw_ysize() < 4) {
	/*
	  +--------------------+
	  |Last: 44.1kHz 121%  |
	  | D,Rt 10.20s,10.2s  |
	  +--------------------+
	*/
        if (load == 0.0) {
            tw_putstring(0, 0, "Last: out of memory");
        } else {
            tw_printf(0, 0, "Last: %.1fkHz %d%%",
                      (double)sample_rate / 1000.0, (int)(100.0 * load));
        }
    } else {
	/*
	  +--------------------+
	  |Last settings:      |
	  | D,Rt: 10.20s,10.2s |
	  | 44.1 kHz, 121% load|
	  |Now testing new...  |
	  +--------------------+
	*/
	tw_putstring(0, 0, "Last settings:");
        if (load == 0.0) {
            tw_putstring(0, 2, " Not enough memory");
        } else {
            tw_printf(0, 2, " %.1fkHz, %d%% load",
                      (double)sample_rate / 1000.0, (int)(100.0 * load));
        }
	tw_putstring(0, 3, "Now testing new...");
    }
    tw_refresh();
}

void
vcu_progress_load_program(const struct filterprg *prg,
                          int channel_index,
                          int n_channels,
                          int from_rate,
                          int to_rate,
                          bool_t deriving)
{
    if (tw_ysize() < 4) {
	/*
	  +--------------------+
	  |Berwald hall    1/8 |
	  |resamp. 44.1 to 48.0|
	  +--------------------+
	*/
	if (channel_index == 0) {
	    tw_clear();
	    tw_add_scrollstring(0, 0, prg->name, tw_xsize() - 6, 0, 0,
				false);
	}
	tw_putstring(tw_xsize() - 5, 0, "     ");
        if (deriving) {
            tw_printf(tw_xsize() - 5, 0, "%d/%d", channel_index + 1,
                      n_channels);
            tw_putstring(0, 1, "making filters      ");
        } else {
            tw_printf(tw_xsize() - 5, 0, "%d/%d", channel_index + 1,
                      n_channels);
            tw_putstring(0, 1, "                    ");
            if (from_rate != to_rate) {
                tw_printf(0, 1, "resamp. %.1f to %.1f",
                          (double)from_rate / 1000.0,
                          (double)to_rate / 1000.0);
            } else {
                tw_printf(0, 1, "native %.1f kHz", (double)to_rate / 1000.0);
            }
        }
	tw_refresh();
    } else {
	/*
	  +--------------------+
	  |Berwald hall        |
	  | load impulse 1/8   |
	  | resampling:        |
	  |  44.1 to 48.0 kHz  |
	  +--------------------+
          
	  +--------------------+
	  |Berwald hall        |
	  | making filter 20/20|
	*/
	if (channel_index == 0) {
	    tw_clear();
	    tw_add_scrollstring(0, 0, prg->name, tw_xsize(), 0, 0, false);
	}
	tw_putstring(0, 1, "                    ");
	tw_putstring(0, 2, "                    ");
	tw_putstring(0, 3, "                    ");
        if (deriving) {
            tw_printf(0, 1, " making filter %d/%d", channel_index + 1,
                      n_channels);
        } else {
            tw_printf(0, 1, " load impulse %d/%d", channel_index + 1,
                      n_channels);
            if (from_rate != to_rate) {
                tw_putstring(0, 2, " resampling:");
                tw_printf(0, 3, "  %.1f to %.1f kHz",
                          (double)from_rate / 1000.0,
                          (double)to_rate / 1000.0);
            } else {
                tw_printf(0, 2, " native %.1f kHz", (double)to_rate / 1000.0);
            }
        }
	tw_refresh();
    }    
}

static void
state_runtime_main_screen(int action);

static void
cpu_load_callback(void *param)
{
    struct timeval tv;
    int load;

    if (state_stack[stackpos] != state_runtime_main_screen) {
        last_sysinfo.callback_active = false;
        return;
    }
    last_sysinfo.maxload = vcu.sysinfo.get_cpu_maxload();
    last_sysinfo.memused = vcu.sysinfo.get_ram_used();
    last_sysinfo.sample_rate = vcu.runtime.signal.get_input_sample_rate();    
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    vcu.ctl.register_timer(NULL, cpu_load_callback, &tv);
    if (screensaver_screen == NULL) {
        load = (int)(last_sysinfo.maxload * 100.0);
        if (load == 100) {
            load = 99;
        }
        tw_printf(0, 3, "%02d/%.1f", load,
                  (double)last_sysinfo.sample_rate / 1000.0);
        tw_refresh();
    }
}

static void
peak_meter_callback(double peak)
{
    runtime_state.peak = peak;
    if (screensaver_screen == NULL) {
        state_stack[stackpos](ACTION_IGNORE);
    }
}

static void
convolver_terminated(int reason)
{
    bool_t doexit = false;
    
    elog(LOGDEBUG, "reason: %d", reason);
    convolver_last_error = reason;
    do {
        if (stackpos == 1) {
            doexit = true;
        }
	state_stack[stackpos](ACTION_EXIT);
    } while (!doexit);
}

static bool_t
load_reverb_program_helper(const struct filterprg *fprg,
                           bool_t copy_to_cache,
                           bool_t *make_copy,
                           bool_t *copy_ask)
{
    int err;

    *copy_ask = false;
    switch ((err = vcu.reverb.load_program(fprg, copy_to_cache))) {
    case VCU_ERR_NO_MEMORY:
	textmenu_message("Not enough memory to",
			 "load the program    ");
	return false;
    case VCU_ERR_READ_FAILED:
    case VCU_ERR_WRITE_FAILED:
	textmenu_message("Could not store the ",
			 "program in cache    ");
	return false;
    case VCU_ERR_OBJECT_TOO_LARGE:
	textmenu_message("Too large for cache ",
			 "(loaded anyway)     ");
        return false;
    case VCU_ERR_OBJECT_EXISTS:
	textmenu_yesno(make_copy,
                       "Already loaded. Make",
                       "a copy?             ");
        *copy_ask = true;
        return true;
    case VCU_ERR_INVALID_FORMAT:
	textmenu_message("The reverb program  ",
			 "is broken           ");
	return false;
    case VCU_ERR_LOAD_FAILED:
	textmenu_message("Could not load the  ",
			 "reverb program      ");
	return false;
    case VCU_ERR_RESAMPLING_FAILED:
	textmenu_message("Could not resample  ",
			 "reverb program      ");
	return false;
    case VCU_ERR_INVALID_VALUE:
    default:
	if (err >= 0) {
            textmenu_message("Reverb prg success- ",
                             "fully loaded        ");
	    return true;
	}
	elog(LOGDEBUG, "unexpected error: %d", err);
	textmenu_message("An error likely due ",
			 "to a bug occured    ");
	return false;
    }
    return true;
}

static bool_t
load_xtc_program_helper(const struct filterprg *fprg,
                        bool_t copy_to_cache)
{
    int err;
    
    tw_clear();
    tw_putstring(0, 0, "Loading crosstalk  ");
    tw_putstring(0, 1, "cancellation prg...");
    tw_refresh();
    
    switch ((err = vcu.ambiopole.load_program(fprg, copy_to_cache))) {
    case VCU_ERR_NO_MEMORY:
	textmenu_message("Not enough memory to",
			 "load the program    ");
	return false;
    case VCU_ERR_READ_FAILED:
    case VCU_ERR_WRITE_FAILED:
	textmenu_message("Could not store the ",
			 "program in cache    ");
	return false;
    case VCU_ERR_TRUNCATION_NECESSARY:
	textmenu_message("The xtc program is  ",
			 "too long            ");
        return false;
    case VCU_ERR_OBJECT_TOO_LARGE:
	textmenu_message("Program too large to",
			 "fit in cache        ");
        return false;
    case VCU_ERR_OBJECT_EXISTS:
	textmenu_message("The xtc program is  ",
			 "already loaded      ");
        return false;
    case VCU_ERR_INVALID_FORMAT:
	textmenu_message("The xtc program is  ",
			 "broken              ");
	return false;
    case VCU_ERR_LOAD_FAILED:
	textmenu_message("Could not load the  ",
			 "xtc program         ");
	return false;
    case VCU_ERR_RESAMPLING_FAILED:
	textmenu_message("Could not resample  ",
			 "xtc program         ");
	return false;
    case VCU_ERR_INVALID_VALUE:
    default:
	if (err >= 0) {
            textmenu_message("xtc program success-",
                             "fully loaded        ");
	    return true;
	}
	elog(LOGDEBUG, "unexpected error: %d", err);
	textmenu_message("An error likely due ",
			 "to a bug occured    ");
	return false;
    }
    return true;
}

static bool_t
load_eq_program_helper(const struct filterprg *fprg,
                       bool_t copy_to_cache)
{
    int err;
    
    tw_clear();
    tw_putstring(0, 0, "Loading equalisa-  ");
    tw_putstring(0, 1, "tion program...    ");
    tw_refresh();
    
    switch ((err = vcu.eq.load_program(fprg, copy_to_cache))) {
    case VCU_ERR_NO_MEMORY:
	textmenu_message("Not enough memory to",
			 "load the program    ");
	return false;
    case VCU_ERR_READ_FAILED:
    case VCU_ERR_WRITE_FAILED:
	textmenu_message("Could not store the ",
			 "program in cache    ");
	return false;
    case VCU_ERR_TRUNCATION_NECESSARY:
	textmenu_message("The eq program is   ",
			 "too long            ");
        return false;
    case VCU_ERR_OBJECT_TOO_LARGE:
	textmenu_message("Program too large to",
			 "fit in cache        ");
        return false;
    case VCU_ERR_OBJECT_EXISTS:
	textmenu_message("The eq program is   ",
			 "already loaded      ");
        return false;
    case VCU_ERR_INVALID_FORMAT:
	textmenu_message("The eq program is   ",
			 "broken              ");
	return false;
    case VCU_ERR_LOAD_FAILED:
	textmenu_message("Could not load the  ",
			 "eq program          ");
	return false;
    case VCU_ERR_RESAMPLING_FAILED:
	textmenu_message("Could not resample  ",
			 "eq program          ");
	return false;
    case VCU_ERR_INVALID_VALUE:
    default:
	if (err >= 0) {
            textmenu_message("Eq program success- ",
                             "fully loaded        ");
	    return true;
	}
	elog(LOGDEBUG, "unexpected error: %d", err);
	textmenu_message("An error likely due ",
			 "to a bug occured    ");
	return false;
    }
    return true;
}

static bool_t
mount_device_helper(int idx)
{
    char mount_error_str[40];
    
    /* Access to the ROM unit is expected to be slow (and mount can block)
       so we print this message first */
    tw_clear();
    tw_printf(0, 0, "Wait while %s", conf_static.rom[idx].name);
    tw_printf(0, 1, "is accessed...");
    tw_refresh();


    if (!conf_static.rom[idx].mount) {
	return true;
    }
 mount_try:
    if (mount(conf_static.rom[idx].device,
	      conf_static.rom[idx].path,
	      conf_static.rom[idx].filesys,
              (conf_static.rom[idx].write ? 0 : MS_RDONLY) |
              MS_SYNCHRONOUS, NULL) == -1)
    {
	switch (errno) {
	case ELOOP:
	case ENOENT:
	case ENODEV:
	case ENOTBLK:
	case ENOTDIR:
	case ENAMETOOLONG:
	case EINVAL:		    
	    textmenu_message("Invalid static conf ",
			     "for device          ");
	    return false;
	case EPERM:
	    textmenu_message("Insufficient rights ",
			     "to mount device     ");
	    return false;
	case EBUSY:
	    if (umount(conf_static.rom[idx].path) == -1) {
		textmenu_message("Device is busy      ",
				 "                    ");
	    } else {
		goto mount_try;
	    }
	    return false;
	case ENOMEDIUM:
	    textmenu_message("There is no medium  ",
			     "in the device       ");
	    return false;
	case EMEDIUMTYPE:
	    textmenu_message("Wrong medium type in",
			     "the device          ");
	    return false;
        case ENXIO:
	    textmenu_message("Unsupported medium  ",
			     "format              ");
	    return false;
	default:
	    sprintf(mount_error_str, "mount errno: %d", errno);
	    textmenu_message(mount_error_str,
			     "                    ");
	    return false;
	}
	return false;
    }
    mount_path = conf_static.rom[idx].path;
    return true;
}

static void
state_runtime_reverb_setup(int action);

static void
state_runtime_channel_mode(int action);

static void
state_runtime_eq_mode(int action);

static int
handle_runtime_key(int action)
{
    int max_count, grp;

    elog(LOGDEBUG, "handle runtime key %d", action);

    grp = runtime_state.curgrp;
    switch (action) {
    case KI_KEY_MVOL_UP:
        vcu.runtime.volume.set_master(runtime_state.master_volume +
                                      conf_user.volume_step);
        runtime_state.master_volume = vcu.runtime.volume.get_master();
        vcu.runtime.signal.reset_peak_meter();
        break;
    case KI_KEY_MVOL_DOWN:
        vcu.runtime.volume.set_master(runtime_state.master_volume -
                                      conf_user.volume_step);
        runtime_state.master_volume = vcu.runtime.volume.get_master();
        vcu.runtime.signal.reset_peak_meter();
        break;
    case KI_KEY_GVOL_UP:
        if (runtime_state.curgrp < 0) {
            return ACTION_IGNORE;
        }
        vcu.runtime.volume.set_group(grp, runtime_state.group_volume[grp] +
                                     conf_user.volume_step);
        runtime_state.group_volume[grp] = vcu.runtime.volume.get_group(grp);
        vcu.runtime.signal.reset_peak_meter();
        break;
    case KI_KEY_GVOL_DOWN:
        if (runtime_state.curgrp < 0) {
            return ACTION_IGNORE;
        }
        vcu.runtime.volume.set_group(grp, runtime_state.group_volume[grp] -
                                     conf_user.volume_step);
        runtime_state.group_volume[grp] = vcu.runtime.volume.get_group(grp);
        vcu.runtime.signal.reset_peak_meter();
        break;
    case KI_KEY_REVERB_PRG_NEXT:
        if (state_stack[stackpos] == state_runtime_reverb_setup) {
            return ACTION_IGNORE;
        }
        if (runtime_state.revidx < runtime_state.n_programs - 1) {
            runtime_state.revidx++;
        } else {
            runtime_state.revidx = 0;
        }
        vcu.runtime.reverb.set_id(runtime_state.fprgs[runtime_state.revidx].id);
        vcu.runtime.signal.reset_peak_meter();
        break;
    case KI_KEY_CHANNEL_MODE_NEXT:
        if (runtime_state.channel_mode < 0 ||
            state_stack[stackpos] == state_runtime_channel_mode)
        {
            return ACTION_IGNORE;
        }
        max_count = vcu.channel.get_mode_count();
        runtime_state.channel_mode++;
        if (runtime_state.channel_mode == max_count) {
            runtime_state.channel_mode = 0;
        }
        vcu.runtime.channel.set_mode(runtime_state.channel_mode);
        break;
    case KI_KEY_EQ_MODE_NEXT:
        if (!vcu.runtime.eq.get_mode_active(-1) ||
            state_stack[stackpos] == state_runtime_channel_mode)
        {
            return ACTION_IGNORE;
        }
        max_count = vcu.eq.get_mode_count();
    key_eq_mode_next_restart:
        runtime_state.eq_mode++;
        if (runtime_state.eq_mode == max_count) {
            runtime_state.eq_mode = -1;
        }
        if (!vcu.runtime.eq.get_mode_active(runtime_state.eq_mode)) {
            goto key_eq_mode_next_restart;
        }
        vcu.runtime.eq.set_mode(runtime_state.eq_mode);
        break;
    case KI_KEY_GROUP_NEXT:
        if (runtime_state.curgrp < 0) {
            return ACTION_IGNORE;
        }
    key_group_next_restart:
        for (grp = runtime_state.curgrp + 1;
             grp < vcu.channel.get_group_count();
             grp++)
        {
            if (vcu.runtime.channel.get_group_active(grp) &&
                conf_user.vol_groups[grp])
            {
                break;
            }
        }
        if (grp < vcu.channel.get_group_count()) {
            runtime_state.curgrp = grp;            
        } else {
            runtime_state.curgrp = -1;
            goto key_group_next_restart;
        }
        break;
    case KI_KEY_MASTER_MUTE:
        runtime_state.master_mute = !runtime_state.master_mute;
        vcu.runtime.volume.set_master_temp_mute(runtime_state.master_mute);
        break;
    case KI_KEY_GROUP_MUTE:
        if (runtime_state.curgrp < 0) {
            return ACTION_IGNORE;
        }
        runtime_state.group_mute[grp] = !runtime_state.group_mute[grp];
        vcu.runtime.volume.set_group_mute(grp, runtime_state.group_mute[grp]);
        break;
    default:
        return action;
    }
    return ACTION_REFRESH;
}

static void
state_runtime_reverb_setup(int action)
{
    static struct textmenu_state *mstate = NULL;
    static double rev_delay, rev_volume;
    static double balance[3];
    
    if (action == ACTION_INIT) {
	void *entries[20];
	double range[2];
	prgid_t id;
	int pos;
	
	id = vcu.runtime.reverb.get_id();
	pos = 0;
        if (runtime_state.n_programs > 1) {
            entries[pos++] = "Select next prg?";
        }
	rev_delay = vcu.runtime.reverb.get_trim_delay() * 1000.0;
	entries[pos] = alloca(80);
        vcu.runtime.reverb.get_trim_delay_range(range);
	sprintf((char *)entries[pos++], "Delay [%%.1f,1.0,10,%f,%f]ms",
                range[0] * 1000.0, range[1] * 1000.0);
	entries[pos++] = &rev_delay;
	vcu.runtime.reverb.get_trim_volume_range(range);
	rev_volume = vcu.runtime.reverb.get_trim_volume();
	entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Trim vol [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &rev_volume;

        vcu.runtime.reverb.get_xyz_balance_range(range);
        vcu.runtime.reverb.get_xyz_balance(balance);
	entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Xbalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[0];
        entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Ybalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[1];
	entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Zbalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[2];

	entries[pos++] = NULL;
	mstate = textmenu_state_new(runtime_state.fprgs
                                    [runtime_state.revidx].longname, entries);
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action == 0 && runtime_state.n_programs > 1) {
            runtime_state.revidx = (runtime_state.revidx + 1) %
                runtime_state.n_programs;
            vcu.runtime.reverb.set_id(runtime_state.fprgs
                                      [runtime_state.revidx].id);
            textmenu_state_delete(mstate);
            state_runtime_reverb_setup(ACTION_INIT);
            return;
        }
        if (runtime_state.n_programs == 1) {
            action += 1;
        }
        switch (action) {
        case 1:
            vcu.runtime.reverb.set_trim_delay(rev_delay / 1000.0);
            return;
        case 2:
            vcu.runtime.reverb.set_trim_volume(rev_volume);
            return;
        case 3: case 4: case 5:
            vcu.runtime.reverb.set_xyz_balance(balance);
            return;
        }
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_runtime_apole_setup_apole(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprgs;
    static int n_programs;
    static int idx;

    switch (action) {
    case ACTION_INIT: {
	void *entries[20];
	prgid_t id;
	int pos;

	fprgs = vcu.ambiopole.get_programs(&n_programs);
        
        if (!vcu.ambiopole.get_xtc_active(current_ambiopole) ||
            n_programs <= 1)
        {
	    mstate = NULL;
	    textmenu_message("There is nothing to ",
			     "configure           ");
	    return;
        }
        
	id = vcu.runtime.ambiopole.get_id(current_ambiopole);
	for (idx = 0; memcmp(&id, &fprgs[idx].id, sizeof(id)) != 0; idx++);
	pos = 0;
	entries[pos++] = "Select next prg?";
	entries[pos++] = NULL;
	mstate = textmenu_state_new(fprgs[idx].longname, entries);
	break;
    }
    default:
	break;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
	idx = (idx + 1) % n_programs;
        elog(LOGDEBUG, "ambiopole prg index %d (%d programs)", idx, n_programs);
	vcu.runtime.ambiopole.set_id(current_ambiopole, fprgs[idx].id);
	textmenu_state_delete(mstate);
	state_runtime_apole_setup_apole(ACTION_INIT);
        break;
    default:
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_runtime_apole_setup(int action)
{
    static struct textmenu_state *mstate;
    static int pos2idx[VCU_MAX_CHANNELS/2];
    static int n_ambiopoles;

    switch (action) {
    case ACTION_CHILD_POP:
        if (n_ambiopoles == 1) {
            textgui_pop_state();
            return;
        }
        break;
    case ACTION_INIT: {
	int pos, n, count;
	void **entries;

        count = vcu.channel.get_count(VCU_OUT, VCU_CH_AMBIOPOLE) / 2;
        entries = alloca((1 + 2 * count) * sizeof(void *));
        pos = 0;
        n_ambiopoles = 0;
        for (n = 0; n < count; n++) {
            entries[pos] = alloca(40);
            if (!vcu.runtime.channel.get_active(VCU_OUT, VCU_CH_AMBIOPOLE, 
                                                2 * n) &&
                !vcu.runtime.channel.get_active(VCU_OUT, VCU_CH_AMBIOPOLE,
                                                2 * n + 1))
            {
                continue;
            }
            pos2idx[pos>>1] = n;
            sprintf((char *)entries[pos++], "Ambiopole %d...", n + 1);
            entries[pos++] = state_runtime_apole_setup_apole;
            n_ambiopoles++;
        }

        if (n_ambiopoles == 1) {
            current_ambiopole = pos2idx[0];
            mstate = NULL;
            textgui_push_state(state_runtime_apole_setup_apole);
            return;
        }
        
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }

    current_ambiopole = pos2idx[mstate->pos];
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_eq_adjust_fft(int action);

static void
state_runtime_dynamic_eq(int action)
{
    static struct textmenu_state *mstate;
    static const int *indexes;

    const struct vcu_eqprg *eqprg;
    int n;

    switch (action) {
    case ACTION_INIT: {
	int pos, n_dyneq;
	void **entries;

        mstate = NULL;
        indexes = vcu.runtime.eq.get_dynamic_equalisers(&n_dyneq, true);
        memset(&eq_setup_state, 0, sizeof(eq_setup_state));
        if (n_dyneq == 1) {
            eqprg = vcu.eq.get_equaliser(indexes[0]);
            eq_setup_state.current_eq_index = indexes[0];
            eq_setup_state.part = eqprg->parts[0];
            textgui_push_state(state_eq_adjust_fft);
            return;
        }
        entries = alloca((1 + 2 * n_dyneq) * sizeof(void *));
        pos = 0;
        for (n = 0; n < n_dyneq; n++) {
            eqprg = vcu.eq.get_equaliser(indexes[n]);
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++], "%s...", eqprg->name);
            entries[pos++] = state_eq_adjust_fft;
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        return;
    default:
	break;
    }

    eq_setup_state.current_eq_index = indexes[mstate->pos];
    eqprg = vcu.eq.get_equaliser(indexes[mstate->pos]);
    eq_setup_state.part = eqprg->parts[0];
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_runtime_channel_mode(int action)
{
    static struct textmenu_state *mstate;
    const struct vcu_chmode *chmode;

    if (action == ACTION_INIT) {
        char s[VCU_CHMODE_NAME_SIZE];
	void *entries[20];
	int pos;
        
	pos = 0;
	entries[pos++] = "Next channel mode?";
	entries[pos++] = NULL;
        chmode = vcu.channel.get_mode(runtime_state.channel_mode);
        strcpy(s, chmode->name);
        s[0] = toupper(s[0]);
	mstate = textmenu_state_new(s, entries);
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        if (runtime_state.channel_mode < vcu.channel.get_mode_count() - 1) {
            runtime_state.channel_mode++;
        } else {
            runtime_state.channel_mode = 0;
        }
        vcu.runtime.channel.set_mode(runtime_state.channel_mode);
	textmenu_state_delete(mstate);
	state_runtime_channel_mode(ACTION_INIT);
        break;
    default:
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_runtime_eq_mode(int action)
{
    static struct textmenu_state *mstate;
    static int eq_mode;

    eq_mode = runtime_state.eq_mode + 1;
    
    if (action == ACTION_INIT) {
        char s[VCU_EQMODE_NAME_SIZE + 30];
        const struct vcu_eqmode *eqmode;
	void *entries[20];
	int pos, n;

	pos = 0;
        if (vcu.eq.get_mode_count() == 1) {
            entries[pos++] = "Equalisation [off/on]";
            entries[pos++] = &eq_mode;
        } else {
            entries[pos++] = "Next eq mode?";
        }
        if (runtime_state.eq_mode < 0) {
            strcpy(s, "Equalisation off");
        } else {
            vcu.runtime.eq.get_dynamic_equalisers(&n, true);
            if (n > 0) {
                entries[pos++] = "Dynamic eq...";
                entries[pos++] = state_runtime_dynamic_eq;
            }
            eqmode = vcu.eq.get_mode(runtime_state.eq_mode);
            strcpy(s, eqmode->name);
            s[0] = toupper(s[0]);
        }
        entries[pos++] = NULL;
        if (vcu.eq.get_mode_count() == 1) {
            mstate = textmenu_state_new(NULL, entries);
        } else {
            mstate = textmenu_state_new(s, entries);
        }
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        runtime_state.eq_mode++;
        while (runtime_state.eq_mode < vcu.eq.get_mode_count() &&
               !vcu.runtime.eq.get_mode_active(runtime_state.eq_mode))
        {
            runtime_state.eq_mode++;
        }
        if (runtime_state.eq_mode == vcu.eq.get_mode_count()) {
            runtime_state.eq_mode = -1;
        }
        elog(LOGDEBUG, "Setting eq_mode to %d", runtime_state.eq_mode);
        vcu.runtime.eq.set_mode(runtime_state.eq_mode);
	textmenu_state_delete(mstate);
	state_runtime_eq_mode(ACTION_INIT);
        break;
    default:
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_runtime_offset_volume(int action)
{
    static struct textmenu_state *mstate;
    static int group_mute_pos[VCU_MAX_CHANNEL_GROUPS];
    static int group_vol_pos[VCU_MAX_CHANNEL_GROUPS];
    int n;

    if (action == ACTION_INIT) {
	void *entries[200 + 4 * VCU_MAX_CHANNEL_GROUPS];
        const struct vcu_chgrp *chgrp;
	double range[2];
	int pos, count;

	pos = 0;
        vcu.runtime.volume.get_range(range);
        count = vcu.channel.get_group_count();
        for (n = 0; n < count; n++) {
            if (!vcu.runtime.channel.get_group_active(n)) {
                group_vol_pos[n] = ACTION_IGNORE;
                continue;
            }
            chgrp = vcu.channel.get_group(n);
            group_vol_pos[n] = pos / 2;
            entries[pos] = alloca(40 + VCU_CHGRP_NAME_SIZE);
            sprintf((char *)entries[pos], "%s [%%+.1f,0.1,10,%f,%f]dB",
                    chgrp->name, range[0], range[1]);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &runtime_state.group_volume[n];
        }
        for (n = 0; n < count; n++) {
            if (!vcu.runtime.channel.get_group_active(n)) {
                group_mute_pos[n] = ACTION_IGNORE;
                continue;
            }
            chgrp = vcu.channel.get_group(n);
            group_mute_pos[n] = pos / 2;
            entries[pos] = alloca(40 + VCU_CHGRP_NAME_SIZE);
            sprintf((char *)entries[pos], "%s mute [off/on]", chgrp->name);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &runtime_state.group_mute[n];
        }
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            if (action == group_mute_pos[n]) {
                vcu.runtime.volume.set_group_mute
                    (n, runtime_state.group_mute[n]);
            } else if (action == group_vol_pos[n]) {
                vcu.runtime.volume.set_group(n, runtime_state.group_volume[n]);
            }
        }
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_runtime_info(int action);

static void
state_runtime_info_callback(void *param)
{
    state_runtime_info(ACTION_REFRESH);
}

static void
state_runtime_info(int action)
{
    static int timer_id = -1;
    
    double maxload, memused, rate, cputemp, mbtemp;
    struct timeval tv;

    switch (action) {
    case ACTION_INIT:
    case ACTION_REFRESH:
        if (state_stack[stackpos] != state_runtime_info) {
            return;
        }
        break;
    case ACTION_EXIT:
    case KI_KEY_LEFT:
        vcu.ctl.unregister_timer(timer_id);
        textgui_pop_state();
    default:
        return;
    }
    maxload = vcu.sysinfo.get_cpu_maxload();
    memused = vcu.sysinfo.get_ram_used();
    cputemp = vcu.sysinfo.get_temp_cpu();
    mbtemp = vcu.sysinfo.get_temp_motherboard();
    rate = vcu.runtime.signal.get_input_sample_rate();
    if (mbtemp == VCU_SYSINFO_NA) {
        mbtemp = 0.0;
    }
    if (cputemp == VCU_SYSINFO_NA) {
        cputemp = 0.0;
    }
    if (screensaver_screen != NULL) {
        return;
    }
    tw_clear();
    if (tw_ysize() < 4) {
        /*
          +--------------------+
          |CPU 66%/66C  MB 99C |
          |Rate 44.1kHz Mem 99%|
          +--------------------+
        */
        tw_printf(0, 0, "CPU %.2d%%/%dC", (int)(maxload * 100.0), (int)cputemp);
        tw_printf(13, 0, "MB %dC", (int)mbtemp);
        tw_printf(0, 1, "Rate %.1fkHz", rate / 1000.0);
        tw_printf(13, 1, "Mem %.2d%%", (int)(memused * 100.0));
    } else {
        /*
          +--------------------+
          |Load CPU/Mem 99%/99%|
          |Temp CPU/MB 99C/99C |
          |Sample rate 44.1kHz |
          |                    |
          +--------------------+
        */
        tw_printf(0, 0, "Load CPU/Mem %.2d%%/%.2d%%",
                  (int)(maxload * 100.0), (int)(memused * 100.0));
        tw_printf(0, 1, "Temp CPU/MB %dC/%dC", (int)cputemp, (int)mbtemp);
        tw_printf(0, 2, "Sample rate %.1fkHz", rate / 1000.0);
    }
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    timer_id = vcu.ctl.register_timer(NULL, state_runtime_info_callback, &tv);
    tw_refresh();
}

static void
state_trim_volume_in(int action);

static void
state_trim_volume_out(int action);

static void
state_trim_delay_in(int action);

static void
state_trim_delay_out(int action);

static void
state_runtime_channel_control(int action)
{
    static struct textmenu_state *mstate;
    static int master_mute_pos, group_mute_pos[VCU_MAX_CHANNEL_GROUPS];
    static int group_vol_pos[VCU_MAX_CHANNEL_GROUPS];
    static int menupos = 0, channel_mode_pos = ACTION_IGNORE;
    int n;

    switch (action) {
    case ACTION_CHILD_POP:
        if (menupos == channel_mode_pos) {
            textmenu_state_delete(mstate);
            state_runtime_channel_control(ACTION_INIT);
            return;
        }
        break;
    case ACTION_INIT: {
	void *entries[200 + 4 * VCU_MAX_CHANNEL_GROUPS];
        const struct vcu_chgrp *chgrp;
	double range[2];
	int pos, count;

	pos = 0;

        /* volume and mute */
        vcu.runtime.volume.get_range(range);
        entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Master [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &runtime_state.master_volume;

        count = vcu.channel.get_group_count();
        if (active_group_count() == 1) {
            for (n = 0; n < count; n++) {
                group_vol_pos[n] = ACTION_IGNORE;
                group_mute_pos[n] = ACTION_IGNORE;
            }
            count = 0;
        }
        for (n = 0; n < count; n++) {
            if (!vcu.runtime.channel.get_group_active(n)) {
                group_vol_pos[n] = ACTION_IGNORE;
                continue;
            }
            chgrp = vcu.channel.get_group(n);
            group_vol_pos[n] = pos / 2;
            entries[pos] = alloca(40 + VCU_CHGRP_NAME_SIZE);
            sprintf((char *)entries[pos], "%s [%%+.1f,0.1,10,%f,%f]dB",
                    chgrp->name, range[0], range[1]);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &runtime_state.group_volume[n];
        }
        
	master_mute_pos = pos / 2;
	entries[pos++] = "Master mute [off/on]";
	entries[pos++] = &runtime_state.master_mute;

        for (n = 0; n < count; n++) {
            if (!vcu.runtime.channel.get_group_active(n)) {
                group_mute_pos[n] = ACTION_IGNORE;
                continue;
            }
            chgrp = vcu.channel.get_group(n);
            group_mute_pos[n] = pos / 2;
            entries[pos] = alloca(40 + VCU_CHGRP_NAME_SIZE);
            sprintf((char *)entries[pos], "%s mute [off/on]", chgrp->name);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &runtime_state.group_mute[n];
        }
        
        /* sub menus */
        if (runtime_state.channel_mode >= 0 &&
            vcu.channel.get_mode_count() > 0 &&
            conf_user.channel_mode_option != SELECT_OPTION_LOCKED)
        {
            if (menupos == channel_mode_pos) {
                menupos = pos / 2;
            }
            channel_mode_pos = pos / 2;
            entries[pos++] = "Channel mode...";
            entries[pos++] = state_runtime_channel_mode;            
        }

	if (vcu.runtime.eq.get_mode_active(-1) &&
            conf_user.eq_mode_option != SELECT_OPTION_LOCKED)
        {
            entries[pos++] = "Eq mode...";
            entries[pos++] = state_runtime_eq_mode;
        }
        
        if (vcu.runtime.channel.get_type_active(VCU_CH_AMBIOPOLE)) {
	    entries[pos++] = "A-pole setup...";
	    entries[pos++] = state_runtime_apole_setup;
	}
        
	if (vcu.runtime.channel.get_type_active(VCU_CH_REVERB)) {
	    entries[pos++] = "Reverb setup...";
	    entries[pos++] = state_runtime_reverb_setup;
	}

        if (runtime_state.channel_mode >= 0 &&
            vcu.channel.get_mode_count() > 0 &&
            active_group_count() == 1)
        {
	    entries[pos++] = "Offset volume...";
	    entries[pos++] = state_runtime_offset_volume;
        }

	entries[pos++] = "Trim volume in...";
	entries[pos++] = state_trim_volume_in;
	entries[pos++] = "Trim delay in...";
	entries[pos++] = state_trim_delay_in;
	entries[pos++] = "Trim volume out...";
	entries[pos++] = state_trim_volume_out;
	entries[pos++] = "Trim delay out...";
	entries[pos++] = state_trim_delay_out;
	entries[pos++] = "Runtime info...";
	entries[pos++] = state_runtime_info;
	entries[pos++] = NULL;
        
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
	break;
    }
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            if (action == group_mute_pos[n]) {
                vcu.runtime.volume.set_group_mute(
                    n, runtime_state.group_mute[n]);
            } else if (action == group_vol_pos[n]) {
                vcu.runtime.volume.set_group(n, runtime_state.group_volume[n]);
            }
        }
	if (action == 0) {
	    vcu.runtime.volume.set_master(runtime_state.master_volume);
	} else if (action == master_mute_pos) {
	    vcu.runtime.volume.set_master_temp_mute(runtime_state.master_mute);
        }
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        channel_mode_pos = ACTION_IGNORE;
        menupos = 0;
	return;
    }    
}

#define CONTROL_MODE_MIN 0
#define CONTROL_MODE_PICK_REVERB 0
#define CONTROL_MODE_ALTER_MASTER_VOL 1
#define CONTROL_MODE_ALTER_GROUP_VOL 2
#define CONTROL_MODE_SELECT_EQ_MODE 3
#define CONTROL_MODE_SELECT_CHANNEL_MODE 4
#define CONTROL_MODE_SELECT_MENU 5
#define CONTROL_MODE_MAX 5

#define SELECT_MENU_MIN 0
#define SELECT_MENU_MAIN_SCREEN 0
#define SELECT_MENU_CHANNEL_CONTROL 1
#define SELECT_MENU_SETUP 2
#define SELECT_MENU_MAX 2

static void
runtime_screen_update_top_line(int control_mode,
                               int select_menu,
                               bool_t enlarge_reverb)
{
    const struct vcu_chmode *chmode;
    const struct vcu_eqmode *eqmode;
    char s[100], *p;
    int count;
    
    switch (control_mode) {
    case CONTROL_MODE_PICK_REVERB:
    case CONTROL_MODE_ALTER_MASTER_VOL:
    case CONTROL_MODE_ALTER_GROUP_VOL:
        if (vcu.runtime.channel.get_type_active(VCU_CH_REVERB) &&
            (control_mode == CONTROL_MODE_PICK_REVERB ||
             runtime_state.channel_mode < 0))
        {
            strncpy(s, runtime_state.fprgs[runtime_state.revidx].name, 21);
            s[20] = '\0';
            p = s;
            if (control_mode == CONTROL_MODE_PICK_REVERB || enlarge_reverb) {
                while(*p != '\0') {
                    *p = toupper(*p);
                    p++;
                }
            }
            tw_putstring((20 - strlen(s)) / 2, 0, s);
        } else {
            if (runtime_state.channel_mode < 0) {
                tw_putstring(0, 0, "      AlmusVCU      ");
            } else {
                chmode = vcu.channel.get_mode(runtime_state.channel_mode);
                strncpy(s, chmode->name, 21);
                s[20] = '\0';
                s[0] = toupper(s[0]);
                tw_putstring((20 - strlen(s)) / 2, 0, s);
            }
        }
        break;
    case CONTROL_MODE_SELECT_CHANNEL_MODE:
        chmode = vcu.channel.get_mode(runtime_state.channel_mode);
        strncpy(s, chmode->name, 21);
        s[20] = '\0';
        p = s;
        while(*p != '\0') {
            *p = toupper(*p);
            p++;
        }
        tw_putstring((20 - strlen(s)) / 2, 0, s);
        break;
    case CONTROL_MODE_SELECT_EQ_MODE:
        if (runtime_state.channel_mode >= 0) {
            chmode = vcu.channel.get_mode(runtime_state.channel_mode);
            count = bit_count(chmode->eqmodes, vcu.eq.get_mode_count());
        } else {
            count = vcu.eq.get_mode_count();
        }
        if (runtime_state.eq_mode < 0) {
            strcpy(s, "EQUALISATION OFF");
        } else if (count == 1) {
            strcpy(s, "EQUALISATION ON");
        } else {
            eqmode = vcu.eq.get_mode(runtime_state.eq_mode);
            strncpy(s, eqmode->name, 21);
            s[20] = '\0';
            p = s;
            while(*p != '\0') {
                *p = toupper(*p);
                p++;
            }
        }
        tw_putstring((20 - strlen(s)) / 2, 0, s);
        break;
    case CONTROL_MODE_SELECT_MENU:
        switch (select_menu) {
        case SELECT_MENU_MAIN_SCREEN:
            tw_putstring(0, 0, "    MAIN SCREEN     ");
            break;
        case SELECT_MENU_CHANNEL_CONTROL:
            tw_putstring(0, 0, "  CHANNEL CONTROL   ");
            break;
        case SELECT_MENU_SETUP:
            tw_putstring(0, 0, "       SETUP        ");
            break;
        }
        break;
    }
}

static void
state_runtime_main_screen(int action)
{
    static int control_mode, select_menu;
    static bool_t use_master;

    const struct vcu_chgrp *chgrp;
    const struct vcu_chmode *chmode;
    struct timeval tv;
    bool_t breaknow;
    int n, count;
    char s[100];
    prgid_t id;

    switch (action) {
    case KI_KEY_UP:
	switch (control_mode) {
	case CONTROL_MODE_PICK_REVERB:
            handle_runtime_key(KI_KEY_REVERB_PRG_NEXT);
	    break;
	case CONTROL_MODE_ALTER_MASTER_VOL:
            handle_runtime_key(KI_KEY_MVOL_UP);
	    break;
	case CONTROL_MODE_ALTER_GROUP_VOL:
            vcu.runtime.volume.set_group
                (runtime_state.curgrp,
                 runtime_state.group_volume[runtime_state.curgrp] +
                 conf_user.volume_step);
            runtime_state.group_volume[runtime_state.curgrp] =
                vcu.runtime.volume.get_group(runtime_state.curgrp);
	    break;
        case CONTROL_MODE_SELECT_CHANNEL_MODE:
	    if (runtime_state.channel_mode < vcu.channel.get_mode_count() - 1) {
		runtime_state.channel_mode++;
	    } else {
		runtime_state.channel_mode = 0;
	    }
            vcu.runtime.channel.set_mode(runtime_state.channel_mode);
            state_runtime_main_screen(ACTION_REFRESH);
            return;
        case CONTROL_MODE_SELECT_EQ_MODE:
	    if (runtime_state.eq_mode < vcu.eq.get_mode_count() - 1) {
		runtime_state.eq_mode++;
	    } else {
		runtime_state.eq_mode = -1;
	    }
            if (!vcu.runtime.eq.get_mode_active(runtime_state.eq_mode)) {
                state_runtime_main_screen(KI_KEY_UP);
                return;
            }
            vcu.runtime.eq.set_mode(runtime_state.eq_mode);
            state_runtime_main_screen(ACTION_REFRESH);
            return;
	case CONTROL_MODE_SELECT_MENU:
	    if (select_menu < SELECT_MENU_MAX) {
		select_menu++;
	    } else {
		select_menu = SELECT_MENU_MIN;
	    }
	    break;
	}
	break;
    case KI_KEY_DOWN:
	switch (control_mode) {
	case CONTROL_MODE_PICK_REVERB:
            if (runtime_state.revidx > 0) {
                runtime_state.revidx--;
            } else {
                runtime_state.revidx =
                    runtime_state.n_programs - 1;
            }
            vcu.runtime.reverb.set_id(runtime_state.fprgs
                                      [runtime_state.revidx].id);
            vcu.runtime.signal.reset_peak_meter();
	    break;
	case CONTROL_MODE_ALTER_MASTER_VOL:
            handle_runtime_key(KI_KEY_MVOL_DOWN);
	    break;
	case CONTROL_MODE_ALTER_GROUP_VOL:
            vcu.runtime.volume.set_group
                (runtime_state.curgrp,
                 runtime_state.group_volume[runtime_state.curgrp] -
                 conf_user.volume_step);
            runtime_state.group_volume[runtime_state.curgrp] =
                vcu.runtime.volume.get_group(runtime_state.curgrp);
	    break;
        case CONTROL_MODE_SELECT_CHANNEL_MODE:
	    if (runtime_state.channel_mode > 0) {
		runtime_state.channel_mode--;
	    } else {
		runtime_state.channel_mode = vcu.channel.get_mode_count() - 1;
	    }
            vcu.runtime.channel.set_mode(runtime_state.channel_mode);
            state_runtime_main_screen(ACTION_REFRESH);
            return;
        case CONTROL_MODE_SELECT_EQ_MODE:
	    if (runtime_state.eq_mode > -1) {
		runtime_state.eq_mode--;
	    } else {
		runtime_state.eq_mode = vcu.eq.get_mode_count() - 1;
	    }
            if (!vcu.runtime.eq.get_mode_active(runtime_state.eq_mode)) {
                state_runtime_main_screen(KI_KEY_DOWN);
                return;
            }
            vcu.runtime.eq.set_mode(runtime_state.eq_mode);
            state_runtime_main_screen(ACTION_REFRESH);
            return;
	case CONTROL_MODE_SELECT_MENU:
	    if (select_menu > SELECT_MENU_MIN) {
		select_menu--;
	    } else {
		select_menu = SELECT_MENU_MAX;
	    }
	    break;
	}
	break;
    case KI_KEY_LEFT:
	if (control_mode == CONTROL_MODE_SELECT_MENU) {
	    break;
	}
        switch (control_mode) {
        case CONTROL_MODE_ALTER_GROUP_VOL:
            handle_runtime_key(KI_KEY_GROUP_MUTE);
            break;
        default:
            handle_runtime_key(KI_KEY_MASTER_MUTE);
            break;
        }
	break;
    case KI_KEY_RIGHT:
        breaknow = false;
        switch (control_mode) {
        case CONTROL_MODE_ALTER_GROUP_VOL:
            if (runtime_state.curgrp < 0) {
                break;
            }
            n = runtime_state.curgrp;
            handle_runtime_key(KI_KEY_GROUP_NEXT);
            if (runtime_state.curgrp <= n) {
                runtime_state.curgrp = n;
            } else {
                breaknow = true;
            }
            break;
        case CONTROL_MODE_SELECT_MENU:
	    switch (select_menu) {
            case SELECT_MENU_MAIN_SCREEN:
                break;
	    case SELECT_MENU_CHANNEL_CONTROL:
		textgui_push_state(state_runtime_channel_control);
		return;
	    case SELECT_MENU_SETUP:
		vcu.runtime.stop();
		return;
	    }
            break;
        }
        if (breaknow) {
            break;
        }
        control_mode++;
        if (control_mode > CONTROL_MODE_MAX) {
	    control_mode = CONTROL_MODE_MIN;
	}
        if (!use_master &&
            control_mode == CONTROL_MODE_ALTER_MASTER_VOL)
        {
            state_runtime_main_screen(KI_KEY_RIGHT);
            return;
        }
        if ((runtime_state.curgrp < 0 || active_group_count() == 1) &&
            control_mode == CONTROL_MODE_ALTER_GROUP_VOL)
        {
            state_runtime_main_screen(KI_KEY_RIGHT);
            return;
        }
        if ((runtime_state.channel_mode < 0 ||
             conf_user.channel_mode_option != SELECT_OPTION_SEL_ON) &&
            control_mode == CONTROL_MODE_SELECT_CHANNEL_MODE)
        {
            state_runtime_main_screen(KI_KEY_RIGHT);
            return;
        }
        if ((!vcu.runtime.eq.get_mode_active(-1) ||
             conf_user.eq_mode_option != SELECT_OPTION_SEL_ON) &&
            control_mode == CONTROL_MODE_SELECT_EQ_MODE)
        {
            state_runtime_main_screen(KI_KEY_RIGHT);
            return;
        }
        if (!vcu.runtime.channel.get_type_active(VCU_CH_REVERB) &&
            control_mode == CONTROL_MODE_PICK_REVERB)
        {
            state_runtime_main_screen(KI_KEY_RIGHT);
            return;
        }
        if (control_mode == CONTROL_MODE_ALTER_GROUP_VOL) {
            for (runtime_state.curgrp = 0;
                 runtime_state.curgrp < vcu.channel.get_group_count();
                 runtime_state.curgrp++)
            {
                if (vcu.runtime.channel.get_group_active(runtime_state.curgrp)
                    &&
                    conf_user.vol_groups[runtime_state.curgrp])
                {
                    break;
                }
            }
        }
	break;
    case ACTION_REFRESH:
    case ACTION_CHILD_POP:
    case ACTION_INIT:
        runtime_state.master_mute = vcu.runtime.volume.get_master_temp_mute();
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            runtime_state.group_volume[n] = vcu.runtime.volume.get_group(n);
            runtime_state.group_mute[n] = vcu.runtime.volume.get_group_mute(n);
        }
	runtime_state.master_volume = vcu.runtime.volume.get_master();
        runtime_state.channel_mode = vcu.runtime.channel.get_mode();
        runtime_state.eq_mode = vcu.runtime.eq.get_mode();
        use_master = conf_user.vol_master;
        
        if (action != ACTION_REFRESH ||
            !vcu.runtime.channel.get_group_active(runtime_state.curgrp))
        {
            for (runtime_state.curgrp = 0;
                 runtime_state.curgrp < vcu.channel.get_group_count();
                 runtime_state.curgrp++)
            {
                if (vcu.runtime.channel.get_group_active(runtime_state.curgrp)
                    &&
                    conf_user.vol_groups[runtime_state.curgrp])
                {
                    break;
                }
            }
            
            if (runtime_state.curgrp == vcu.channel.get_group_count()) {
                runtime_state.curgrp = -1;
            }
        }
        if (runtime_state.curgrp == -1 || active_group_count() < 2) {
            use_master = true;
        }
        if (vcu.runtime.channel.get_type_active(VCU_CH_REVERB)) {
            runtime_state.fprgs =
                vcu.reverb.get_programs(&runtime_state.n_programs);
            id = vcu.runtime.reverb.get_id();
            for (runtime_state.revidx = 0;
                 memcmp(&id, &runtime_state.fprgs
                        [runtime_state.revidx].id, sizeof(id)) != 0;
                 runtime_state.revidx++);
        }
        switch (action) {
        case ACTION_INIT:
            if (vcu.runtime.channel.get_type_active(VCU_CH_REVERB)) {
                control_mode = CONTROL_MODE_PICK_REVERB;
            } else {
                control_mode = CONTROL_MODE_ALTER_MASTER_VOL;
            }
            select_menu = SELECT_MENU_MAIN_SCREEN;
            break;
        case ACTION_REFRESH:
            break;
        case ACTION_CHILD_POP:
            control_mode = CONTROL_MODE_SELECT_MENU;
            break;
        }
        if ((control_mode == CONTROL_MODE_ALTER_GROUP_VOL &&
             runtime_state.curgrp == -1) ||
            (control_mode == CONTROL_MODE_PICK_REVERB &&
             !vcu.runtime.channel.get_type_active(VCU_CH_REVERB)))
        {
            control_mode = CONTROL_MODE_ALTER_MASTER_VOL;
        }
        if (control_mode == CONTROL_MODE_ALTER_MASTER_VOL && !use_master) {
            control_mode = CONTROL_MODE_ALTER_GROUP_VOL;
        }
        if (!last_sysinfo.callback_active) {
            last_sysinfo.maxload = vcu.sysinfo.get_cpu_maxload();
            last_sysinfo.memused = vcu.sysinfo.get_ram_used();
            last_sysinfo.sample_rate =
                vcu.runtime.signal.get_input_sample_rate();
            last_sysinfo.callback_active = true;
            tv.tv_sec = 5;
            tv.tv_usec = 0;
            vcu.ctl.register_timer(NULL, cpu_load_callback, &tv);
        }
	break;
    case ACTION_EXIT:
	textgui_pop_state();
	return;
    default:
	break;
    }

    tw_clear();
    if (tw_ysize() < 4) {
	/* update first line */
        runtime_screen_update_top_line(control_mode, select_menu, false);

	/* update second line */
        if (runtime_state.peak == VCU_PEAK_NEG_INFINITE) {
            tw_printf(0, 1, "p-inf");
        } else {
            tw_printf(0, 1, "p%+.1f", runtime_state.peak);
        }
        if (runtime_state.curgrp >= 0 && active_group_count() > 1) {
            if (use_master) {
                if (vcu.runtime.volume.get_master_temp_mute()) {
                    tw_printf(7, 1, "m:mute");
                } else {
                    tw_printf(7, 1, "m%+.1f", vcu.runtime.volume.get_master());
                }
            }
            if (vcu.runtime.volume.get_group_mute(runtime_state.curgrp)) {
                tw_printf(14, 1, "g:mute");
            } else {
                tw_printf(14, 1, "g%+.1f",
                          vcu.runtime.volume.get_group(runtime_state.curgrp));
            }
        } else {
            if (vcu.runtime.volume.get_master_temp_mute()) {
                tw_printf(14, 1, "m:mute");
            } else {
                tw_printf(14, 1, "m%+.1f", vcu.runtime.volume.get_master());
            }
        }
	switch (control_mode) {
	case CONTROL_MODE_ALTER_MASTER_VOL:            
	    tw_putchar(runtime_state.curgrp >= 0 && active_group_count() > 1 ?
                       7 : 14, 1, 'M');
	    break;
	case CONTROL_MODE_ALTER_GROUP_VOL:
            tw_putchar(14, 1, 'G');
	    break;
	}
    } else {
	/* update first line */
        runtime_screen_update_top_line(control_mode, select_menu, true);

	/* update second line */
	switch (control_mode) {
	case CONTROL_MODE_PICK_REVERB:
	    tw_putstring(0, 1, " [pick reverb prog] ");
	    break;
	case CONTROL_MODE_ALTER_MASTER_VOL:
	    tw_putstring(0, 1, " [alter master vol] ");
	    break;
	case CONTROL_MODE_ALTER_GROUP_VOL:
            chgrp = vcu.channel.get_group(runtime_state.curgrp);
            sprintf(s, "[alter %s vol]", chgrp->name);
            tw_putstring((20 - strlen(s)) / 2, 1, s);
	    break;
	case CONTROL_MODE_SELECT_CHANNEL_MODE:
	    tw_putstring(0, 1, "[pick channel mode] ");
	    break;
	case CONTROL_MODE_SELECT_EQ_MODE:
            if (runtime_state.channel_mode >= 0) {
                chmode = vcu.channel.get_mode(runtime_state.channel_mode);
                count = bit_count(chmode->eqmodes, vcu.eq.get_mode_count());
            } else {
                count = vcu.eq.get_mode_count();
            }
            if (count == 1) {
                tw_putstring(0, 1, "    [toggle eq]     ");
            } else {
                tw_putstring(0, 1, "   [pick eq mode]   ");
            }
	    break;
	case CONTROL_MODE_SELECT_MENU:
	    tw_putstring(0, 1, "   [select menu]    ");
	    break;
	}

	/* update third line */
        if (runtime_state.peak == VCU_PEAK_NEG_INFINITE) {
            tw_printf(0, 2, "pk-inf");
        } else {
            tw_printf(0, 2, "pk%+.1f", runtime_state.peak);
        }
        if (use_master) {
            if (vcu.runtime.volume.get_master_temp_mute()) {
                tw_printf(9, 2, "master:mute");
            } else {
                tw_printf(9, 2, "master%+.1f", vcu.runtime.volume.get_master());
            }
        }
        
	/* update fourth line */
        n = (int)(last_sysinfo.maxload * 100.0);
        if (n == 100) {
            n = 99;
        }
        tw_printf(0, 3, "%02d/%.1f      ", n,
                  (double)last_sysinfo.sample_rate / 1000.0);
        if (runtime_state.curgrp >= 0 && active_group_count() > 1) {
            chgrp = vcu.channel.get_group(runtime_state.curgrp);
            if (vcu.runtime.volume.get_group_mute(runtime_state.curgrp)) {
                tw_printf(15 - strlen(chgrp->name),
                          use_master ? 3 : 2,
                          "%s:mute", chgrp->name);
            } else {
                tw_printf(15 - strlen(chgrp->name),
                          use_master ? 3 : 2,
                          "%s%+.1f",
                          chgrp->name,
                          vcu.runtime.volume.get_group(runtime_state.curgrp));
            }
        }
        if (!use_master && vcu.runtime.volume.get_master_temp_mute()) {
            tw_putstring(11, 3, "all muted");
        }
    }
    tw_refresh();
}

static void
state_mix_channels_sub(int action)
{
    static struct textmenu_state *mstate;
    static bool_t active[VCU_MAX_CHANNELS];
    static int count;

    uint32_t mask[VCU_MAX_CHANNELS/32];
    int n;
    
    if (action == ACTION_INIT) {
        void *entries[2 * VCU_MAX_CHANNELS + 10];
        char name[10], aname[10];
        const uint32_t *curmask;
        int pos;

        elog(LOGDEBUG, "io %d; type %d; idx %d",
             mix_channels_state.io, mix_channels_state.type,
             mix_channels_state.idx);
        pos = 0;
        strcpy(name, "ARBP");
        strcpy(aname, "wxyzrstuv");
        count = vcu.channel.get_count(mix_channels_state.io, VCU_CH_PHYSICAL);
        curmask = vcu.channel.get_phys_map(mix_channels_state.io,
                                           mix_channels_state.type,
                                           mix_channels_state.idx);
        array_bit2bool(curmask, active, count);
        for (n = 0; n < count; n++) {
            entries[pos] = alloca(40);
            if (mix_channels_state.io == IN) {
                switch (mix_channels_state.type) {
                case VCU_CH_AMBIOPOLE:
                case VCU_CH_PASSTHRU:
                    if (n > 8 && mix_channels_state.idx > 8) {
                        sprintf((char *)entries[pos++],
                                "link i%c%d-i%d[off/on]",
                                name[mix_channels_state.type],
                                mix_channels_state.idx + 1, n + 1);
                    } else {
                        sprintf((char *)entries[pos++],
                                "link i%c%d-i%d [off/on]",
                                name[mix_channels_state.type],
                                mix_channels_state.idx + 1, n + 1);
                    }
                    break;
                case VCU_CH_REVERB:
                    switch (vcu.channel.get_count(VCU_IN, VCU_CH_REVERB)) {
                    case 1:
                        sprintf((char *)entries[pos++],
                                "link iRc-i%d [off/on]", n + 1);
                        break;
                    case 2:
                        sprintf((char *)entries[pos++],
                                "link iR%c-i%d [off/on]",
                                mix_channels_state.idx == 0 ? 'l' : 'r', n + 1);
                        break;
                    case 3:
                        sprintf((char *)entries[pos++],
                                "link iR%c-i%d [off/on]",
                                mix_channels_state.idx == 0 ? 'l' :
                                (mix_channels_state.idx == 1 ? 'r' : 'c'),
                                n + 1);
                        break;
                    }
                    break;
                case VCU_CH_AMBISONIC:
                    switch (vcu.ambisonics.get_input_format()) {
                    case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:                
                    case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                        sprintf((char *)entries[pos++], "link iB%c-i%d "
                                "[off/on]",
                                aname[mix_channels_state.idx], n + 1);
                        break;
                    case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                        sprintf((char *)entries[pos++], "link iB%c-i%d "
                                "[off/on]",
                                mix_channels_state.idx == 0 ? 'l' : 'r', n + 1);
                        break;
                    }
                    break;
                }
            } else {
                if (n > 8 && mix_channels_state.idx > 8) {
                    sprintf((char *)entries[pos++], "link o%c%d-o%d[off/on]",
                            name[mix_channels_state.type],
                            mix_channels_state.idx + 1, n + 1);
                } else {
                    sprintf((char *)entries[pos++], "link o%c%d-o%d [off/on]",
                            name[mix_channels_state.type],
                            mix_channels_state.idx + 1, n + 1);
                }
            }
            entries[pos++] = &active[n];
        }
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        for (n = 0; n < count; n++) {
            if (active[n]) {
                break;
            }
        }
        if (n == count) {
	    textmenu_message("There must be at    ",
			     "least one link      ");
	    return;
        }
        memset(mask, 0, sizeof(mask));
        array_bool2bit(active, mask, count);
        if (vcu.channel.set_phys_map(mix_channels_state.io,
                                     mix_channels_state.type,
                                     mix_channels_state.idx,
                                     mask) != 0)
        {
	    textmenu_message("The total link count",
			     "is too high         ");
	    return;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_mix_channels(int action)
{
    static struct textmenu_state *mstate;
    static int typebase[VCU_CH_COUNT + 1];
    static int menupos = 0;
    int type;

    switch (action) {
    case ACTION_CHILD_POP:
        if (mstate == NULL) {
            textgui_pop_state();
            return;
        }
        textmenu_state_delete(mstate);
        state_mix_channels(ACTION_INIT);
        return;
    case ACTION_INIT: {
	void *entries[2 * VCU_MAX_CHANNELS + 10];
	int pos, inte_count, phys, phys_count, count, inte, io;
        char name[10], aname[10], chname[10], *str, *p;
        const uint32_t *mask;

	if (!have_channels()) {
	    mstate = NULL;
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
        io = mix_channels_state.io;
        phys_count = vcu.channel.get_count(io, VCU_CH_PHYSICAL);
        pos = 0;
        strcpy(name, "ARBP");
        strcpy(aname, "wxyzrstuv");
        count = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            inte_count = vcu.channel.get_count(io, type);
            typebase[type] = pos / 2;
            for (inte = 0; inte < inte_count; inte++) {
                if (io == IN) {
                    switch (type) {
                    case VCU_CH_AMBIOPOLE:
                    case VCU_CH_PASSTHRU:
                        sprintf(chname, "i%c%d", name[type], inte + 1);
                        break;
                    case VCU_CH_REVERB:
                        switch (inte_count) {
                        case 1:
                            strcpy(chname, "iRc");
                            break;
                        case 2:
                            sprintf(chname, "iR%c", inte == 0 ? 'l' : 'r');
                            break;
                        case 3:
                            sprintf(chname, "iR%c", inte == 0 ? 'l' :
                                    (inte == 1 ? 'r' : 'c'));
                            break;
                        }
                        break;
                    case VCU_CH_AMBISONIC:
                        switch (vcu.ambisonics.get_input_format()) {
                        case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
                        case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                            sprintf(chname, "iB%c", aname[inte]);
                            break;
                        case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                            sprintf(chname, "iB%c", inte == 0 ? 'l' : 'r');
                            break;
                        }
                        break;
                    }
                } else {
                    sprintf(chname, "o%c%d", name[type], inte + 1);
                }
                mask = vcu.channel.get_phys_map(io, type, inte);
                count = bit_count(mask, phys_count);
                if (count < 6) {
                    str = alloca(64);
                    phys = -1;
                    sprintf(str, "%s (", chname);
                    p = str + strlen(str);
                    while ((phys = bit_find(mask, phys + 1, phys_count - 1))
                           != -1)
                    {
                        sprintf(p, "%c%d,", io == IN ? 'i' : 'o', phys + 1);
                        p += strlen(p);
                    }
                    p[-1] = ')';
                    strcat(p, "...");
                    entries[pos++] = str;
                } else {
                    entries[pos] = alloca(30);
                    sprintf((char *)entries[pos++], "%s (%d links)...",
                            chname, count);
                }
                entries[pos++] = state_mix_channels_sub;
            }
        }
        typebase[type] = pos;
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    }

    for (type = 0; type < VCU_CH_COUNT; type++) {
        if (mstate->pos < typebase[type + 1]) {
            break;
        }
    }
    mix_channels_state.type = type;
    mix_channels_state.idx = mstate->pos - typebase[type];
    menupos = mstate->pos;
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }
}

static void
state_mix_inputs(int action)
{
    mix_channels_state.io = VCU_IN;
    state_mix_channels(action);
}

static void
state_mix_outputs(int action)
{
    mix_channels_state.io = VCU_OUT;
    state_mix_channels(action);
}

static void
state_reverb_setup(int action)
{
    static struct textmenu_state *mstate;
    static int source_layout;
    
    if (action == ACTION_INIT) {
        source_layout = vcu.reverb.get_source_layout() - 1;
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
		"Layout [center/stereo/lrc]", &source_layout,
		NULL
	    });
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.reverb.set_source_layout(source_layout + 1);
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_ambisonic_setup(int action)
{
    static struct textmenu_state *mstate;
    static int input_format;
    
    if (action == ACTION_INIT) {
        input_format = vcu.ambisonics.get_input_format() - 1;
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
		"Format [1st order/2nd order]", &input_format,
		NULL
	    });
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.ambisonics.set_input_format(input_format + 1);
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_channel_group_select_output_channels(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_chgrp chgrp;
    static bool_t channels[VCU_CH_COUNT][VCU_MAX_CHANNELS];
    static int channel_count[VCU_CH_COUNT];
    int type, n;
    
    if (action == ACTION_INIT) {
        char name[VCU_CH_COUNT + 1];
	void **entries;
	int pos;

        mstate = NULL;
        memcpy(&chgrp, vcu.channel.get_group(current_channel_group),
               sizeof(chgrp));
        if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
        }
        for (type = 0; type < VCU_CH_COUNT; type++) {
            channel_count[type] = vcu.channel.get_count(VCU_OUT, type);
        }

        strcpy(name, "ARBP");
        entries = alloca((2 * total_channel_count(VCU_OUT) + 1) *
                         sizeof(void *));
	pos = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            array_bit2bool(chgrp.channels[OUT][type], channels[type], 
                           VCU_MAX_CHANNELS);
            for (n = 0; n < channel_count[type]; n++) {
                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++], "control o%c%d [off/on]",
                        name[type], n + 1);
                entries[pos++] = &channels[type][n];
            }
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }
    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        for (type = 0; type < VCU_CH_COUNT; type++) {
            array_bool2bit(channels[type], chgrp.channels[OUT][type],
                           VCU_MAX_CHANNELS);
        }
        vcu.channel.modify_group(current_channel_group, &chgrp);
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_channel_group_select_input_channels(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_chgrp chgrp;
    static bool_t channels[VCU_CH_COUNT][VCU_MAX_CHANNELS];
    static int channel_count[VCU_CH_COUNT];
    int n, type;

    if (action == ACTION_INIT) {
        char channel_name[10];
	void **entries;
	int pos;

        mstate = NULL;
        memcpy(&chgrp, vcu.channel.get_group(current_channel_group),
               sizeof(chgrp));
        if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
        }
        for (type = 0; type < VCU_CH_COUNT; type++) {
            channel_count[type] = vcu.channel.get_count(VCU_IN, type);
            array_bit2bool(chgrp.channels[IN][type], channels[type], 
                           VCU_MAX_CHANNELS);
        }

	pos = 0;
        entries = alloca((1 + 2 * total_channel_count(VCU_IN)) *
                         sizeof(void *));

        /* ambiopole */
	for (n = 0; n < channel_count[VCU_CH_AMBIOPOLE]; n++) {
            entries[pos] = alloca(40);
	    sprintf((char *)entries[pos++], "control iA%d [off/on]", n + 1);
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_AMBIOPOLE][n];
	}

        /* reverb */
        switch (channel_count[VCU_CH_REVERB]) {
        case 1:
	    entries[pos++] = "control iRc [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][0];
            break;
        case 2:
	    entries[pos++] = "control iRl [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][0];
	    entries[pos++] = "control iRr [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][1];
            break;
        case 3:
	    entries[pos++] = "control iRl [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][0];
	    entries[pos++] = "control iRr [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][1];
	    entries[pos++] = "control iRc [off/on]";
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_REVERB][2];
            break;
        }

        /* ambisonic */
        if (channel_count[VCU_CH_AMBISONIC] > 0) {
            switch (vcu.ambisonics.get_input_format()) {
            case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:                
            case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                strcpy(channel_name, "wxyzrstuv");
                for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                    entries[pos] = alloca(40);
                    sprintf((char *)entries[pos++], "control iB%c [off/on]",
                            channel_name[n]);
                    entries[pos++] = &chgrp.channels[VCU_IN]
                        [VCU_CH_AMBISONIC][n];
                }
                break;
            case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                entries[pos++] = "control iBl [off/on]";
                entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_AMBISONIC][0];
                entries[pos++] = "control iBr [off/on]";
                entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_AMBISONIC][1];
                break;
            }
        }
        
        /* passthru */
	for (n = 0; n < channel_count[VCU_CH_PASSTHRU]; n++) {
            entries[pos] = alloca(40);
	    sprintf((char *)entries[pos++], "control iP%d [off/on]", n + 1);
	    entries[pos++] = &chgrp.channels[VCU_IN][VCU_CH_PASSTHRU][n];
	}
        
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }
    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        for (type = 0; type < VCU_CH_COUNT; type++) {
            array_bool2bit(channels[type], chgrp.channels[IN][type],
                           VCU_MAX_CHANNELS);
        }
        vcu.channel.modify_group(current_channel_group, &chgrp);
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_channel_edit_group(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_chgrp chgrp;
    static bool_t created_new;
    static double volume;
    static double mute;
    static int menupos = 0;

    switch (action) {
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        /* fall through */
    case ACTION_INIT: {
	int pos, type, count;
	void *entries[100];
        double range[2];
	
        mstate = NULL;
	if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
        if (action != ACTION_CHILD_POP) {
            created_new = false;
        }
        if (current_channel_group < 0) {
            memset(&chgrp, 0, sizeof(chgrp));
            strcpy(chgrp.name, "unnamed");
            if (vcu.channel.new_group(&chgrp) != 0) {
                textmenu_message("Cannot create more  ",
                                 "channel groups      ");
                return;
            }
            current_channel_group = vcu.channel.get_group_count() - 1;
            created_new = true;
        }
        memcpy(&chgrp, vcu.channel.get_group(current_channel_group),
               sizeof(chgrp));
        volume = vcu.runtime.volume.get_group(current_channel_group);
        mute = vcu.runtime.volume.get_group_mute(current_channel_group);
	vcu.runtime.volume.get_range(range);
        pos = 0;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Name [%%s,%d]",
                VCU_CHGRP_NAME_SIZE - 1);
        entries[pos++] = chgrp.name;
        entries[pos] = alloca(40);
	sprintf((char *)entries[pos++], "Volume [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
        entries[pos++] = &volume;
        entries[pos++] = "Mute [off/on]";
        entries[pos++] = &mute;
        count = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            count += bit_count(chgrp.channels[VCU_IN][type],
                               vcu.channel.get_count(VCU_IN, type));
        }
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Inputs (%d)...", count);
        entries[pos++] = state_channel_group_select_input_channels;
        count = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            count += bit_count(chgrp.channels[VCU_OUT][type],
                               vcu.channel.get_count(VCU_OUT, type));
        }
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Outputs (%d)...", count);
        entries[pos++] = state_channel_group_select_output_channels;
        if (created_new) {
            entries[pos++] = "Create this group?";
            entries[pos++] = NULL;
        } else {
            entries[pos++] = "Delete this group?ask";
            entries[pos++] = NULL;
        }
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.channel.modify_group(current_channel_group, &chgrp);
        break;
    case 1:
        vcu.runtime.volume.set_group(current_channel_group, volume);
        break;
    case 2:
        vcu.runtime.volume.set_group_mute(current_channel_group, volume);
        break;
    case 5:
        menupos = 0;
        if (created_new) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
        vcu.channel.delete_group(current_channel_group);
        state_channel_edit_group(ACTION_EXIT);
        break;
    case ACTION_EXIT:
        menupos = 0;
        if (created_new) {
            vcu.channel.delete_group(current_channel_group);
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_channel_reorder_groups(int action)
{
    static struct textmenu_state *mstate;
    static int groups[VCU_MAX_CHANNEL_GROUPS];
    static int menupos = 0;
    static bool_t move_up;

    const struct vcu_chgrp *chgrp;
    bool_t do_refresh = false;
    void **entries;
    int pos, n;
    
    switch (action) {
    case ACTION_REFRESH:
	textmenu_state_delete(mstate);
        do_refresh = true;
        /* fall through */
    case ACTION_INIT: 
        mstate = NULL;
        if (vcu.channel.get_group_count() < 2) {
            textmenu_message("There are less than ",
                             "two groups          ");
            return;
        }        
        entries = alloca((vcu.channel.get_group_count() + 20) * sizeof(void *));
	pos = 0;
        entries[pos++] = "Move [down/up]";
        entries[pos++] = &move_up;
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            if (do_refresh) {
                chgrp = vcu.channel.get_group(groups[n]);
            } else {
                move_up = true;
                groups[n] = n;
                chgrp = vcu.channel.get_group(n);
            }
            entries[pos] = alloca(VCU_CHGRP_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Move %s?", chgrp->name);
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    if (mstate == NULL) {
        textgui_pop_state();
        menupos = 0;
        return;
    }
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 1) {
            return;
        }        
        action--;
        if (move_up) {
            if (action == 0) {
                return;
            }
            n = groups[action - 1];
            groups[action - 1] = groups[action];
            groups[action] = n;
        } else {
            if (action == vcu.channel.get_group_count() - 1) {
                return;
            }
            n = groups[action + 1];
            groups[action + 1] = groups[action];
            groups[action] = n;
        }
        menupos = mstate->pos;
        state_channel_reorder_groups(ACTION_REFRESH);
        return;
    case ACTION_EXIT:
        vcu.channel.reorder_groups(groups);
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_channel_groups(int action)
{
    static struct textmenu_state *mstate;
    static int menupos = 0;

    switch (action) {
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        state_channel_groups(ACTION_INIT);
        return;
    case ACTION_INIT: {
        const struct vcu_chgrp *chgrp;
	void **entries;
	int pos, n;
	
        entries = alloca((vcu.channel.get_group_count() + 20) * sizeof(void *));
	pos = 0;
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            chgrp = vcu.channel.get_group(n);
            entries[pos] = alloca(VCU_CHGRP_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Edit \"%s\"...", chgrp->name);
            entries[pos++] = state_channel_edit_group;
        }
        entries[pos++] = "New channel grp...";
        entries[pos++] = state_channel_edit_group;
        entries[pos++] = "Reorder groups...";
        entries[pos++] = state_channel_reorder_groups;
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    }

    if (mstate->pos >= vcu.channel.get_group_count()) {
        current_channel_group = -1;
    } else {
        current_channel_group = mstate->pos;
    }
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_channel_edit_mode(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_chmode mode;
    static bool_t groups[VCU_MAX_CHANNEL_GROUPS];
    static int delete_mode_pos;
    static bool_t created_new;
    
    if (action == ACTION_INIT) {
	void *entries[20 + vcu.channel.get_group_count()];
        char fmt[20], name[VCU_CHGRP_NAME_SIZE + 1];
        const struct vcu_chgrp *chgrp;
	int pos, n;

	delete_mode_pos = -1;
        mstate = NULL;
	if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
	if (vcu.channel.get_group_count() <= 0) {
	    textmenu_message("There are no confi- ",
			     "gured groups        ");
	    return;
	}
        created_new = false;
        if (current_channel_mode < 0) {
            memset(&mode, 0, sizeof(mode));
            mode.active_eq_mode = -1;
            strcpy(mode.name, "unnamed");
            if (vcu.channel.new_mode(&mode) != 0) {
                textmenu_message("Cannot create more  ",
                                 "channel modes       ");
                return;
            }
            current_channel_mode = vcu.channel.get_mode_count() - 1;
            created_new = true;
        }
        memcpy(&mode, vcu.channel.get_mode(current_channel_mode), sizeof(mode));
        array_bit2bool(mode.groups, groups, VCU_MAX_CHANNEL_GROUPS);
        pos = 0;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Name [%%s,%d]",
                VCU_CHMODE_NAME_SIZE - 1);
        entries[pos++] = mode.name;
        sprintf(fmt, "%%-%ds [off/on]", VCU_CHGRP_NAME_SIZE + 1);
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            chgrp = vcu.channel.get_group(n);
            entries[pos] = alloca(strlen(chgrp->name) + 20);
            sprintf(name, "\"%s\"", chgrp->name);
            sprintf((char *)entries[pos], fmt, name);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &groups[n];
        }
        delete_mode_pos = pos / 2;
        if (created_new) {
            entries[pos++] = "Create this mode?";
            entries[pos++] = NULL;
        } else {
            entries[pos++] = "Delete this mode?ask";
            entries[pos++] = NULL;
        }
	mstate = textmenu_state_new(NULL, entries);
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);
    
    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        if (action == delete_mode_pos) {
            if (created_new) {
                textmenu_state_delete(mstate);
                textgui_pop_state();
                return;
            }
            vcu.channel.delete_mode(current_channel_mode);
            state_channel_edit_mode(ACTION_EXIT);
            return;
        }
        array_bool2bit(groups, mode.groups, VCU_MAX_CHANNEL_GROUPS);
        vcu.channel.modify_mode(current_channel_mode, &mode);
        break;
    case ACTION_EXIT:
        if (created_new) {
            vcu.channel.delete_mode(current_channel_mode);
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

/* FIXME: very similar to reorder_groups function */
static void
state_channel_reorder_modes(int action)
{
    static struct textmenu_state *mstate;
    static int modes[VCU_MAX_CHANNEL_MODES];
    static int menupos = 0;
    static bool_t move_up;

    const struct vcu_chmode *chmode;
    bool_t do_refresh = false;
    void **entries;
    int pos, n;
    
    switch (action) {
    case ACTION_REFRESH:
	textmenu_state_delete(mstate);
        do_refresh = true;
        /* fall through */
    case ACTION_INIT: 
        mstate = NULL;
        if (vcu.channel.get_mode_count() < 2) {
            textmenu_message("There are less than ",
                             "two modes           ");
            return;
        }        
        entries = alloca((vcu.channel.get_mode_count() + 20) * sizeof(void *));
	pos = 0;
        entries[pos++] = "Move [down/up]";
        entries[pos++] = &move_up;
        for (n = 0; n < vcu.channel.get_mode_count(); n++) {
            if (do_refresh) {
                chmode = vcu.channel.get_mode(modes[n]);
            } else {
                move_up = true;
                modes[n] = n;
                chmode = vcu.channel.get_mode(n);
            }
            entries[pos] = alloca(VCU_CHMODE_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Move %s?", chmode->name);
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    if (mstate == NULL) {
        textgui_pop_state();
        menupos = 0;
        return;
    }
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 1) {
            return;
        }        
        action--;
        if (move_up) {
            if (action == 0) {
                return;
            }
            n = modes[action - 1];
            modes[action - 1] = modes[action];
            modes[action] = n;
        } else {
            if (action == vcu.channel.get_mode_count() - 1) {
                return;
            }
            n = modes[action + 1];
            modes[action + 1] = modes[action];
            modes[action] = n;
        }
        menupos = mstate->pos;
        state_channel_reorder_modes(ACTION_REFRESH);
        return;
    case ACTION_EXIT:
        vcu.channel.reorder_modes(modes);
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_channel_modes(int action)
{
    static int mode_base, mode, menupos = 0;
    static struct textmenu_state *mstate;

    switch (action) {
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        state_channel_modes(ACTION_INIT);
        return;
    case ACTION_INIT: {
        const struct vcu_chmode *chmode;
	void **entries;
	int pos, n;
        char *p;
	
        mode = vcu.runtime.channel.get_mode() + 1;
        entries = alloca((vcu.channel.get_mode_count() + 20) * sizeof(void *));
	pos = 0;
        entries[pos] = alloca(40 + vcu.channel.get_mode_count() *
                              VCU_CHMODE_NAME_SIZE);
        if (vcu.channel.get_mode_count() > 1) {
            p = (char *)entries[pos++];
            
            strcpy(p, "Mode [off/");
            p = &p[10];
            for (n = 0; n < vcu.channel.get_mode_count(); n++) {
                chmode = vcu.channel.get_mode(n);
                sprintf(p, "%s/", chmode->name);
                p += strlen(p);
            }
            p[-1] = ']';
            entries[pos++] = &mode;
        } else {
            vcu.runtime.channel.set_mode(-1);
            entries[pos++] = "Modes off";
        }
        mode_base = 1;
        
        for (n = 0; n < vcu.channel.get_mode_count(); n++) {
            chmode = vcu.channel.get_mode(n);
            entries[pos] = alloca(VCU_CHMODE_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Edit \"%s\"...", chmode->name);
            entries[pos++] = state_channel_edit_mode;
        }
        entries[pos++] = "New mode...";
        entries[pos++] = state_channel_edit_mode;
        entries[pos++] = "Reorder modes...";
        entries[pos++] = state_channel_reorder_modes;
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }
    }

    if (mstate->pos >= mode_base) {
        if (mstate->pos == mode_base + vcu.channel.get_mode_count()) {
            current_channel_mode = -1;
        } else {
            current_channel_mode = mstate->pos - mode_base;
        }
    }
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.runtime.channel.set_mode(mode - 1);
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_speaker_layout_per_speaker(int action)
{
    static struct textmenu_state *mstate;
    static bool_t active[VCU_MAX_CHANNELS], init_active[VCU_MAX_CHANNELS];
    static double azimuth[VCU_MAX_CHANNELS];
    static double elevation[VCU_MAX_CHANNELS];
    static double distance[VCU_MAX_CHANNELS];
    static int menupos = 0;

    switch (action) {
    case ACTION_INIT: {
	void *entries[10 + 8 * VCU_MAX_CHANNELS];
	int pos, n, output_count;
	double distmax;

	pos = 0;
	distmax = vcu.channel.distance_max() * 100.0;
        output_count = vcu.channel.get_count(OUT, VCU_CH_PHYSICAL);
	for (n = 0; n < output_count; n++) {
            vcu.channel.position_get(n, &active[n], &azimuth[n], &elevation[n],
                                     &distance[n]);
            init_active[n] = active[n];
            distance[n] *= 100.0;
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++], "o%d position [off/on]", n + 1);
            entries[pos++] = &active[n];
            if (active[n]) {
                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++],
                        "o%d a[%%.1f,0.1,10,-180,180]deg", n + 1);
                entries[pos++] = &azimuth[n];

                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++],
                        "o%d e[%%.1f,0.1,10,-90,90]deg", n + 1);
                entries[pos++] = &elevation[n];

                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++],
                        "o%d d[%%.1f,0.1,100,0,%f]cm", n + 1, distmax);
                entries[pos++] = &distance[n];
            } else {
                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++], "o%d a ---deg", n + 1);

                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++], "o%d e ---deg", n + 1);
                
                entries[pos] = alloca(40);
                sprintf((char *)entries[pos++], "o%d d ---cm", n + 1);
            }
	}
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
	break;
    }
    default:
	break;
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    /* post textmenu_action action handling */
    switch (action) {
    default: {
        int n, output_count;
        bool_t changed;
        
        if (action < 0) {
            return;
        }
        changed = false;
        output_count = vcu.channel.get_count(OUT, VCU_CH_PHYSICAL);
        for (n = 0; n < output_count; n++) {
            vcu.channel.position_set(n, active[n], azimuth[n], elevation[n],
                                     distance[n] / 100.0);
            if (active[n] != init_active[n]) {
                changed = true;
            }
        }
        if (changed) {
            textmenu_state_delete(mstate);
            state_speaker_layout_per_speaker(ACTION_INIT);
        }
	return;
    }
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_speaker_layout(int action)
{
    static struct textmenu_state *mstate;
    static bool_t distance_compensate;
    int n;
    
    if (action == ACTION_INIT) {
        distance_compensate = vcu.channel.get_distance_compensate_active();
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
		"Dist. comp. [off/on]", &distance_compensate,
                "Clear layout?ask",
                "Specify layout...", state_speaker_layout_per_speaker,
		NULL
	    });
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.channel.set_distance_compensate_active(distance_compensate);        
        break;
    case 1:
        for (n = 0; n < vcu.channel.get_count(VCU_OUT, VCU_CH_PHYSICAL); n++) {
            vcu.channel.position_set(n, false, 0, 0, 0);
        }
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_stage_layout(int action)
{
    static double azimuth[VCU_MAX_STAGES];
    static double elevation[VCU_MAX_STAGES];
    static double width[VCU_MAX_STAGES];
    static double height[VCU_MAX_STAGES];
    static struct textmenu_state *mstate;
    static int add_stage_pos, del_stage_pos;
    static int menupos = 0;
    
    if (action == ACTION_INIT) {
        int pos, n, offset;
        void **entries;
        
        entries = alloca((8 * vcu.channel.get_stage_count() + 10) *
                         sizeof(void *));
        pos = 0;
        for (n = 0; n < vcu.channel.get_stage_count(); n++) {
            vcu.channel.get_stage(n, &azimuth[n], &elevation[n],
                                  &width[n], &height[n]);
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++],
                    "S%d azim [%%.0f,1,10,-180,180]deg", n + 1);
            entries[pos++] = &azimuth[n];
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++],
                    "S%d elev [%%.0f,1,10,-90,90]deg", n + 1);
            entries[pos++] = &elevation[n];
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++],
                    "S%d width [%%.0f,1,10,0,360]deg", n + 1);
            entries[pos++] = &width[n];
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++],
                    "S%d height [%%.0f,1,10,0,180]deg", n + 1);
            entries[pos++] = &height[n];
        }
        add_stage_pos = -1;
        del_stage_pos = -1;
        offset = 0;
        if (vcu.channel.get_stage_count() < vcu.channel.get_stage_max_count()) {
            add_stage_pos = pos / 2;
            offset = 1;
            entries[pos++] = "Add stage?";
        }
        if (vcu.channel.get_stage_count() > 0) {
            del_stage_pos = pos / 2 + offset;
            entries[pos++] = "Delete last stage?ask";
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        if (action == add_stage_pos) {
            vcu.channel.new_stage(0, 0, 0, 0);
            menupos = (vcu.channel.get_stage_count() - 1) * 4;
            textmenu_state_delete(mstate);
            state_stage_layout(ACTION_INIT);
        } else if (action == del_stage_pos) {
            vcu.channel.delete_stage(vcu.channel.get_stage_count() - 1);
            menupos = vcu.channel.get_stage_count() * 4;
            if (menupos > 0) {
                menupos++;
            }
            textmenu_state_delete(mstate);
            state_stage_layout(ACTION_INIT);
        } else {
            action /= 4;
            vcu.channel.modify_stage(action,
                                     azimuth[action],
                                     elevation[action],
                                     width[action],
                                     height[action]);
        }
        return;
    case ACTION_EXIT:
        menupos = 0;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_inout_volume_edit(int action)
{
    static struct textmenu_state *mstate;
    static int action2type[VCU_CH_WP_COUNT * VCU_MAX_CHANNELS];
    static int action2idx[VCU_CH_WP_COUNT * VCU_MAX_CHANNELS];
    static double volume[VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
    static bool_t mute[VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
    static int channel_count[VCU_CH_WP_COUNT];
    static int max_action, offset;
    static bool_t isruntime;

    double vol;
    int n;

    switch (action) {
    case ACTION_INIT: {
        char name[VCU_CH_COUNT + 1 + 10], c;
	int pos, type, io;
	void **entries;
        double range[2];

        io = inout_generic_state.io;
        isruntime = vcu.runtime.isruntime();
	if (!have_channels()) {
	    mstate = NULL;
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
        
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            channel_count[type] = vcu.channel.get_count(io, type);
            for (n = 0; n < channel_count[type]; n++) {
                volume[type][n] = vcu.runtime.volume.get_trim(io, type, n);
                mute[type][n] = vcu.runtime.volume.get_mute(io, type, n);
            }
        }

        vcu.runtime.volume.get_range(range);
        pos = 0;
        entries = alloca((20 + VCU_CH_WP_COUNT * VCU_MAX_CHANNELS) *
                         sizeof(void *));
        entries[pos++] = "mute all?ask";
        entries[pos++] = "unmute all?ask";
        offset = 2;
        if (io == IN && inout_generic_state.type != VCU_CH_PHYSICAL) {

            /* ambiopole */
            for (n = 0; n < channel_count[VCU_CH_AMBIOPOLE]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_AMBIOPOLE, n))
                {
                    continue;
                }
                action2type[(pos-offset)>>2] = VCU_CH_AMBIOPOLE;
                action2idx[(pos-offset)>>2] = n;
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iA%d [%%+.1f,0.1,10,%f,%f]dB",
                        n + 1, range[0], range[1]);
                entries[pos++] = &volume[VCU_CH_AMBIOPOLE][n];                

                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iA%d mute [off/on]", n + 1);
                entries[pos++] = &mute[VCU_CH_AMBIOPOLE][n];
            }

            /* reverb */
            for (n = 0; n < channel_count[VCU_CH_REVERB]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_REVERB, n))
                {
                    continue;
                }
                c = 0;
                switch (channel_count[VCU_CH_REVERB]) {
                case 1:
                    c = 'c';
                    break;
                case 2:
                    c = n == 0 ? 'l' : 'r';
                    break;
                case 3:
                    c = n == 0 ? 'l' : (n == 1 ? 'r' : 'c');
                    break;
                }
                
                action2type[(pos-offset)>>2] = VCU_CH_REVERB;
                action2idx[(pos-offset)>>2] = n;

                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++],
                        "iR%c [%%+.1f,0.1,10,%f,%f]dB", c, range[0], range[1]);
                entries[pos++] = &volume[VCU_CH_REVERB][n];
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++],
                        "iR%c mute [off/on]", c);
                entries[pos++] = &mute[VCU_CH_REVERB][n];
            }
            
            /* ambisonic */
            if (channel_count[VCU_CH_AMBISONIC] > 0) {
                switch (vcu.ambisonics.get_input_format()) {
                case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:                
                case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                    strcpy(name, "wxyzrstuv");
                    for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                        if (isruntime &&
                            !vcu.runtime.channel.get_active(io,
                                                            VCU_CH_AMBISONIC,
                                                            n))
                        {
                            continue;
                        }
                        action2type[(pos-offset)>>2] = VCU_CH_AMBISONIC;
                        action2idx[(pos-offset)>>2] = n;
                        
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++],
                                "iB%c [%%+.1f,0.1,10,%f,%f]dB",
                                name[n], range[0], range[1]);
                        entries[pos++] = &volume[VCU_CH_AMBISONIC][n];
                        
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++], "iB%c mute [off/on]",
                                name[n]);
                        entries[pos++] = &mute[VCU_CH_AMBISONIC][n];
                    }
                    break;
                case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                    for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                        if (isruntime &&
                            !vcu.runtime.channel.get_active(io,
                                                            VCU_CH_AMBISONIC,
                                                            n))
                        {
                            continue;
                        }
                        action2type[(pos-offset)>>2] = VCU_CH_AMBISONIC;
                        action2idx[(pos-offset)>>2] = n;
                        
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++],
                                "iB%c [%%+.1f,0.1,10,%f,%f]dB",
                                n == 0 ? 'l' : 'r', range[0], range[1]);
                        entries[pos++] = &volume[VCU_CH_AMBISONIC][n];
                        
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++],
                                "iB%c mute [off/on]", n == 0 ? 'l' : 'r');
                        entries[pos++] = &mute[VCU_CH_AMBISONIC][n];
                    }
                    break;
                }
            }
        
            /* passthru */
            for (n = 0; n < channel_count[VCU_CH_PASSTHRU]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_PASSTHRU, n))
                {
                    continue;
                }
                action2type[(pos-offset)>>2] = VCU_CH_PASSTHRU;
                action2idx[(pos-offset)>>2] = n;
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iP%d [%%+.1f,0.1,10,%f,%f]dB",
                        n + 1, range[0], range[1]);
                entries[pos++] = &volume[VCU_CH_PASSTHRU][n];
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iP%d mute [off/on]", n + 1);
                entries[pos++] = &mute[VCU_CH_PASSTHRU][n];
            }
            
        }

        if (io == OUT && inout_generic_state.type != VCU_CH_PHYSICAL) {
            
            strcpy(name, "ARBP");
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (n = 0; n < channel_count[type]; n++) {
                    if (isruntime &&
                        !vcu.runtime.channel.get_active(io, type, n))
                    {
                        continue;
                    }
                    action2type[(pos-offset)>>2] = type;
                    action2idx[(pos-offset)>>2] = n;
                    
                    entries[pos] = alloca(80);
                    sprintf((char *)entries[pos++],
                            "o%c%d [%%+.1f,0.1,10,%f,%f]dB",
                            name[type], n + 1, range[0], range[1]);
                    entries[pos++] = &volume[type][n];
                    
                    entries[pos] = alloca(80);
                    sprintf((char *)entries[pos++], "o%c%d mute [off/on]",
                            name[type], n + 1);
                    entries[pos++] = &mute[type][n];
                }
            }
        }

        if (inout_generic_state.type == VCU_CH_PHYSICAL) {
            type = VCU_CH_PHYSICAL;
            for (n = 0; n < channel_count[type]; n++) {
                if (isruntime && !vcu.runtime.channel.get_active(io, type, n)) {
                    continue;
                }
                action2type[(pos-offset)>>2] = type;
                action2idx[(pos-offset)>>2] = n;
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++],
                        "%c%d [%%+.1f,0.1,10,%f,%f]dB",
                        io == IN ? 'i' : 'o', n + 1, range[0], range[1]);
                entries[pos++] = &volume[type][n];
                
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "%c%d mute [off/on]",
                        io == IN ? 'i' : 'o', n + 1);
                entries[pos++] = &mute[type][n];
            }
        }
        max_action = (pos - offset) / 4;
        
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        elog(LOGDEBUG, "mute all");
        vcu.runtime.buffer_commands(true);
        for (n = 0; n < max_action; n++) {
            elog(LOGDEBUG, "mute [%d][%d][%d] (%d)", inout_generic_state.io,
                 action2type[n], action2idx[n], n);
            mute[action2type[n]][action2idx[n]] = true;
            vcu.runtime.volume.set_mute(inout_generic_state.io, action2type[n],
                                        action2idx[n], true);
        }
        vcu.runtime.buffer_commands(false);
        state_inout_volume_edit(ACTION_REFRESH);
        return;
    case 1:
        elog(LOGDEBUG, "unmute all");
        vcu.runtime.buffer_commands(true);
        for (n = 0; n < max_action; n++) {
            mute[action2type[n]][action2idx[n]] = false;
            vcu.runtime.volume.set_mute(inout_generic_state.io, action2type[n],
                                        action2idx[n], false);
        }
        vcu.runtime.buffer_commands(false);
        state_inout_volume_edit(ACTION_REFRESH);
        return;
    default:
        if (action < 0) {
            return;
        }
        action -= 2;
        action >>= 1;
        elog(LOGDEBUG, "set state [%d][%d][%d]",
             inout_generic_state.io,
             action2type[action],
             action2idx[action]);
        /* get rid of rounding errors from volume */
        vol = volume[action2type[action]][action2idx[action]];
        vol = (double)((int)(vol * 10)) / 10.0;
        vcu.runtime.volume.set_trim(inout_generic_state.io,
                                    action2type[action],
                                    action2idx[action],
                                    vol);
        vcu.runtime.volume.set_mute
            (inout_generic_state.io,
             action2type[action],
             action2idx[action],
             mute[action2type[action]][action2idx[action]]);
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_inout_delay_edit(int action)
{
    static struct textmenu_state *mstate;
    static int action2type[VCU_CH_WP_COUNT * VCU_MAX_CHANNELS];
    static int action2idx[VCU_CH_WP_COUNT * VCU_MAX_CHANNELS];
    static double delay[VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
    static int channel_count[VCU_CH_WP_COUNT];
    static bool_t isruntime;

    double dly;

    switch (action) {
    case ACTION_INIT: {
        char name[VCU_CH_COUNT + 1 + 10], str[80];
	int pos, n, type, io;
	void **entries;
        
        sprintf(str, "[%%.2f,0.01,10,0,%f]ms",
                vcu.runtime.delay.get_max() * 1000.0);
        io = inout_generic_state.io;
        isruntime = vcu.runtime.isruntime();
	if (!have_channels()) {
	    mstate = NULL;
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
        
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            channel_count[type] = vcu.channel.get_count(io, type);
            for (n = 0; n < channel_count[type]; n++) {
                delay[type][n] =
                    1000.0 * vcu.runtime.delay.get_trim(io, type, n);
            }
        }

        pos = 0;
        entries = NULL;
        if (io == IN && inout_generic_state.type != VCU_CH_PHYSICAL) {
            entries = alloca((1 + 2 * total_channel_count(VCU_IN)) *
                             sizeof(void *));

            /* ambiopole */
            for (n = 0; n < channel_count[VCU_CH_AMBIOPOLE]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_AMBIOPOLE, n))
                {
                    continue;
                }
                action2type[pos>>1] = VCU_CH_AMBIOPOLE;
                action2idx[pos>>1] = n;
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iA%d %s", n + 1, str);
                entries[pos++] = &delay[VCU_CH_AMBIOPOLE][n];                
            }

            /* reverb */
            for (n = 0; n < channel_count[VCU_CH_REVERB]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_REVERB, n))
                {
                    continue;
                }
                action2type[pos>>1] = VCU_CH_REVERB;
                action2idx[pos>>1] = n;
                entries[pos] = alloca(80);
                switch (channel_count[VCU_CH_REVERB]) {
                case 1:
                    sprintf((char *)entries[pos++], "iRc %s", str);
                    break;
                case 2:
                    sprintf((char *)entries[pos++],
                            "iR%c %s", n == 0 ? 'l' : 'r', str);
                    break;
                case 3:
                    sprintf((char *)entries[pos++],
                            "iR%c %s", n == 0 ? 'l' : (n == 1 ? 'r' : 'c'),
                            str);
                    break;
                }
                entries[pos++] = &delay[VCU_CH_REVERB][n];
            }
            
            /* ambisonic */
            if (channel_count[VCU_CH_AMBISONIC] > 0) {
                switch (vcu.ambisonics.get_input_format()) {
                case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:                
                case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                    strcpy(name, "wxyzrstuv");
                    for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                        if (isruntime &&
                            !vcu.runtime.channel.get_active(io,
                                                            VCU_CH_AMBISONIC,
                                                            n))
                        {
                            continue;
                        }
                        action2type[pos>>1] = VCU_CH_AMBISONIC;
                        action2idx[pos>>1] = n;
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++], "iB%c %s",
                                name[n], str);
                        entries[pos++] = &delay[VCU_CH_AMBISONIC][n];
                    }
                    break;
                case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                    for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                        if (isruntime &&
                            !vcu.runtime.channel.get_active(io,
                                                            VCU_CH_AMBISONIC,
                                                            n))
                        {
                            continue;
                        }
                        action2type[pos>>1] = VCU_CH_AMBISONIC;
                        action2idx[pos>>1] = n;
                        entries[pos] = alloca(80);
                        sprintf((char *)entries[pos++],
                                "iB%c %s", n == 0 ? 'l' : 'r', str);
                        entries[pos++] = &delay[VCU_CH_AMBISONIC][n];
                    }
                    break;
                }
            }
        
            /* passthru */
            for (n = 0; n < channel_count[VCU_CH_PASSTHRU]; n++) {
                if (isruntime &&
                    !vcu.runtime.channel.get_active(io, VCU_CH_PASSTHRU, n))
                {
                    continue;
                }
                action2type[pos>>1] = VCU_CH_PASSTHRU;
                action2idx[pos>>1] = n;
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "iP%d %s", n + 1, str);
                entries[pos++] = &delay[VCU_CH_PASSTHRU][n];
            }
            
        }

        if (io == OUT && inout_generic_state.type != VCU_CH_PHYSICAL) {
            
            strcpy(name, "ARBP");
            entries = alloca((2 * total_channel_count(io) + 1) *
                             sizeof(void *));
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (n = 0; n < channel_count[type]; n++) {
                    if (isruntime &&
                        !vcu.runtime.channel.get_active(io, type, n))
                    {
                        continue;
                    }
                    action2type[pos>>1] = type;
                    action2idx[pos>>1] = n;
                    entries[pos] = alloca(80);
                    sprintf((char *)entries[pos++], "o%c%d %s",
                            name[type], n + 1, str);
                    entries[pos++] = &delay[type][n];
                }
            }
        }

        if (inout_generic_state.type == VCU_CH_PHYSICAL) {
            type = VCU_CH_PHYSICAL;
            entries = alloca((1 + 2 * total_channel_count(VCU_OUT)) *
                             sizeof(void *));
            for (n = 0; n < channel_count[type]; n++) {
                if (isruntime && !vcu.runtime.channel.get_active(io, type, n)) {
                    continue;
                }
                action2type[pos>>1] = type;
                action2idx[pos>>1] = n;
                entries[pos] = alloca(80);
                sprintf((char *)entries[pos++], "%c%d %s",
                        io == IN ? 'i' : 'o', n + 1, str);
                entries[pos++] = &delay[type][n];
            }
        }
        
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        elog(LOGDEBUG, "set delay [%d][%d][%d]",
             inout_generic_state.io,
             action2type[action],
             action2idx[action]);
        dly = delay[action2type[action]][action2idx[action]];
        dly = (double)((int)(dly * 100)) / 100.0;
        vcu.runtime.delay.set_trim(inout_generic_state.io,
                                   action2type[action],
                                   action2idx[action],
                                   dly / 1000.0);
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_inout_generic(int action)
{
    static struct textmenu_state *mstate;
    static int logical_count, physical_count, menupos = 0;
    
    switch (action) {
    case ACTION_CHILD_POP:
        if (mstate == NULL) {
            break;
        }
        if (logical_count != inout_generic_state.logical_count ||
            physical_count != inout_generic_state.physical_count)
        {
            textmenu_state_delete(mstate);
        } else {
            break;
        }
        /* fall through */
    case ACTION_INIT: {
        void *entries[20];
        int pos;
        
	if (!have_channels()) {
	    mstate = NULL;
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
        logical_count = inout_generic_state.logical_count;
        physical_count = inout_generic_state.physical_count;
        
        pos = 0;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Logical (%d)...",
                inout_generic_state.logical_count);
        entries[pos++] = inout_generic_state.state;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Physical (%d)...",
                inout_generic_state.physical_count);
        entries[pos++] = inout_generic_state.state;
        entries[pos++] = NULL;
	mstate = textmenu_state_new(inout_generic_state.heading, entries);
        mstate->pos = menupos;
        break;
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }

    menupos = mstate->pos;
    if (mstate->pos == 0) {
        inout_generic_state.type = -1;
    } else {
        inout_generic_state.type = VCU_CH_PHYSICAL;
    }
    
    textmenu_action(mstate, &action);
    
    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }
}

static bool_t
dummy_false(int io,
            int type,
            int idx)
{
    return false;
}

static void
count_nonzero_logi_phys(int io,
                        double (*get_value)(int io, int type, int idx),
                        bool_t (*get_bool)(int io, int type, int idx))
{
    int count, n, type;
    bool_t isruntime;
    
    isruntime = vcu.runtime.isruntime();
    inout_generic_state.logical_count = 0;
    for (type = 0; type < VCU_CH_COUNT; type++) {
        count = vcu.channel.get_count(io, type);
        for (n = 0; n < count; n++) {
            if ((!isruntime || vcu.runtime.channel.get_active(io, type, n)) &&
                (get_value(io, type, n) != 0.0 || get_bool(io, type, n)))
            {
                inout_generic_state.logical_count++;
            }
        }
    }
    inout_generic_state.physical_count = 0;
    type = VCU_CH_PHYSICAL;
    count = vcu.channel.get_count(io, type);
    for (n = 0; n < count; n++) {
        if ((!isruntime || vcu.runtime.channel.get_active(io, type, n)) &&
            (get_value(io, type, n) != 0.0 || get_bool(io, type, n)))
        {
            inout_generic_state.physical_count++;
        }
    }
}

static void
state_trim_volume_in(int action)
{
    if (action == ACTION_INIT || action == ACTION_CHILD_POP) {
        inout_generic_state.io = VCU_IN;
        inout_generic_state.state = state_inout_volume_edit;
        count_nonzero_logi_phys(VCU_IN, vcu.runtime.volume.get_trim,
                                vcu.runtime.volume.get_mute);
        strcpy(inout_generic_state.heading, "Trim volume in");
    }
    state_inout_generic(action);
}

static void
state_trim_volume_out(int action)
{
    if (action == ACTION_INIT || action == ACTION_CHILD_POP) {
        inout_generic_state.io = VCU_OUT;
        inout_generic_state.state = state_inout_volume_edit;
        count_nonzero_logi_phys(VCU_OUT, vcu.runtime.volume.get_trim,
                                vcu.runtime.volume.get_mute);
        strcpy(inout_generic_state.heading, "Trim volume out");
    }
    state_inout_generic(action);
}

static void
state_trim_delay_in(int action)
{
    if (action == ACTION_INIT || action == ACTION_CHILD_POP) {
        inout_generic_state.io = VCU_IN;
        inout_generic_state.state = state_inout_delay_edit;
        count_nonzero_logi_phys(VCU_IN, vcu.runtime.delay.get_trim,
                                dummy_false);
        strcpy(inout_generic_state.heading, "Trim delay in");
    }
    state_inout_generic(action);
}

static void
state_trim_delay_out(int action)
{
    if (action == ACTION_INIT || action == ACTION_CHILD_POP) {
        inout_generic_state.io = VCU_OUT;
        inout_generic_state.state = state_inout_delay_edit;
        count_nonzero_logi_phys(VCU_OUT, vcu.runtime.delay.get_trim,
                                dummy_false);
        strcpy(inout_generic_state.heading, "Trim delay out");
    }
    state_inout_generic(action);
}

static bool_t
rev_too_long(double length)
{
    int n, n_rates, rates[VCU_VALID_RATES_COUNT];
    double maxrevlen;
    
    if (vcu.iodelay.need_to_run_benchmark()) {
        /* we don't know maximum length */
        return false;
    }
    
    vcu.signal.get_rates(rates, &n_rates);
    for (n = 0; n < n_rates; n++) {
        vcu.iodelay.get_delay_length_pair(rates[n], NULL, &maxrevlen, NULL);
        if (maxrevlen < length) {
            return true;
        }
    }    
    return false;
}

static void
state_load_program_from_rom_or_cache(int action)
{
    static bool_t make_copy, success, copy_ask, mounted_path;
    static int load_reverb, n_programs, idx;
    static struct textmenu_state *mstate;
    static const struct filterprg *fprg;
    struct filterprg fprgcopy;    
    
    switch (action) {
    case ACTION_INIT: {
	void **entries;
	int n, pos;
	
	mstate = NULL;
        success = false;
        copy_ask = false;
        load_reverb = -1;
        idx = access_program_on_rom;
        mounted_path = false;
        if (access_program_on_rom != -1) {
            mounted_path = conf_static.rom[idx].mount;
            if (!mount_device_helper(idx)) {
                mstate = NULL;
                return;
            }
            fprg = vcu.ctl.scan_fprg_dir(conf_static.rom[idx].path,
                                         current_program_type,
                                         &n_programs);
        } else {
            fprg = vcu.cache.get_programs(&n_programs,
                                          current_program_type);
        }
        
	if (n_programs == 0) {
	    mstate = NULL;
	    switch (current_program_type) {
	    case VCU_PROGRAM_TYPE_REVERB:
                if (access_program_on_rom >= 0) {
                    textmenu_message("The medium stores no",
                                     "reverb programs     ");
                } else {
                    textmenu_message("There are no cached ",
                                     "reverb programs     ");
                }
		break;
	    case VCU_PROGRAM_TYPE_XTC:
                if (access_program_on_rom >= 0) {
                    textmenu_message("The medium stores no",
                                     "xtc programs        ");
                } else {
                    textmenu_message("There are no cached ",
                                     "xtc programs        ");
                }
		break;
	    case VCU_PROGRAM_TYPE_EQ:
                if (access_program_on_rom >= 0) {
                    textmenu_message("The medium stores no",
                                     "eq programs         ");
                } else {
                    textmenu_message("There are no cached ",
                                     "eq programs         ");
                }
		break;
	    }
	    return;
	}
	
	pos = 0;
	entries = alloca((1 + n_programs) * sizeof(void *));
	for (n = 0; n < n_programs; n++) {
	    /* FIXME: menu command characters in longname can make menu do dumb
	       stuff */
	    entries[pos] = alloca(40 + strlen(fprg[n].longname));
	    sprintf((char *)entries[pos], "%s", fprg[n].longname);
	    pos++;
	}
	entries[pos++] = NULL;
	switch (current_program_type) {
	case VCU_PROGRAM_TYPE_REVERB:
	    mstate = textmenu_state_new("Load reverb program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_XTC:
	    mstate = textmenu_state_new("Load xtc program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_EQ:
	    mstate = textmenu_state_new("Load eq program:", entries);
	    break;
	}
        break;
    }
    case ACTION_CHILD_POP:
        if (success) {
            if (copy_ask) {
                if (make_copy) {
                    tw_clear();
                    tw_putstring(0, 0, "Creating copy...    ");
                    tw_refresh();
                    /* FIXME: handle errors */
                    vcu.reverb.copy_program(&fprg[load_reverb],
                                            &fprgcopy);
                    load_reverb = -1;
                    success = load_reverb_program_helper(&fprgcopy,
                                                         true,
                                                         &make_copy,
                                                         &copy_ask);
                    return;
                }
            } else {
                tw_clear();
                state_load_program_from_rom_or_cache(ACTION_REFRESH);
            }
            return;
        }
        break;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_CHILD_POP:
        if (load_reverb == -1) {
            break;
        }
        action = load_reverb;
        /* fall through */
    default:
	if (action < n_programs) {
	    switch (current_program_type) {
	    case VCU_PROGRAM_TYPE_REVERB:
                if (load_reverb == -1 && rev_too_long(fprg[action].length)) {
                    textmenu_message("Reverb program will ",
                                     "be shortened        ");
                    load_reverb = action;
                    return;
                }
                load_reverb = -1;
		success = load_reverb_program_helper(&fprg[action],
                                                     mounted_path,
                                                     &make_copy,
                                                     &copy_ask);
                if (copy_ask) {
                    load_reverb = action;
                }
		break;
	    case VCU_PROGRAM_TYPE_XTC:
		success = load_xtc_program_helper(&fprg[action], mounted_path);
		break;
	    case VCU_PROGRAM_TYPE_EQ:
		success = load_eq_program_helper(&fprg[action], mounted_path);
		break;
	    }
	}
	return;
    case ACTION_EXIT:        
	if (access_program_on_rom != -1 && conf_static.rom[idx].mount) {
	    umount(conf_static.rom[idx].path);
            mount_path = NULL;
	}
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_REFRESH:
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_load_program(int action)
{
    static struct textmenu_state *mstate;
    static int cur_programs;
    static bool_t do_pop;
    static int menupos = 0;
    int n;

    switch (action) {
    case ACTION_CHILD_POP:
        if (do_pop) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
	textmenu_state_delete(mstate);
        /* fall through */
    case ACTION_INIT: {
	char heading[40], from_rom[ROM_COUNT][40];
	void *entries[5 + ROM_COUNT * 2];
	int pos = 0;
        
        do_pop = false;
	switch (current_program_type) {
	case VCU_PROGRAM_TYPE_REVERB:
            if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) == 0) {
                mstate = NULL;
                textmenu_message("There are no reverb ",
                                 "channels            ");
                do_pop = true;
                return;
            }
	    vcu.reverb.get_programs(&cur_programs);
	    sprintf(heading, "Space left: %.1fs",
		    vcu.reverb.get_capacity(NULL));
	    break;
	case VCU_PROGRAM_TYPE_XTC:
	    vcu.ambiopole.get_programs(&cur_programs);
	    sprintf(heading, "Space left: %.1fs",
		    vcu.ambiopole.get_capacity(NULL));
	    break;
	case VCU_PROGRAM_TYPE_EQ:
	    vcu.eq.get_programs(&cur_programs);
	    sprintf(heading, "Space left: %.1fs",
		    vcu.eq.get_capacity(NULL));
	    break;
	}

        for (n = 0; n < conf_static.n_roms; n++) {
            snprintf(from_rom[n], 40, "From %s...", conf_static.rom[n].name);
            from_rom[n][39] = '\0';	
            entries[pos++] = from_rom[n];
            entries[pos++] = state_load_program_from_rom_or_cache;
        }
        entries[pos++] = "From cache...";
	entries[pos++] = state_load_program_from_rom_or_cache;
	entries[pos++] = NULL;
	mstate = textmenu_state_new(heading, entries);
        mstate->pos = menupos;
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
        return;
    }

    if (mstate->pos < conf_static.n_roms) {
        access_program_on_rom = mstate->pos;
    } else {
        access_program_on_rom = -1;
    }
    
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        menupos = 0;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

/* FIXME: merge with _from_cache like the function
   state_load_program_from_rom_or_cache is made */
static void
state_del_program_from_memory(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprg;
    static int n_programs;
    
    if (action == ACTION_INIT) {
	void **entries;
	int n, pos;
	
	switch (current_program_type) {
	case VCU_PROGRAM_TYPE_REVERB:
            fprg = vcu.reverb.get_programs(&n_programs);	
	    break;
	case VCU_PROGRAM_TYPE_XTC:
            fprg = vcu.ambiopole.get_programs(&n_programs);	
	    break;
	case VCU_PROGRAM_TYPE_EQ:
            fprg = vcu.eq.get_programs(&n_programs);	
	    break;
	}
	if (n_programs == 0) {
	    mstate = NULL;
	    switch (current_program_type) {
	    case VCU_PROGRAM_TYPE_REVERB:
		textmenu_message("There are no reverb ",
				 "programs loaded     ");
		break;
	    case VCU_PROGRAM_TYPE_XTC:
		textmenu_message("There are no xtc    ",
				 "programs loaded     ");
		break;
	    case VCU_PROGRAM_TYPE_EQ:
		textmenu_message("There are no eq pro-",
				 "grams loaded        ");
		break;
	    }
	    return;
	}
	
	pos = 0;
	entries = alloca((1 + n_programs) * sizeof(void *));
	for (n = 0; n < n_programs; n++) {
	    /* FIXME: menu command characters in longname can make menu do dumb
	       stuff */
	    entries[pos] = alloca(40 + strlen(fprg->longname));
	    sprintf((char *)entries[pos], "%s|Delete program?",
                    fprg[n].longname);
	    pos++;
	}
	entries[pos++] = NULL;
	switch (current_program_type) {
	case VCU_PROGRAM_TYPE_REVERB:
	    mstate = textmenu_state_new("Del reverb program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_XTC:
	    mstate = textmenu_state_new("Del xtc program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_EQ:
	    mstate = textmenu_state_new("Del eq program:", entries);
	    break;
	}
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
	if (action < n_programs) {
            tw_clear();
            tw_putstring(0, 0, "Deleting program... ");
            tw_refresh();
	    switch (current_program_type) {
	    case VCU_PROGRAM_TYPE_REVERB:
		vcu.reverb.delete_program(fprg[action].id);
		break;
	    case VCU_PROGRAM_TYPE_XTC:
		vcu.ambiopole.delete_program(fprg[action].id);
		break;
	    case VCU_PROGRAM_TYPE_EQ:
		vcu.eq.delete_program(fprg[action].id);
		break;
	    }
	    textmenu_state_delete(mstate);
	    if (n_programs == 1) {
		/* the last program was deleted */
		textgui_pop_state();
	    } else {
		state_del_program_from_memory(ACTION_INIT);
	    }
	}
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_CHILD_POP:
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_del_program_from_cache(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprg;
    static int n_programs;
    
    if (action == ACTION_INIT) {
	void **entries;
	int n, pos;
	
	fprg = vcu.cache.get_programs(&n_programs, current_program_type);
	if (n_programs == 0) {
	    mstate = NULL;
	    switch (current_program_type) {
	    case VCU_PROGRAM_TYPE_REVERB:
		textmenu_message("There are no cached ",
				 "reverb programs     ");
		break;
	    case VCU_PROGRAM_TYPE_XTC:
		textmenu_message("There are no cached ",
				 "xtc programs        ");
		break;
	    case VCU_PROGRAM_TYPE_EQ:
		textmenu_message("There are no cached ",
				 "eq programs         ");
		break;
	    }
	    return;
	}
	
	pos = 0;
	entries = alloca((1 + n_programs) * sizeof(void *));
	for (n = 0; n < n_programs; n++) {
	    /* FIXME: menu command characters in longname can make menu do dumb
	       stuff */
	    entries[pos] = alloca(40 + strlen(fprg->longname));
	    sprintf((char *)entries[pos], "%s|Delete program?",
		    fprg[n].longname);
	    pos++;
	}
	entries[pos++] = NULL;
	switch (current_program_type) {
	case VCU_PROGRAM_TYPE_REVERB:
	    mstate = textmenu_state_new("Del reverb program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_XTC:
	    mstate = textmenu_state_new("Del xtc program:", entries);
	    break;
	case VCU_PROGRAM_TYPE_EQ:
	    mstate = textmenu_state_new("Del eq program:", entries);
	    break;
	}
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
	if (action < n_programs) {
	    vcu.cache.delete_program(fprg[action].id);
	    textmenu_state_delete(mstate);
	    if (n_programs == 1) {
		/* the last program was deleted */
		textgui_pop_state();
	    } else {
		state_del_program_from_cache(ACTION_INIT);
	    }
	}
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_CHILD_POP:
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_del_program(int action)
{
    static struct textmenu_state *mstate;

    switch (action) {
    case ACTION_INIT: {
	void *entries[5];
	int pos = 0;

	entries[pos++] = "From memory...";
	entries[pos++] = state_del_program_from_memory;
	entries[pos++] = "From cache...";
	entries[pos++] = state_del_program_from_cache;
	entries[pos++] = NULL;
	mstate = textmenu_state_new("Delete program:", entries);
    }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static int
scandir_select_pcm(const struct dirent *dirent)
{
    int slen;

    elog(LOGDEBUG, ">> (%s)", dirent->d_name);
    slen = strlen(dirent->d_name);
    if (strstr(dirent->d_name, ".pcm") == &dirent->d_name[slen - 4] ||
        strstr(dirent->d_name, ".raw") == &dirent->d_name[slen - 4] ||
        strstr(dirent->d_name, ".PCM") == &dirent->d_name[slen - 4] ||
        strstr(dirent->d_name, ".RAW") == &dirent->d_name[slen - 4])
    {
	return 1;
    }
    return 0;
}

static void
state_select_pcm_file_sub(int action)
{
    static struct textmenu_state *mstate;
    static struct dirent **namelist = NULL;
    static int n_files;

    char path[PATH_MAX];
    
    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        textgui_pop_state();
        return;
    case ACTION_INIT: {
	void **entries;
        int n, pos;
        
        if (!mount_device_helper(access_program_on_rom)) {
            mstate = NULL;
            return;
        }
        if (namelist != NULL) {
            while (n_files--) {
                free(namelist[n_files]);
            }
            free(namelist);
            namelist = NULL;
        }
        n_files = scandir(conf_static.rom[access_program_on_rom].path,
                          &namelist, scandir_select_pcm, alphasort);
	if (access_program_on_rom != -1 &&
            conf_static.rom[access_program_on_rom].mount)
        {
	    umount(conf_static.rom[access_program_on_rom].path);
            mount_path = NULL;
	}
        if (n_files == 0) {
            mstate = NULL;
            textmenu_message("The medium stores no",
                             ".pcm/.raw files     ");
            return;
        } else if (n_files < 0) {
            elog(LOGDEBUG, "scandir in \"%s\" failed: %s",
                 conf_static.rom[access_program_on_rom].path, strerror(errno));
            mstate = NULL;
            textmenu_message("Failed to list dir  ",
                             "on medium           ");
            return;
        }
        entries = alloca((30 + n_files) * sizeof(void *));
        pos = 0;
        for (n = 0; n < n_files; n++) {
            entries[pos] = alloca(strlen(namelist[n]->d_name) + 10);
            sprintf((char *)entries[pos++], "\"%s\"", namelist[n]->d_name);
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new("Select PCM file:", entries);
    }    
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        snprintf(path, PATH_MAX, "%s/%s",
                 conf_static.rom[access_program_on_rom].path,
                 namelist[action]->d_name);
        path[PATH_MAX-1] = '\0';
        free(conf_user.file.path);
        conf_user.file.path = strdup(path);
        vcu.ctl.schedule_save_config();
        /* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_select_pcm_file(int action)
{
    static struct textmenu_state *mstate;
    int n, pos;

    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[30 + ROM_COUNT];

        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            snprintf(rom[n], 40, "From %s...", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
            entries[pos++] = state_select_pcm_file_sub;
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new("Select PCM file:", entries);
    }
    }

    access_program_on_rom = mstate->pos;
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_pcm_file_setup(int action)
{
    static struct textmenu_state *mstate = NULL;
    static int sample_format;

    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        state_pcm_file_setup(ACTION_INIT);
        return;
    case ACTION_INIT: {
        char sformat[100 + BF_SAMPLE_FORMAT_MAX * 20], *fname;
        void *entries[100];
        int pos = 0, n;

        strcpy(sformat, "Format [");
        for (n = BF_SAMPLE_FORMAT_MIN; n <= BF_SAMPLE_FORMAT_MAX; n++) {
            strcat(sformat, bf_strsampleformat(n));
            strcat(sformat, "/");
        }
        strcpy(&sformat[strlen(sformat)-1], "]");

        sample_format = conf_user.file.sample_format - BF_SAMPLE_FORMAT_MIN;
        entries[pos++] = "Select file...";
        entries[pos++] = state_select_pcm_file;
        entries[pos++] = sformat;
        entries[pos++] = &sample_format;
        entries[pos] = alloca(80);
        sprintf((char *)entries[pos++], "Channel count [%%d,1,10,1,%d]",
                VCU_MAX_CHANNELS);
        entries[pos++] = &conf_user.file.channel_count;
        entries[pos++] = "Loop [off/on]";
        entries[pos++] = &conf_user.file.loop;
	entries[pos++] = NULL;
        if (conf_user.file.path == NULL || conf_user.file.path[0] == '\0') {
            fname = alloca(40);
            strcpy(fname, "No file selected");
        } else {
            fname = alloca(strlen(conf_user.file.path) + 10);
            if (strrchr(conf_user.file.path, '/') != NULL) {
                sprintf(fname, "\"%s\"", strrchr(conf_user.file.path, '/') + 1);
            } else {
                sprintf(fname, "\"%s\"", conf_user.file.path);
            }
        }
	mstate = textmenu_state_new(fname, entries);
        break;
    }
    }

    textmenu_action(mstate, &action);
    
    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        if (action == 1) {
            conf_user.file.sample_format = sample_format + BF_SAMPLE_FORMAT_MIN;
        }
        vcu.ctl.schedule_save_config();
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_signal_setup(int action)
{
    static struct textmenu_state *mstate;
    static int output_bits;
    static bool_t allow_rate[VCU_VALID_RATES_COUNT];
    static bool_t init_allow_rate[VCU_VALID_RATES_COUNT];    
    static int bits[64], n_bits;
    static int rates[VCU_VALID_RATES_COUNT], n_rates;
    static int suggested_rate, suggested_rate_pos, allowed_rates_base;
    static int apply_dither_pos, watch_clock_pos, set_bits_pos, dp_pos;
    static bool_t apply_dither, watch_clock, double_precision;
    static int menupos = 0;

    int usedrates[VCU_VALID_RATES_COUNT], n_usedrates, n;

    /* pre textmenu_action action handling */
    switch (action) {
    case ACTION_INIT: {
	char bit_output[200], allow_rate_str[VCU_VALID_RATES_COUNT][30];
	char suggest[200], s[100];
	void *entries[100];
	int pos, i, posoffset;

	apply_dither = vcu.signal.get_dither();
        watch_clock = vcu.signal.get_watch_clock();
        double_precision = vcu.signal.get_double_precision();
        apply_dither_pos = -1;
        watch_clock_pos = -1;
        dp_pos = -1;
        set_bits_pos = -1;
        suggested_rate_pos = -1;
        allowed_rates_base = -1;

	pos = 0;
        posoffset = 0;

        entries[pos++] = "Input [sound card/pcm file]";
        entries[pos++] = &conf_user.file.use;

        entries[pos++] = "PCM file setup...";
        entries[pos++] = state_pcm_file_setup;
        
	vcu.signal.get_allowed_output_bits(bits, &n_bits);
	if (n_bits > 1) {
	    output_bits = vcu.signal.get_output_bits();
	    for (n = 0; n < n_bits; n++) {
		if (bits[n] == output_bits) {
		    output_bits = n;
		    break;
		}
	    }
	    /* create a string like "[24/16] bit output" */
	    strcpy(bit_output, "[");
	    for (n = 0; n < n_bits - 1; n++) {
		sprintf(s, "%d/", bits[n]);
		strcat(bit_output, s);
	    }
	    sprintf(s, "%d] bit output", bits[n]);
	    strcat(bit_output, s);
            set_bits_pos = pos / 2;
	    entries[pos++] = bit_output;
	    entries[pos++] = &output_bits;
	}
        dp_pos =  pos / 2;;
        entries[pos++] = "[32/64] bit internal";
        entries[pos++] = &double_precision;

        if (vcu.signal.get_max_dither_bits() != 0) {
            if (bits[output_bits] > vcu.signal.get_max_dither_bits()) {
                entries[pos++] = "Dither off";
                posoffset += 1;
            } else {
                apply_dither_pos = pos / 2;
                entries[pos++] = "Dither [off/on]";
                entries[pos++] = &apply_dither;
            }
        }
        watch_clock_pos =  (pos + posoffset) / 2;;
        entries[pos++] = "Watch clock [off/on]";
        entries[pos++] = &watch_clock;
        
	vcu.signal.get_allowed_rates(rates, &n_rates);
	vcu.signal.get_rates(usedrates, &n_usedrates);
	if (n_rates > 1) {
            allowed_rates_base = (pos + posoffset) / 2;
	    memset(allow_rate, 0, sizeof(allow_rate));
	    memset(init_allow_rate, 0, sizeof(init_allow_rate));
	    for (n = 0; n < n_rates; n++) {
		for (i = 0; i < n_usedrates; i++) {
		    if (usedrates[i] == rates[n]) {
			allow_rate[n] = true;
			init_allow_rate[n] = true;
			break;
		    }
		}
		sprintf(allow_rate_str[n], "Allow %.1fkHz[off/on]",
			(double)rates[n] / 1000.0);
		entries[pos++] = allow_rate_str[n];
		entries[pos++] = &allow_rate[n];
	    }
            
	} else {
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++], "Rate %.1fkHz",
                    (double)rates[0] / 1000.0);
            posoffset += 1;
        }
	if (n_usedrates > 1) {
            /* create a string like "Suggest [44.1/48.0]kHz" */
            strcpy(suggest, "Suggest [");
            suggested_rate = 0;
            for (n = 0; n < n_rates; n++) {
                if (vcu.signal.get_suggested_rate() == rates[n]) {
                    suggested_rate = n;
                }
                sprintf(s, "%.1f/", (double)rates[n] / 1000.0);
                strcat(suggest, s);
            }
            suggest[strlen(suggest) - 1] = ']';
            strcat(suggest, "kHz");
            suggested_rate_pos = (pos + posoffset) / 2;
            entries[pos++] = suggest;
            entries[pos++] = &suggested_rate;
        }
	entries[pos] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
	break;
    }
    case ACTION_CHILD_POP:
	break;
    default:
	break;
    }
    
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    /* post textmenu_action action handling */
    switch (action) {
    case 0:
        vcu.ctl.schedule_save_config();
        break;
    default:
        if (action < 0) {
            return;
        }
        if (action == apply_dither_pos) {
            vcu.signal.set_dither(apply_dither);
        } else if (action == watch_clock_pos) {
            vcu.signal.set_watch_clock(watch_clock);
        } else if (action == dp_pos) {
            vcu.signal.set_double_precision(double_precision);
            textmenu_state_delete(mstate);
            state_signal_setup(ACTION_INIT);
            return;
        } else if (action == set_bits_pos) {
	    vcu.signal.set_output_bits(bits[output_bits]);
            textmenu_state_delete(mstate);
            state_signal_setup(ACTION_INIT);
            return;
        } else if (action == suggested_rate_pos) {
            vcu.signal.set_suggested_rate(rates[suggested_rate]);
        } else if (action >= allowed_rates_base) {
            n_usedrates = 0;
            for (n = 0; n < n_rates; n++) {
                if (allow_rate[n]) {
                    usedrates[n_usedrates] = rates[n];
                    n_usedrates++;
                }
            }
            if (n_usedrates == 0) {
                n_usedrates = 1;
                if (action == allowed_rates_base) {
                    usedrates[0] = rates[1];
                } else {
                    usedrates[0] = rates[0];
                }
            }
            vcu.signal.set_rates(usedrates, n_usedrates);
            textmenu_state_delete(mstate);
            state_signal_setup(ACTION_INIT);
            return;
        }
	break;
    case ACTION_EXIT:	
        elog(LOGDEBUG, "calling update_hard_settings");
        UPDATE_HARD_SETTINGS_CALL;
	textmenu_state_delete(mstate);
	textgui_pop_state();	
        menupos = 0;
	break;
    case ACTION_CHILD_POP:
        UPDATE_HARD_SETTINGS_CHILD_POP;
        menupos = 0;
	return;
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }    
}

static void
state_channel_setup(int action)
{
    static struct textmenu_state *mstate;
    static int channel_count[VCU_CH_COUNT];
    int type;

    for (type = 0; type < VCU_CH_COUNT; type++) {
        channel_count[type] = vcu.channel.get_count(VCU_OUT, type);
    }
    channel_count[VCU_CH_AMBIOPOLE] /= 2;
	
    if (action == ACTION_INIT) {
	void *entries[100];
	int pos;

	pos = 0;
        entries[pos] = alloca(40);
	sprintf((char *)entries[pos++], "[%%d,1,10,0,%d] (A)mbiopole",
		vcu.channel.get_max_count(VCU_OUT, VCU_CH_AMBIOPOLE) / 2);
	entries[pos++] = &channel_count[VCU_CH_AMBIOPOLE];
        
        entries[pos] = alloca(40);
	sprintf((char *)entries[pos++], "[%%d,1,10,0,%d] (R)everb",
		vcu.channel.get_max_count(VCU_OUT, VCU_CH_REVERB));
	entries[pos++] = &channel_count[VCU_CH_REVERB];
        
        entries[pos] = alloca(40);
	sprintf((char *)entries[pos++], "[%%d,1,10,0,%d] am(B)isonics",
		vcu.channel.get_max_count(VCU_OUT, VCU_CH_AMBISONIC));
	entries[pos++] = &channel_count[VCU_CH_AMBISONIC];
        
        entries[pos] = alloca(40);
	sprintf((char *)entries[pos++], "[%%d,1,10,0,%d] (P)assthru",
		vcu.channel.get_max_count(VCU_OUT, VCU_CH_PASSTHRU));
	entries[pos++] = &channel_count[VCU_CH_PASSTHRU];

	entries[pos++] = "Reverb setup...";
	entries[pos++] = state_reverb_setup;
	entries[pos++] = "Ambisonic setup...";
	entries[pos++] = state_ambisonic_setup;
        entries[pos++] = "Input matrix...";
	entries[pos++] = state_mix_inputs;
	entries[pos++] = "Output matrix...";
	entries[pos++] = state_mix_outputs;
	entries[pos++] = "Speaker layout...";
	entries[pos++] = state_speaker_layout;
	entries[pos++] = "Stage layout...";
	entries[pos++] = state_stage_layout;
        entries[pos++] = "Channel groups...";
        entries[pos++] = state_channel_groups;
        entries[pos++] = "Channel modes...";
        entries[pos++] = state_channel_modes;
	entries[pos++] = "Trim volume in...";
	entries[pos++] = state_trim_volume_in;
	entries[pos++] = "Trim delay in...";
	entries[pos++] = state_trim_delay_in;
	entries[pos++] = "Trim volume out...";
	entries[pos++] = state_trim_volume_out;
	entries[pos++] = "Trim delay out...";
	entries[pos++] = state_trim_delay_out;
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action >= 0 && action <= VCU_CH_COUNT) {
            channel_count[VCU_CH_AMBIOPOLE] *= 2;
            vcu.channel.set_count(action, channel_count[action]);
            channel_count[VCU_CH_AMBIOPOLE] /= 2;
        }
        return;
    case ACTION_EXIT:
	if (!have_channels()) {
	    textmenu_message("Must have more than ",
			     "zero channels       ");
	    return;
	}
        elog(LOGDEBUG, "calling update_hard_settings");
        UPDATE_HARD_SETTINGS_CALL;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	break;
    case ACTION_CHILD_POP:
        UPDATE_HARD_SETTINGS_CHILD_POP;
	return;
    }    
}

static void
state_eq_custom(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprg;
    static int n_programs;
    
    if (action == ACTION_INIT) {
	void **entries;
	int n, pos;
        
        fprg = vcu.eq.get_programs(&n_programs);
        if (n_programs == 0) {
            mstate = NULL;
            textmenu_message("There are no custom ",
                             "programs loaded     ");
            return;
        }
	pos = 0;
	entries = alloca((2 + n_programs) * sizeof(void *));
	for (n = 0; n < n_programs; n++) {
	    /* FIXME: menu command characters in longname can make menu do dumb
	       stuff */
	    entries[pos] = alloca(40 + strlen(fprg[n].longname));
	    sprintf((char *)entries[pos], "%s", fprg[n].longname);
	    pos++;
	}
        if (eq_setup_state.modify) {
            entries[pos++] = "Delete filter?";
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqpart));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_CUSTOM;
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
	if (action < 0 || action > n_programs) {
	    return;
	}
        if (action == n_programs) {
            eq_setup_state.delete = true;            
        } else {
            eq_setup_state.part.params.custom.id = fprg[action].id;
        }
        eq_setup_state.did_something = true;
	/* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_eq_del_op(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        mstate = textmenu_state_new(
            NULL,
            (void *[]){ "Delete operation?", NULL });
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        eq_setup_state.delete = true;
	eq_setup_state.did_something = true;
        /* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_multiply(int action)
{
    static struct textmenu_state *mstate;
    static double amp;
    static int sign;
    
    if (action == ACTION_INIT) {
        void *entries[10];
        double range[2];
        int pos;
        
        vcu.runtime.volume.get_range(range);
	eq_setup_state.did_something = false;
        pos = 0;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Amplify [%%+.1f,0.1,10,%f,%f]dB",
                range[0], range[1]);
        entries[pos++] = &amp;
        entries[pos++] = "Sign [+/-]";
        entries[pos++] = &sign;
        if (eq_setup_state.modify) {
            amp = eq_setup_state.part.params.multiply.amp;
            sign = (eq_setup_state.part.params.multiply.sign > 0) ? 0 : 1;
            entries[pos++] = "Delete operation?";
        } else {
            eq_setup_state.part.type = VCU_EQPRG_TYPE_OP_MULTIPLY;
            amp = 0;
            sign = 0;
            entries[pos++] = "Create operation?";
        }
        entries[pos++] = NULL;
        mstate = textmenu_state_new("Multiply:", entries);
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 2:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        }
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 2) {
            eq_setup_state.did_something = true;
            eq_setup_state.part.params.multiply.amp = amp;
            eq_setup_state.part.params.multiply.sign = (sign == 0) ? 1 : -1;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_lowpass(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Lowpass settings:",
                (void *[]){ "Cutoff [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.lowpass.cutoff,
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.lowpass.slope,
                            "Delete filter?", NULL});
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqpart));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_LOWPASS;
            eq_setup_state.part.params.lowpass.cutoff = 2000.0;
            eq_setup_state.part.params.lowpass.phase = VCU_PHASE_LINEAR;
            eq_setup_state.part.params.lowpass.slope = 0.5;
            mstate = textmenu_state_new(
                "Lowpass settings:",
                (void *[]){ "Cutoff [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.lowpass.cutoff,
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.lowpass.slope,
                            "Create filter?", NULL});
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 2:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        }
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 2) {
            eq_setup_state.did_something = true;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_highpass(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Highpass settings:",
                (void *[]){ "Cutoff [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.highpass.cutoff,
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.highpass.slope,
                            "Delete filter?", NULL });
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_HIGHPASS;
            eq_setup_state.part.params.highpass.cutoff = 2000.0;
            eq_setup_state.part.params.highpass.phase = VCU_PHASE_LINEAR;
            eq_setup_state.part.params.highpass.slope = 0.5;
            mstate = textmenu_state_new(
                "Highpass settings:",
                (void *[]){ "Cutoff [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.highpass.cutoff,
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.highpass.slope,
                            "Create filter?", NULL });
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 2:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        }
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 2) {
            eq_setup_state.did_something = true;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_bandpass(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Bandpass settings:",
                (void *[]){ "Cutoff 1 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandpass.cutoff[0],
                            "Cutoff 2 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandpass.cutoff[1],
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.bandpass.slope,
                            "Delete filter?", NULL });
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_BANDPASS;
            eq_setup_state.part.params.bandpass.cutoff[0] = 2000.0;
            eq_setup_state.part.params.bandpass.cutoff[1] = 3000.0;
            eq_setup_state.part.params.bandpass.phase = VCU_PHASE_LINEAR;
            eq_setup_state.part.params.bandpass.slope = 0.5;
            mstate = textmenu_state_new(
                "Bandpass settings:",
                (void *[]){ "Cutoff 1 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandpass.cutoff[0],
                            "Cutoff 2 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandpass.cutoff[1],
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.bandpass.slope,
                            "Create filter?", NULL });
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 3:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        } else if (eq_setup_state.part.params.bandstop.cutoff[0] >=
                   eq_setup_state.part.params.bandstop.cutoff[1])
        {
	    textmenu_message("Cutoff 1 is not less",
			     "than cutoff 2       ");
            return;
        }
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 3) {
            eq_setup_state.did_something = true;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_bandstop(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Bandstop settings:",
                (void *[]){ "Cutoff 1 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandstop.cutoff[0],
                            "Cutoff 2 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandstop.cutoff[1],
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.bandstop.slope,
                            "Delete filter?", NULL });
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_BANDSTOP;
            eq_setup_state.part.params.bandstop.cutoff[0] = 2000.0;
            eq_setup_state.part.params.bandstop.cutoff[1] = 3000.0;
            eq_setup_state.part.params.bandstop.phase = VCU_PHASE_LINEAR;
            eq_setup_state.part.params.bandstop.slope = 0.5;
            mstate = textmenu_state_new(
                "Bandstop settings:",
                (void *[]){ "Cutoff 1 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandstop.cutoff[0],
                            "Cutoff 2 [%.0f,1,1000,5,20000]Hz",
                            &eq_setup_state.part.params.bandstop.cutoff[1],
                            "Sharpness [%.2f,0.01,0.1,0,1]",
                            &eq_setup_state.part.params.bandstop.slope,
                            "Create filter?", NULL });
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 3:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        } else if (eq_setup_state.part.params.bandstop.cutoff[0] >=
                   eq_setup_state.part.params.bandstop.cutoff[1])
        {
	    textmenu_message("Cutoff 1 is not less",
			     "than cutoff 2       ");
            return;
        }
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 3) {
            eq_setup_state.did_something = true;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_allpass(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Allpass:", (void *[]){ "Delete filter?", NULL });
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_ALLPASS;
            mstate = textmenu_state_new(
                "Allpass:", (void *[]){ "Create filter?", NULL });
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        }
	eq_setup_state.did_something = true;
        /* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_random_phase(int action)
{
    static struct textmenu_state *mstate;
    static double length;
    
    if (action == ACTION_INIT) {
	eq_setup_state.did_something = false;
        if (eq_setup_state.modify) {
            mstate = textmenu_state_new(
                "Random phase:",
                (void *[]){ "Length [%.0f,1,1000,1,10000]ms", &length,
                            "Delete filter?", NULL });
        } else {
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = VCU_EQPRG_TYPE_RANDOM_PHASE;
            length = 10.0;
            mstate = textmenu_state_new(
                "Random phase:",
                (void *[]){ "Length [%.0f,1,1000,1,10000]ms", &length,
                            "Create filter?", NULL });
        }
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 1:
        if (eq_setup_state.modify) {
            eq_setup_state.delete = true;
        }
	eq_setup_state.did_something = true;
        /* fall through */
    case ACTION_EXIT:
        if (eq_setup_state.modify || action == 1) {
            eq_setup_state.part.params.random_phase.length =
                length / 1000.0;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    default:
        break;
    }
}

static void
state_eq_adjust_fft(int action)
{
    static struct textmenu_state *mstate;

    if (action == ACTION_INIT) {
        void *entries[2 * VCU_EQPRG_MAX_BANDS + 2];
        double range[2];
        int pos, n;

        eq_setup_state.did_something = false;
        pos = 0;
        vcu.runtime.eq.get_dynamic_equaliser_volume_range(range);
        for (n = 0; n < eq_setup_state.part.params.fft.n_bands; n++) {
            entries[pos] = alloca(80);
            if (eq_setup_state.part.params.fft.band[n] < 1000.0) {
                sprintf((char *)entries[pos++],
                        "%.1fHz [%%+.1f,0.1,10,%f,%f]dB",
                        eq_setup_state.part.params.fft.band[n],
                        range[0], range[1]);
            } else {
                sprintf((char *)entries[pos++],
                        "%.0fHz [%%+.1f,0.1,10,%f,%f]dB",
                        eq_setup_state.part.params.fft.band[n],
                        range[0], range[1]);
            }
            entries[pos++] = &eq_setup_state.part.params.fft.amp[n];
        }
        if (!vcu.runtime.isruntime()) {
            if (eq_setup_state.modify) {
                entries[pos++] = "Delete filter?";
            } else {
                entries[pos++] = "Create filter?";
            }
        }
        entries[pos++] = NULL;        
	mstate = textmenu_state_new(NULL, entries);
    }    
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action >= 0) {
            if (eq_setup_state.modify) {
                eq_setup_state.did_something = true;
            }
            if (vcu.runtime.isruntime()) {
                /* assume success */
                vcu.runtime.eq.modify_dynamic_equaliser(
                    eq_setup_state.current_eq_index,
                    &eq_setup_state.part);
            }
            if (action == eq_setup_state.part.params.fft.n_bands) {
                eq_setup_state.did_something = true;
                if (eq_setup_state.modify) {
                    eq_setup_state.delete = true;
                }
                state_eq_adjust_fft(ACTION_EXIT);
            }
        }
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        return;
    }    
}

static void
state_eq_custom_bands(int action)
{
    static struct textmenu_state *mstate;
    static int menupos = 0;
    int n_bands, n;
    
    if (action == ACTION_INIT) {
        void *entries[2 * VCU_EQPRG_MAX_BANDS + 20];
        int pos, type;

        if (!eq_setup_state.modify) {
            type = eq_setup_state.part.type;
            memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
            eq_setup_state.part.type = type;
            eq_setup_state.part.params.fft.n_bands = 2;
            eq_setup_state.part.params.fft.band[0] = 2000;
            eq_setup_state.part.params.fft.band[1] = 4000;
        }
        eq_setup_state.modify = true;

        pos = 0;        
        for (n = 0; n < eq_setup_state.part.params.fft.n_bands; n++) {
            entries[pos] = alloca(80);
            sprintf((char *)entries[pos++], "Band%d [%%.0f,1,1000,5,20000]Hz",
                    n + 1);
            entries[pos++] = &eq_setup_state.part.params.fft.band[n];
        }
        entries[pos++] = "Add band?";
        entries[pos++] = "Accept current?";
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }    
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action == eq_setup_state.part.params.fft.n_bands) {
            /* Add band */
            if (eq_setup_state.part.params.fft.n_bands ==
                VCU_EQPRG_MAX_BANDS)
            {
                textmenu_message("Cannot add more     ",
                                 "bands               ");
                return;
            }
            n_bands = eq_setup_state.part.params.fft.n_bands;
            if (eq_setup_state.part.params.fft.n_bands > 0) {
                eq_setup_state.part.params.fft.band[n_bands] =
                    eq_setup_state.part.params.fft.band[n_bands - 1];
            }
            eq_setup_state.part.params.fft.n_bands += 1;
            menupos = n_bands;
            textmenu_state_delete(mstate);
            state_eq_custom_bands(ACTION_INIT);
            return;
        }
        if (action == eq_setup_state.part.params.fft.n_bands + 1) {
            /* Accept current */
            for (n = 1;
                 n < eq_setup_state.part.params.fft.n_bands;
                 n++)
            {
                if (eq_setup_state.part.params.fft.band[n] <=
                    eq_setup_state.part.params.fft.band[n-1])
                {
                    textmenu_message("The bands are not   ",
                                     "properly sorted     ");
                    return;
                }
            }            
            for (n = 1;
                 n < eq_setup_state.part.params.fft.n_bands;
                 n++)
            {
                if (eq_setup_state.part.params.fft.band[n] -
                    vcu.runtime.eq.get_dynamic_equaliser_smallest_band() <
                    eq_setup_state.part.params.fft.band[n-1])
                {
                    char str[100];
                    sprintf(str, "Must be at least %.0f", vcu.runtime.eq.
                            get_dynamic_equaliser_smallest_band());
                    textmenu_message(str,
                                     "Hz between bands    ");
                    return;
                }
            }            
            eq_setup_state.did_something = true;
            state_eq_custom_bands(ACTION_EXIT);
            return;
        }
        return;
    case ACTION_EXIT:
        menupos = 0;
        eq_setup_state.modify = false;
	textmenu_state_delete(mstate);
	textgui_pop_state();        
        return;
    }
}

static void
state_eq_fft(int action)
{
    static struct textmenu_state *mstate;
    
    if (action == ACTION_INIT) {
        int type;
        
        eq_setup_state.did_something = false;
        eq_setup_state.modify = false;
        type = eq_setup_state.part.type;
        memset(&eq_setup_state.part, 0, sizeof(struct vcu_eqprg));
        eq_setup_state.part.type = type;
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
		"ISO oct bands?",
		"ISO 1/3 oct bands?",
		"Custom bands...", state_eq_custom_bands,
		NULL
	    });
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        /* ISO octave bands */
        eq_setup_state.part.params.fft.n_bands = 10;
        eq_setup_state.part.params.fft.band[0] = 31.5;
        eq_setup_state.part.params.fft.band[1] = 63;
        eq_setup_state.part.params.fft.band[2] = 125;
        eq_setup_state.part.params.fft.band[3] = 250;
        eq_setup_state.part.params.fft.band[4] = 500;
        eq_setup_state.part.params.fft.band[5] = 1000;
        eq_setup_state.part.params.fft.band[6] = 2000;
        eq_setup_state.part.params.fft.band[7] = 4000;
        eq_setup_state.part.params.fft.band[8] = 8000;
        eq_setup_state.part.params.fft.band[9] = 16000;
        goto eq_fft_finished_bands;
    case 1:
        /* ISO 1/3 octave bands */
        eq_setup_state.part.params.fft.n_bands = 31;
        eq_setup_state.part.params.fft.band[0] = 20;
        eq_setup_state.part.params.fft.band[1] = 25;
        eq_setup_state.part.params.fft.band[2] = 31;
        eq_setup_state.part.params.fft.band[3] = 40;
        eq_setup_state.part.params.fft.band[4] = 50;
        eq_setup_state.part.params.fft.band[5] = 63;
        eq_setup_state.part.params.fft.band[6] = 80;
        eq_setup_state.part.params.fft.band[7] = 100;
        eq_setup_state.part.params.fft.band[8] = 125;
        eq_setup_state.part.params.fft.band[9] = 160;
        eq_setup_state.part.params.fft.band[10] = 200;
        eq_setup_state.part.params.fft.band[11] = 250;
        eq_setup_state.part.params.fft.band[12] = 315;
        eq_setup_state.part.params.fft.band[13] = 400;
        eq_setup_state.part.params.fft.band[14] = 500;
        eq_setup_state.part.params.fft.band[15] = 630;
        eq_setup_state.part.params.fft.band[16] = 800;
        eq_setup_state.part.params.fft.band[17] = 1000;
        eq_setup_state.part.params.fft.band[18] = 1250;
        eq_setup_state.part.params.fft.band[19] = 1600;
        eq_setup_state.part.params.fft.band[20] = 2000;
        eq_setup_state.part.params.fft.band[21] = 2500;
        eq_setup_state.part.params.fft.band[22] = 3150;
        eq_setup_state.part.params.fft.band[23] = 4000;
        eq_setup_state.part.params.fft.band[24] = 5000;
        eq_setup_state.part.params.fft.band[25] = 6300;
        eq_setup_state.part.params.fft.band[26] = 8000;
        eq_setup_state.part.params.fft.band[27] = 10000;
        eq_setup_state.part.params.fft.band[28] = 12500;
        eq_setup_state.part.params.fft.band[29] = 16000;
        eq_setup_state.part.params.fft.band[30] = 20000;
        goto eq_fft_finished_bands;
    case ACTION_CHILD_POP:
	if (!eq_setup_state.did_something) {
	    return;
	}
    eq_fft_finished_bands:
        state_stack[stackpos] = state_eq_adjust_fft;
        state_eq_adjust_fft(ACTION_INIT);
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    default:
	return;
    }        
}

static void
state_eq_add_op(int action)
{
    static struct textmenu_state *mstate;

    switch (action) {
    case ACTION_INIT: {
        void *entries[100];
        int pos;
	eq_setup_state.did_something = false;
        pos = 0;
        entries[pos++] = "Convolve?";
        entries[pos++] = "Add?";
        entries[pos++] = "Subtract?";
        entries[pos++] = "Multiply...";
        entries[pos++] = state_eq_multiply;
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        memset(&eq_setup_state.part, 0, sizeof(eq_setup_state.part));
        eq_setup_state.part.type = VCU_EQPRG_TYPE_OP_CONVOLVE;
        eq_setup_state.did_something = true;
        state_eq_add_op(ACTION_EXIT);
        return;
    case 1:
        memset(&eq_setup_state.part, 0, sizeof(eq_setup_state.part));
        eq_setup_state.part.type = VCU_EQPRG_TYPE_OP_ADD;
        eq_setup_state.did_something = true;
        state_eq_add_op(ACTION_EXIT);
        return;
    case 2:
        memset(&eq_setup_state.part, 0, sizeof(eq_setup_state.part));
        eq_setup_state.part.type = VCU_EQPRG_TYPE_OP_SUBTRACT;
        eq_setup_state.did_something = true;
        state_eq_add_op(ACTION_EXIT);
        return;
    case ACTION_CHILD_POP:
	if (!eq_setup_state.did_something) {
	    return;
	}
	/* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    default:
	return;
    }        
}

static void
state_eq_add_filter(int action)
{
    static struct textmenu_state *mstate;
    static int dyneq_pos = -1;

    switch (action) {
    case ACTION_INIT: {
        void *entries[100];
        int pos;
	eq_setup_state.did_something = false;
        pos = 0;
        dyneq_pos = -1;
        entries[pos++] = "Lowpass...";
        entries[pos++] = state_eq_lowpass;
        entries[pos++] = "Highpass...";
        entries[pos++] = state_eq_highpass;
        entries[pos++] = "Bandpass...";
        entries[pos++] = state_eq_bandpass;
        entries[pos++] = "Bandstop...";
        entries[pos++] = state_eq_bandstop;
        entries[pos++] = "Allpass...";
        entries[pos++] = state_eq_allpass;
        entries[pos++] = "Random phase...";
        entries[pos++] = state_eq_random_phase;
        entries[pos++] = "Custom...";
        entries[pos++] = state_eq_custom;
        entries[pos++] = "FFT static...";
        entries[pos++] = state_eq_fft;
        if (eq_setup_state.current_eq_index < 0 ||
            eq_setup_state.current_eq_index >= vcu.eq.get_equaliser_count())
        {
            dyneq_pos = pos / 2;
            entries[pos++] = "FFT dynamic...";
            entries[pos++] = state_eq_fft;
        }
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
	break;
    }
    case ACTION_CHILD_POP:
	if (!eq_setup_state.did_something) {
            break;
	}
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    default:
	break;
    }

    if (mstate->pos == dyneq_pos) {
        eq_setup_state.part.type = VCU_EQPRG_TYPE_FFT_DYNAMIC;
    } else {
        eq_setup_state.part.type = VCU_EQPRG_TYPE_FFT_STATIC;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    default:
	return;
    }        
}

static const char *
eq_part_str(const struct vcu_eqpart *part)
{
    static char str[100];
    const struct filterprg *fprgs;
    int i, n_programs;
    
    switch (part->type) {
    case VCU_EQPRG_TYPE_OP_CONVOLVE:
        sprintf(str, "op: convolve");
        break;
    case VCU_EQPRG_TYPE_OP_ADD:
        sprintf(str, "op: add");
        break;
    case VCU_EQPRG_TYPE_OP_SUBTRACT:
        sprintf(str, "op: subtract");
        break;
    case VCU_EQPRG_TYPE_OP_MULTIPLY:
        sprintf(str, "op: multiply (%+.1f %c)",
                part->params.multiply.amp,
                part->params.multiply.sign < 0 ? '-' : '+');
        break;
    case VCU_EQPRG_TYPE_LOWPASS:
        sprintf(str, "lowpass %dHz (%.2f)",
                (int)part->params.lowpass.cutoff,
                part->params.lowpass.slope);
        break;
    case VCU_EQPRG_TYPE_HIGHPASS:
        sprintf(str, "highpass %dHz (%.2f)",
                (int)part->params.highpass.cutoff,
                part->params.highpass.slope);
        break;
    case VCU_EQPRG_TYPE_BANDPASS:
        sprintf(str, "bandpass %d-%dHz (%.2f)",
                (int)part->params.bandpass.cutoff[0],
                (int)part->params.bandpass.cutoff[1],
                part->params.bandpass.slope);
        break;
    case VCU_EQPRG_TYPE_BANDSTOP:
        sprintf(str, "bandstop %d-%dHz (%.2f)",
                (int)part->params.bandstop.cutoff[0],
                (int)part->params.bandstop.cutoff[1],
                part->params.bandstop.slope);
        break;
    case VCU_EQPRG_TYPE_ALLPASS:
        sprintf(str, "allpass");
        break;
    case VCU_EQPRG_TYPE_CUSTOM:
        fprgs = vcu.eq.get_programs(&n_programs);
        for (i = 0; i < n_programs; i++) {
            if (memcmp(&fprgs[i].id, &part->params.custom.id,
                       sizeof(prgid_t)) == 0)
            {
                break;
            }
        }
        if (i == n_programs) {
            sprintf(str, "custom (NO PROGRAM)");
        } else {
            sprintf(str, "custom (%s)", fprgs[i].longname);
        }
        break;
    case VCU_EQPRG_TYPE_RANDOM_PHASE:
        sprintf(str, "random phase %.0fms",
                1000.0 * part->params.random_phase.length);
        break;
    case VCU_EQPRG_TYPE_FFT_STATIC:
        sprintf(str, "%d band fft eq", part->params.fft.n_bands);
        break;
    case VCU_EQPRG_TYPE_FFT_DYNAMIC:
        sprintf(str, "%d band dynamic eq",
                part->params.fft.n_bands);
        break;
    default:
        elog(LOGERROR, "unknown type: %d", part->type);
        exit(EXIT_FAILURE);
    }
    return str;
}

static void
state_eq_reorder_stack(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_eqprg *eqprg;
    static int parts[VCU_EQ_COMPONENTS_MAX];
    static int menupos = 0;
    static bool_t move_up;

    struct vcu_eqpart *part;
    bool_t do_refresh = false;
    void **entries;
    int pos, n;
    
    switch (action) {
    case ACTION_REFRESH:
	textmenu_state_delete(mstate);
        do_refresh = true;
        /* fall through */
    case ACTION_INIT: 
        mstate = NULL;
        eqprg = eq_setup_state.prg;
        if (eqprg->part_count < 2) {
            textmenu_message("There are less than ",
                             "two components      ");
            return;
        }        
        entries = alloca((eqprg->part_count + 20) * sizeof(void *));
	pos = 0;
        entries[pos++] = "Move [down/up]";
        entries[pos++] = &move_up;
        for (n = 0; n < eqprg->part_count; n++) {
            if (do_refresh) {
                part = &eqprg->parts[parts[n]];
            } else {
                move_up = true;
                parts[n] = n;
                part = &eqprg->parts[n];
            }
            entries[pos] = alloca(200);
            sprintf((char *)entries[pos++], "Move %s?", eq_part_str(part));
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    if (mstate == NULL) {
        textgui_pop_state();
        menupos = 0;
        return;
    }
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 1) {
            return;
        }        
        eq_setup_state.did_something = true;
        action--;
        if (move_up) {
            if (action == 0) {
                return;
            }
            n = parts[action - 1];
            parts[action - 1] = parts[action];
            parts[action] = n;
        } else {
            if (action == eqprg->part_count - 1) {
                return;
            }
            n = parts[action + 1];
            parts[action + 1] = parts[action];
            parts[action] = n;
        }
        menupos = mstate->pos;
        state_eq_reorder_stack(ACTION_REFRESH);
        return;
    case ACTION_EXIT:
        part = alloca(eqprg->part_count * sizeof(struct vcu_eqpart));
        for (n = 0; n < eqprg->part_count; n++) {
            part[n] = eqprg->parts[parts[n]];
        }
        memcpy(eqprg->parts, part,
               eqprg->part_count * sizeof(struct vcu_eqpart));
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_edit_eq(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_eqprg eqprg;
    static struct vcu_eqpart parts[VCU_EQ_COMPONENTS_MAX];
    static int menupos = 0;
    static int add_filter_pos, add_op_pos, delete_eq_pos, reorder_stack_pos;
    static bool_t refresh_equaliser = false;

    int n, err;

    switch (action) {
    case ACTION_INIT: {
        const struct vcu_eqprg *eqprg_ptr;
        void (*function)(int);
	void **entries;
	int pos;

	mstate = NULL;
        add_filter_pos = -1;
        add_op_pos = -1;
        reorder_stack_pos = -1;
	eq_setup_state.delete = false;
        eq_setup_state.modify = true;
        eq_setup_state.did_something = false;
        eq_setup_state.prg = &eqprg;
        if (!refresh_equaliser) {
            eqprg_ptr = vcu.eq.get_equaliser(eq_setup_state.current_eq_index);
            eqprg = *eqprg_ptr;
            eqprg.parts = parts;
            memcpy(parts, eqprg_ptr->parts,
                   eqprg_ptr->part_count * sizeof(struct vcu_eqpart));
        }
        refresh_equaliser = false;
	entries = alloca((6 * eqprg.part_count + 5) * sizeof(void *));
	pos = 0;
        function = NULL;
	for (n = 0; n < eqprg.part_count; n++) {
	    switch (eqprg.parts[n].type) {
            case VCU_EQPRG_TYPE_OP_CONVOLVE:
            case VCU_EQPRG_TYPE_OP_ADD:
            case VCU_EQPRG_TYPE_OP_SUBTRACT:
                function = state_eq_del_op;
                break;
            case VCU_EQPRG_TYPE_OP_MULTIPLY:
                function = state_eq_multiply;
                break;
	    case VCU_EQPRG_TYPE_LOWPASS:
                function = state_eq_lowpass;
		break;
	    case VCU_EQPRG_TYPE_HIGHPASS:
                function = state_eq_highpass;
		break;
	    case VCU_EQPRG_TYPE_BANDPASS:
                function = state_eq_bandpass;
		break;
	    case VCU_EQPRG_TYPE_BANDSTOP:
                function = state_eq_bandstop;
		break;
	    case VCU_EQPRG_TYPE_ALLPASS:
                function = state_eq_allpass;
		break;
	    case VCU_EQPRG_TYPE_CUSTOM:
                function = state_eq_custom;
		break;
	    case VCU_EQPRG_TYPE_RANDOM_PHASE:
                function = state_eq_random_phase;
		break;
            case VCU_EQPRG_TYPE_FFT_STATIC:
            case VCU_EQPRG_TYPE_FFT_DYNAMIC:
                function = state_eq_adjust_fft;
                break;
	    default:
		elog(LOGERROR, "unknown type: %d", eqprg.parts[n].type);
		exit(EXIT_FAILURE);
	    }
	    entries[pos] = alloca(200);
	    sprintf((char *)entries[pos++], "%s...",
                    eq_part_str(&eqprg.parts[n]));
            entries[pos++] = function;
	}
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Name [%%s,%d]",
                VCU_EQPRG_NAME_SIZE - 1);
        entries[pos++] = eqprg.name;
        if (eqprg.parts[0].type != VCU_EQPRG_TYPE_FFT_DYNAMIC) {
            add_filter_pos = pos / 2;
            entries[pos++] = "Add a filter...";
            entries[pos++] = state_eq_add_filter;
            add_op_pos = pos / 2;
            entries[pos++] = "Add an op...";
            entries[pos++] = state_eq_add_op;
            reorder_stack_pos = pos / 2;
            entries[pos++] = "Reorder stack...";
            entries[pos++] = state_eq_reorder_stack;
        }
        delete_eq_pos = pos / 2;
        entries[pos++] = "Delete this eq?ask";
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	mstate->pos = menupos;
        break;
    }
    case ACTION_CHILD_POP:
	if (!eq_setup_state.did_something) {
            break;
        }
        eq_setup_state.did_something = false;
        
        if (mstate->pos == add_filter_pos ||
            mstate->pos == add_op_pos)
        {
            elog(LOGDEBUG, "add part");
            eqprg.parts[eqprg.part_count] = eq_setup_state.part;
            eqprg.part_count++;
            menupos = mstate->pos + 1;
        } else if (eq_setup_state.delete) {
            elog(LOGDEBUG, "delete part");
            eq_setup_state.delete = false;
            
            if (eqprg.part_count == 1) {
                /* the whole equaliser is deleted */
                eq_setup_state.delete = true;
                state_edit_eq(ACTION_EXIT);
                return;
            }
            eqprg.part_count--;
	    for (n = mstate->pos; n < eqprg.part_count; n++) {
                eqprg.parts[n] = eqprg.parts[n + 1];
	    }
            menupos = mstate->pos;
            if (menupos >= eqprg.part_count) {
                menupos--;
            }
        } else if (mstate->pos == reorder_stack_pos) {
            elog(LOGDEBUG, "reorder parts");
            menupos = mstate->pos;
        } else {
            elog(LOGDEBUG, "modify part");
            eqprg.parts[mstate->pos] = eq_setup_state.part;
            menupos = mstate->pos;
        }
        refresh_equaliser = true;
        textmenu_state_delete(mstate);
        state_edit_eq(ACTION_INIT);
        return;
    }

    if (mstate->pos < eqprg.part_count) {
        eq_setup_state.part = eqprg.parts[mstate->pos];
        eq_setup_state.modify = true;
    }
    if (mstate->pos == add_filter_pos || mstate->pos == add_op_pos) {
        eq_setup_state.modify = false;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action == delete_eq_pos) {
            eq_setup_state.delete = true;
            state_edit_eq(ACTION_EXIT);
            return;
        }
        return;
    case ACTION_EXIT:
        if (!eq_setup_state.delete) {
            err = vcu.eq.modify_equaliser(eq_setup_state.current_eq_index,
                                          &eqprg);
            if (err != 0) {
		elog(LOGDEBUG, "Modify eq failed: %s", vcu.ctl.strerror(err));
		textmenu_message("Invalid equaliser   ",
				 "stack               ");
		return;
            }
        }
        eq_setup_state.modify = false;
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }        
}

static char *
ampstring(struct vcu_eqamp *amp)
{
    static char s[100];

    if (eq_properties_state.mean_mode) {
        if (amp->mean > 99.95) {
            sprintf(s, "+lim       ");
        } else if (amp->mean < -99.95) {
            sprintf(s, "-lim       ");
        } else {
            sprintf(s, "%+.1f      ", amp->mean);
        }
        
        if (amp->min > 99.95) {
            sprintf(&s[5], "+lim       ");
        } else if (amp->min < -99.95) {
            sprintf(&s[5], "-lim       ");
        } else {
            sprintf(&s[5], "%+.1f      ", amp->min);
        }
        
        if (amp->max > 99.95) {
            sprintf(&s[10], "+lim");
        } else if (amp->max < -99.95) {
            sprintf(&s[10], "-lim");
        } else {
            sprintf(&s[10], "%+.1f", amp->max);
        }
    } else {
        if (amp->in > 99.95) {
            sprintf(s, "+lim       ");
        } else if (amp->in < -99.95) {
            sprintf(s, "-lim       ");
        } else {
            sprintf(s, "%+.1f      ", amp->in);
        }
        
        if (amp->thru > 99.95) {
            sprintf(&s[5], "+lim       ");
        } else if (amp->thru < -99.95) {
            sprintf(&s[5], "-lim       ");
        } else {
            sprintf(&s[5], "%+.1f      ", amp->thru);
        }
        
        if (amp->out > 99.95) {
            sprintf(&s[10], "+lim");
        } else if (amp->out < -99.95) {
            sprintf(&s[10], "-lim");
        } else {
            sprintf(&s[10], "%+.1f", amp->out);
        }
    }
    return s;
}

static void
state_eq_properties_bands(int action)
{
    static char strings[VCU_FREQ_ISO_3OCT_BANDS+10][40];
    static char *lines[VCU_FREQ_ISO_3OCT_BANDS+10];
    static int pos, n_lines;
    
    if (action == ACTION_INIT) {
        struct vcu_eqamp *amp;
        int n;

        amp = eq_properties_state.amp;
        n = 0;
        sprintf(strings[n++], "Eq \"%s\":", eq_properties_state.name);
        sprintf(strings[n++], " delay: %.1fms",
                eq_properties_state.delay * 1000.0);
        sprintf(strings[n++], " peak amp: %+.1fdB",
                eq_properties_state.peak_amp);
        sprintf(strings[n++], " ISO 1/3 oct bands:");
        if (eq_properties_state.mean_mode) {
            sprintf(strings[n++], "band mean min  max");
        } else {
            sprintf(strings[n++], "band in   at   out");
        }
        sprintf(strings[n++], " 20 %s", ampstring(&amp[0]));
        sprintf(strings[n++], " 25 %s", ampstring(&amp[1]));
        sprintf(strings[n++], " 31 %s", ampstring(&amp[2]));
        sprintf(strings[n++], " 40 %s", ampstring(&amp[3]));
        sprintf(strings[n++], " 50 %s", ampstring(&amp[4]));
        sprintf(strings[n++], " 63 %s", ampstring(&amp[5]));
        sprintf(strings[n++], " 80 %s", ampstring(&amp[6]));
        sprintf(strings[n++], "100 %s", ampstring(&amp[7]));
        sprintf(strings[n++], "125 %s", ampstring(&amp[8]));
        sprintf(strings[n++], "160 %s", ampstring(&amp[9]));
        sprintf(strings[n++], "200 %s", ampstring(&amp[10]));
        sprintf(strings[n++], "250 %s", ampstring(&amp[11]));
        sprintf(strings[n++], "315 %s", ampstring(&amp[12]));
        sprintf(strings[n++], "400 %s", ampstring(&amp[13]));
        sprintf(strings[n++], "500 %s", ampstring(&amp[14]));
        sprintf(strings[n++], "630 %s", ampstring(&amp[15]));
        sprintf(strings[n++], "800 %s", ampstring(&amp[16]));
        sprintf(strings[n++], "1000%s", ampstring(&amp[17]));
        sprintf(strings[n++], "1250%s", ampstring(&amp[18]));
        sprintf(strings[n++], "1600%s", ampstring(&amp[19]));
        sprintf(strings[n++], "2000%s", ampstring(&amp[20]));
        sprintf(strings[n++], "2500%s", ampstring(&amp[21]));
        sprintf(strings[n++], "3150%s", ampstring(&amp[22]));
        sprintf(strings[n++], "4000%s", ampstring(&amp[23]));
        sprintf(strings[n++], "5000%s", ampstring(&amp[24]));
        sprintf(strings[n++], "6300%s", ampstring(&amp[25]));
        sprintf(strings[n++], "8000%s", ampstring(&amp[26]));
        sprintf(strings[n++], "10k %s", ampstring(&amp[27]));
        sprintf(strings[n++], "12.5%s", ampstring(&amp[28]));
        sprintf(strings[n++], "16k %s", ampstring(&amp[29]));
        sprintf(strings[n++], "20k %s", ampstring(&amp[30]));
	n_lines = n;
	for (n = 0; n < n_lines; n++) {
	    lines[n] = strings[n];
	}
	lines[n] = NULL;
    }
    
    if (n_lines == -1) {
	textgui_pop_state();
        return;
    }

    textmenu_showpage(&action, &pos, (const char **)lines, n_lines);

    if (action == ACTION_EXIT) {
	textgui_pop_state();
    }
}

static void
state_eq_properties(int action)
{
    static struct textmenu_state *mstate;

    switch (action) {
    case ACTION_INIT: {
        char heading[40 + VCU_EQPRG_NAME_SIZE];
        const struct vcu_eqprg *eqprg;
	void *entries[30];
        int err, pos;

        mstate = NULL;
        eqprg = vcu.eq.get_equaliser(eq_setup_state.current_eq_index);
        tw_clear();
        tw_putstring(0, 0, "Getting equaliser   ");
        tw_putstring(0, 1, "properties...       ");
        tw_refresh();
        strcpy(eq_properties_state.name, eqprg->name);
        err = vcu.eq.get_equaliser_properties(eq_setup_state.current_eq_index,
                                              eq_properties_state.amp,
                                              &eq_properties_state.peak_amp,
                                              &eq_properties_state.delay);
        switch (err) {
        case 0:
            break;
        default:
            textmenu_message("Could not get prop- ",
                             "erties for eq       ");
            return;
        }        
        pos = 0;
        entries[pos++] = "Mean/min/max...";
        entries[pos++] = state_eq_properties_bands;
        entries[pos++] = "In/thru/out...";
        entries[pos++] = state_eq_properties_bands;
        entries[pos++] = NULL;
        sprintf(heading, "Eq \"%s\":", eqprg->name);
        mstate = textmenu_state_new(heading, entries);
        break;
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
        return;
    }

    eq_properties_state.mean_mode = (mstate->pos == 0);
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_eq_copy(int action)
{
    static struct textmenu_state *mstate;
    const struct vcu_eqprg *eqprg;
    int err;
    
    /* pre textmenu_action action handling */
    switch (action) {
    case ACTION_INIT: {
	int pos, n, eq_count;
	void **entries;
	char *str;

	eq_count = vcu.eq.get_equaliser_count();
	entries = alloca((5 * eq_count + 3) * sizeof(void *));
	pos = 0;
	for (n = 0; n < eq_count; n++) {
            eqprg = vcu.eq.get_equaliser(n);
	    str = alloca(20 + VCU_EQPRG_NAME_SIZE);
	    sprintf(str, "Copy \"%s\"?", eqprg->name);
	    entries[pos++] = str;
	}
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }

    textmenu_action(mstate, &action);

    /* post textmenu_action action handling */
    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        eqprg = vcu.eq.get_equaliser(action);
        if ((err = vcu.eq.new_equaliser(eqprg)) < 0) {
            elog(LOGDEBUG, "Failed to create new equaliser: %s",
                 vcu.ctl.strerror(err));
            /* assume full table */
            textmenu_message("Equaliser table is  ",
                             "full                ");
            return;
        }
        /* fall through */
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_eq_design(int action)
{
    static struct textmenu_state *mstate;
    static int menupos = 0, eq_count;
    static bool_t new_eq, copy_eq;
    static bool_t ignore_childpop = false;

    struct vcu_eqprg eqprg;
    int n, err;
    
    /* pre textmenu_action action handling */
    switch (action) {
    case ACTION_INIT: {
        const struct vcu_eqprg *_eqprg;
	void **entries;
	char *str;
	int pos;

	new_eq = false;
        copy_eq = false;
	eq_setup_state.current_eq_index = -1;
        eq_setup_state.modify = false;
	eq_setup_state.delete = false;
	eq_setup_state.did_something = false;
	eq_count = vcu.eq.get_equaliser_count();
	entries = alloca((5 * eq_count + 3) * sizeof(void *));
	pos = 0;
	for (n = 0; n < eq_count; n++) {
            _eqprg = vcu.eq.get_equaliser(n);
	    str = alloca(20 + VCU_EQPRG_NAME_SIZE);
	    sprintf(str, "Edit \"%s\"...", _eqprg->name);
	    entries[pos++] = str;
	    entries[pos++] = state_edit_eq;
	    entries[pos++] = " properties...";
	    entries[pos++] = state_eq_properties;
	}
	entries[pos++] = "New equaliser...";
	entries[pos++] = state_eq_add_filter;
        if (eq_count > 0) {
            entries[pos++] = "Copy equaliser...";
            entries[pos++] = state_eq_copy;
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	mstate->pos = menupos;	
	break;
    }
    default:
	break;
    }

    eq_setup_state.current_eq_index = mstate->pos / 2;
    new_eq = false;
    copy_eq = false;

    if (mstate->pos == 2 * eq_count) {
        new_eq = true;
    }
    if (mstate->pos == 2 * eq_count + 1) {
        copy_eq = true;
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    /* post textmenu_action action handling */
    switch (action) {
    default:
	return;
    case ACTION_EXIT:
        vcu.ctl.update_hard_settings();
	textmenu_state_delete(mstate);
	textgui_pop_state();
	menupos = 0;
	return;
    case ACTION_CHILD_POP: {
	elog(LOGDEBUG, "component %d %d %d %d",
	     eq_setup_state.did_something,
	     eq_setup_state.current_eq_index,
	     ignore_childpop,
	     new_eq);

	if (ignore_childpop) {
	    ignore_childpop = false;
	    return;
	}
	if (eq_setup_state.delete) {
            elog(LOGDEBUG, "Delete equaliser");
            /* Delete equaliser */	
            eq_setup_state.delete = false;
            vcu.eq.delete_equaliser(eq_setup_state.current_eq_index);
            eq_count--;
            if (menupos >= 2 * eq_count) {
                menupos = 2 * (eq_count - 1);
            }
            if (menupos < 0) {
                menupos = 0;
            }
            textmenu_state_delete(mstate);
            state_eq_design(ACTION_INIT);
            return;
        }
	if (new_eq) {
	    if (!eq_setup_state.did_something) {
		return;
	    }
            elog(LOGDEBUG, "New equaliser");
	    eq_setup_state.did_something = false;
	    /* Create equaliser */
            memset(&eqprg, 0, sizeof(eqprg));
            eqprg.parts = alloca(sizeof(struct vcu_eqpart));
            eqprg.parts[0] = eq_setup_state.part;
            eqprg.part_count = 1;
            sprintf(eqprg.name, "eq%d", eq_setup_state.current_eq_index + 1);
	    if ((err = vcu.eq.new_equaliser(&eqprg)) < 0) {
		elog(LOGDEBUG, "Failed to create new equaliser: %s",
		     vcu.ctl.strerror(err));
		/* assume full table */
		textmenu_message("Equaliser table is  ",
				 "full                ");
		ignore_childpop = true;
		return;
	    }
            elog(LOGDEBUG, "added new equaliser \"%s\"", eqprg.name);
	    menupos = 2 * eq_count;
	    eq_count++;
	    textmenu_state_delete(mstate);
	    state_eq_design(ACTION_INIT);
	    return;
	}
        if (copy_eq) {
	    menupos = 2 * eq_count + 1;
            eq_count++;
            textmenu_state_delete(mstate);
            state_eq_design(ACTION_INIT);
            return;
        }
	return;
    }
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_inout_eq_mode_channel_edit(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_eqmode mode;
    static int channel_count[VCU_CH_WP_COUNT];
    int type, n;

    switch (action) {
    case ACTION_INIT: {
        char name[VCU_CH_COUNT + 1 + 10], *eqstr, *p, entrylen;
        const struct vcu_eqprg *eqprg;
	void **entries;
	int pos, io;

        mstate = NULL;
        io = inout_generic_state.io;

        eqstr = alloca((VCU_EQPRG_NAME_SIZE + 10) *
                       vcu.eq.get_equaliser_count());
        p = eqstr;
        sprintf(p, "off/");
        p += strlen(p);
        for (n = 0; n < vcu.eq.get_equaliser_count(); n++, p += strlen(p)) {
            eqprg = vcu.eq.get_equaliser(n);
            sprintf(p, "%s/", eqprg->name);
        }
        p[-1] = '\0';
        entrylen = strlen(eqstr) + 20;
        
        memcpy(&mode, vcu.eq.get_mode(current_eq_mode), sizeof(mode));
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            channel_count[type] = vcu.channel.get_count(io, type);
        }

        pos = 0;
        entries = alloca((20 + VCU_CH_WP_COUNT * VCU_MAX_CHANNELS) *
                         sizeof(void *));
        
        if (io == VCU_IN && inout_generic_state.type != VCU_CH_PHYSICAL) {
            /* ambiopole */
            for (n = 0; n < channel_count[VCU_CH_AMBIOPOLE]; n++) {
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iA%d [%s]", n + 1, eqstr);
                mode.prgs[io][VCU_CH_AMBIOPOLE][n]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_AMBIOPOLE][n];
            }

            /* reverb */
            switch (channel_count[VCU_CH_REVERB]) {
            case 1:
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRc [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][0]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][0];
                break;
            case 2:
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRl [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][0]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][0];
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRr [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][1]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][1];
                break;
            case 3:
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRl [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][0]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][0];
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRr [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][1]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][1];
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iRc [%s]", eqstr);
                mode.prgs[io][VCU_CH_REVERB][2]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_REVERB][2];
                break;
            }
            
            /* ambisonic */
            if (channel_count[VCU_CH_AMBISONIC] > 0) {
                switch (vcu.ambisonics.get_input_format()) {
                case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:                
                case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                    strcpy(name, "wxyzrstuv");
                    for (n = 0; n < channel_count[VCU_CH_AMBISONIC]; n++) {
                        entries[pos] = alloca(entrylen);
                        sprintf((char *)entries[pos++],
                                "iB%c [%s]", name[n], eqstr);
                        mode.prgs[io][VCU_CH_AMBISONIC][n]++;
                        entries[pos++] =
                            &mode.prgs[io][VCU_CH_AMBISONIC][n];
                    }
                    break;
                case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                    entries[pos] = alloca(entrylen);
                    sprintf((char *)entries[pos++], "iBl [%s]", eqstr);
                    mode.prgs[io][VCU_CH_AMBISONIC][0]++;
                    entries[pos++] = &mode.prgs[io][VCU_CH_AMBISONIC][0];
                    entries[pos] = alloca(entrylen);
                    sprintf((char *)entries[pos++], "iBr [%s]", eqstr);
                    mode.prgs[io][VCU_CH_AMBISONIC][1]++;
                    entries[pos++] = &mode.prgs[io][VCU_CH_AMBISONIC][1];
                    break;
                }
            }
            
            /* passthru */
            for (n = 0; n < channel_count[VCU_CH_PASSTHRU]; n++) {
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "iP%d [%s]", n + 1, eqstr);
                mode.prgs[io][VCU_CH_PASSTHRU][n]++;
                entries[pos++] = &mode.prgs[io][VCU_CH_PASSTHRU][n];
            }
        }
        
        if (io == OUT && inout_generic_state.type != VCU_CH_PHYSICAL) {

            strcpy(name, "ARBP");
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (n = 0; n < channel_count[type]; n++) {
                    entries[pos] = alloca(entrylen);
                    sprintf((char *)entries[pos++],
                            "o%c%d [%s]", name[type], n + 1, eqstr);
                    mode.prgs[io][type][n]++;
                    entries[pos++] = &mode.prgs[io][type][n];
                }
            }
        }

        if (inout_generic_state.type == VCU_CH_PHYSICAL) {
            type = VCU_CH_PHYSICAL;
            for (n = 0; n < channel_count[type]; n++) {
                entries[pos] = alloca(entrylen);
                sprintf((char *)entries[pos++], "%c%d [%s]",
                        io == IN ? 'i' : 'o', n + 1, eqstr);
                mode.prgs[io][type][n]++;
                entries[pos++] = &mode.prgs[io][type][n];
            }
        }
        
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
    }
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        if (inout_generic_state.type == VCU_CH_PHYSICAL) {
            for (n = 0; n < channel_count[VCU_CH_PHYSICAL]; n++) {
                mode.prgs[inout_generic_state.io][VCU_CH_PHYSICAL][n]--;
            }
        } else {
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (n = 0; n < channel_count[type]; n++) {
                    mode.prgs[inout_generic_state.io][type][n]--;
                }
            }
        }
        vcu.eq.modify_mode(current_eq_mode, &mode);
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_eq_edit_mode_logiphys(int action)
{
    struct vcu_eqmode mode;
    int type, n;
    
    if (action == ACTION_INIT || action == ACTION_CHILD_POP) {
        memcpy(&mode, vcu.eq.get_mode(current_eq_mode), sizeof(mode));
        inout_generic_state.logical_count = 0;
        inout_generic_state.physical_count = 0;
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            for (n = 0; n < vcu.channel.get_count(inout_generic_state.io,
                                                  type); n++)
            {
                if (mode.prgs[inout_generic_state.io][type][n] >= 0) {
                    if (type == VCU_CH_PHYSICAL) {
                        inout_generic_state.physical_count++;
                    } else {
                        inout_generic_state.logical_count++;
                    }
                }
            }
        }
    }
    state_inout_generic(action);
}

static void
state_eq_edit_mode(int action)
{
    static struct textmenu_state *mstate;
    static struct vcu_eqmode mode;
    static int delete_mode_pos;
    static bool_t created_new;
    static int count[2][2];
    static int menupos = 0;

    switch (action) {
    case ACTION_CHILD_POP:
        if (mstate == NULL) {
            break;
        }
        textmenu_state_delete(mstate);
        /* fall through */
    case ACTION_INIT: {
	int pos, type, io, n;
	void *entries[20];

	delete_mode_pos = -1;
        mstate = NULL;
	if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
	if (vcu.eq.get_equaliser_count() == 0) {
	    textmenu_message("There are no equali-",
			     "sers                ");
	    return;
	}
        if (action != ACTION_CHILD_POP) {
            created_new = false;
        }
        if (current_eq_mode < 0) {
            memset(&mode, 0, sizeof(mode));
            strcpy(mode.name, "unnamed");
            memset(&mode.prgs, -1, sizeof(mode.prgs));
            if (vcu.eq.new_mode(&mode) != 0) {
                textmenu_message("Cannot create more  ",
                                 "eq modes            ");
                return;
            }
            current_eq_mode = vcu.eq.get_mode_count() - 1;
            created_new = true;
        }
        memcpy(&mode, vcu.eq.get_mode(current_eq_mode), sizeof(mode));
        pos = 0;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Name [%%s,%d]",
                VCU_EQMODE_NAME_SIZE - 1);
        entries[pos++] = mode.name;
        count[VCU_IN][0] = count[VCU_OUT][0] = 0;
        count[VCU_IN][1] = count[VCU_OUT][1] = 0;
        for (io = VCU_IN; io <= VCU_OUT; io++) {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                for (n = 0; n < vcu.channel.get_count(io, type); n++) {
                    if (mode.prgs[io][type][n] >= 0) {
                        if (type == VCU_CH_PHYSICAL) {
                            count[io][1]++;
                        } else {
                            count[io][0]++;
                        }
                    }
                }
            }
        }
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Inputs (%d)...",
                count[VCU_IN][0] + count[VCU_IN][1]);
        entries[pos++] = state_eq_edit_mode_logiphys;
        entries[pos] = alloca(40);
        sprintf((char *)entries[pos++], "Outputs (%d)...",
                count[VCU_OUT][0] + count[VCU_OUT][1]);
        entries[pos++] = state_eq_edit_mode_logiphys;
        if (created_new) {
            entries[pos++] = "Create this mode?";
        } else {
            entries[pos++] = "Delete this mode?ask";
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        inout_generic_state.io = VCU_IN;
        strcpy(inout_generic_state.heading, "Assign equalisers");
        inout_generic_state.state = state_inout_eq_mode_channel_edit;
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }

    switch (mstate->pos) {
    case 1:
        inout_generic_state.io = VCU_IN;
        break;
    case 2:
        inout_generic_state.io = VCU_OUT;
        break;
    }
    
    textmenu_action(mstate, &action);
    menupos = mstate->pos;
    
    switch (action) {
    case 0:
        vcu.eq.modify_mode(current_eq_mode, &mode);
        break;
    case 3:
        menupos = 0;
        if (created_new) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
        elog(LOGDEBUG, "delete eq mode %d", current_eq_mode);
        vcu.eq.delete_mode(current_eq_mode);
        current_eq_mode = -1;
        state_eq_edit_mode(ACTION_EXIT);
        break;
    case ACTION_EXIT:
        if (created_new) {
            vcu.eq.delete_mode(current_eq_mode);
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

/* FIXME: very similar to reorder_groups function */
static void
state_eq_reorder_modes(int action)
{
    static struct textmenu_state *mstate;
    static int modes[VCU_MAX_EQ_MODES];
    static int menupos = 0;
    static bool_t move_up;

    const struct vcu_eqmode *mode;
    bool_t do_refresh = false;
    void **entries;
    int pos, n;
    
    switch (action) {
    case ACTION_REFRESH:
	textmenu_state_delete(mstate);
        do_refresh = true;
        /* fall through */
    case ACTION_INIT:
        mstate = NULL;
        if (vcu.eq.get_mode_count() < 2) {
            textmenu_message("There are less than ",
                             "two eq modes        ");
            return;
        }        
        entries = alloca((vcu.eq.get_mode_count() + 20) * sizeof(void *));
	pos = 0;
        entries[pos++] = "Move [down/up]";
        entries[pos++] = &move_up;
        for (n = 0; n < vcu.eq.get_mode_count(); n++) {
            if (do_refresh) {
                mode = vcu.eq.get_mode(modes[n]);
            } else {
                move_up = true;
                modes[n] = n;
                mode = vcu.eq.get_mode(n);
            }
            entries[pos] = alloca(VCU_EQMODE_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Move %s?", mode->name);
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
        break;
    }
    if (mstate == NULL) {
        textgui_pop_state();
        menupos = 0;
        return;
    }
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 1) {
            return;
        }        
        action--;
        if (move_up) {
            if (action == 0) {
                return;
            }
            n = modes[action - 1];
            modes[action - 1] = modes[action];
            modes[action] = n;
        } else {
            if (action == vcu.eq.get_mode_count() - 1) {
                return;
            }
            n = modes[action + 1];
            modes[action + 1] = modes[action];
            modes[action] = n;
        }
        menupos = mstate->pos;
        state_eq_reorder_modes(ACTION_REFRESH);
        return;
    case ACTION_EXIT:
        vcu.eq.reorder_modes(modes);
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_assign_eq_modes_sub(int action)
{
    static int menupos = 0, eq_mode_idx, idx2mode[VCU_MAX_EQ_MODES];
    static bool_t active[VCU_MAX_EQ_MODES];
    static struct textmenu_state *mstate;
    static struct vcu_chmode mode;
    
    switch (action) {
    case ACTION_INIT: {
        void *entries[2 * VCU_MAX_EQ_MODES + 10];
        const struct vcu_eqmode *eqmode;
        int n, i, pos;
        char *p;

        pos = 0;
        if (current_channel_mode < 0) {
            if (vcu.eq.get_mode_count() == 0) {
                entries[pos++] = "Equalisation off";
            } else if (vcu.eq.get_mode_count() == 1) {
                eq_mode_idx = vcu.runtime.eq.get_mode() + 1;
                entries[pos++] = "Equalisation [off/on]";
                entries[pos++] = &eq_mode_idx;
            } else {
                eq_mode_idx = vcu.runtime.eq.get_mode() + 1;
                entries[pos] = alloca(vcu.eq.get_mode_count() *
                                      (VCU_EQMODE_NAME_SIZE + 3) + 20);
                p = (char *)entries[pos++];
                
                strcpy(p, "Eq [off/");
                p += strlen(p);
                for (n = 0; n < vcu.eq.get_mode_count(); n++) {
                    eqmode = vcu.eq.get_mode(n);
                    sprintf(p, "%s/", eqmode->name);
                    p += strlen(p);
                }
                p[-1] = ']';
                entries[pos++] = &eq_mode_idx;
            }
            entries[pos++] = NULL;
            mstate = textmenu_state_new("ch modes off", entries);
            break;
        }
        
        memcpy(&mode, vcu.channel.get_mode(current_channel_mode), sizeof(mode));
        array_bit2bool(mode.eqmodes, active, VCU_MAX_EQ_MODES);

        entries[pos] = alloca(40 + vcu.eq.get_mode_count() *
                              VCU_EQMODE_NAME_SIZE);
        if (bit_find(mode.eqmodes, 0, VCU_MAX_EQ_MODES - 1) == -1) {
            entries[pos++] = "All eq modes off";
        } else {
            p = (char *)entries[pos++];
            strcpy(p, "Mode [off/");
            p = &p[10];
            eq_mode_idx = 0;
            idx2mode[0] = -1;
            for (n = 0, i = 1; n < vcu.eq.get_mode_count(); n++) {
                if (active[n]) {
                    idx2mode[i] = n;
                    if (mode.active_eq_mode == n) {
                        eq_mode_idx = i;
                    }
                    eqmode = vcu.eq.get_mode(n);
                    sprintf(p, "%s/", eqmode->name);
                    p += strlen(p);
                    i++;
                }
            }
            p[-1] = ']';
            entries[pos++] = &eq_mode_idx;
        }
        for (n = 0; n < vcu.eq.get_mode_count(); n++) {
            eqmode = vcu.eq.get_mode(n);
            entries[pos] = alloca(VCU_EQMODE_NAME_SIZE + 20);
            if (vcu.eq.mode_is_assignable(n, current_channel_mode)) {
                sprintf((char *)entries[pos++], "\"%s\" [off/on]",
                        eqmode->name);
                entries[pos++] = &active[n];
            } else {
                sprintf((char *)entries[pos++], "\"%s\" off", eqmode->name);
            }
        }
        entries[pos++] = NULL;
        p = alloca(strlen(mode.name) + 3);
        strcpy(&p[1], mode.name);
        p[0] = '"';
        strcat(p, "\"");
	mstate = textmenu_state_new(p, entries);
        mstate->pos = menupos;
        break;
    }
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);
    
    switch (action) {
    case 0:
        if (current_channel_mode < 0) {
            vcu.runtime.eq.set_mode(eq_mode_idx - 1);
            return;
        }
        mode.active_eq_mode = idx2mode[eq_mode_idx];
        vcu.channel.modify_mode(current_channel_mode, &mode);
        return;
    default:
        if (action < 0) {
            return;
        }
        array_bool2bit(active, mode.eqmodes, VCU_MAX_EQ_MODES);
        if (mode.active_eq_mode != -1 && !active[mode.active_eq_mode]) {
            mode.active_eq_mode = -1;
        }
        vcu.channel.modify_mode(current_channel_mode, &mode);
	textmenu_state_delete(mstate);
        state_assign_eq_modes_sub(ACTION_INIT);
        return;
    case ACTION_EXIT:
        menupos = 0;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_assign_eq_modes(int action)
{
    static struct textmenu_state *mstate;
    
    switch (action) {
    case ACTION_INIT: {
        void *entries[2 * VCU_MAX_EQ_MODES + 10];
        const struct vcu_chmode *chmode;
        int n, pos;

        pos = 0;
        entries[pos++] = "ch modes off...";
        entries[pos++] = state_assign_eq_modes_sub;
        for (n = 0; n < vcu.channel.get_mode_count(); n++) {
            chmode = vcu.channel.get_mode(n);
            entries[pos] = alloca(VCU_EQMODE_NAME_SIZE + 20);            
            sprintf((char *)entries[pos++], "\"%s\"...", chmode->name);
            entries[pos++] = state_assign_eq_modes_sub;
        }
        entries[pos++] = NULL;
            
	mstate = textmenu_state_new(NULL, entries);
        break;
    }
    }

    current_channel_mode = mstate->pos - 1;
    textmenu_action(mstate, &action);
    
    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }    
}

static void
state_eq_modes(int action)
{
    static struct textmenu_state *mstate;
    static int mode, mode_base, menupos = 0;

    switch (action) {
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        state_eq_modes(ACTION_INIT);
        return;
    case ACTION_INIT: {
        const struct vcu_eqmode *eqmode;
	void **entries;
	int pos, n;
        char *p;
	
        entries = alloca((vcu.eq.get_mode_count() + 20) * sizeof(void *));
	pos = 0;
        if (vcu.channel.get_mode_count() > 1) {
            entries[pos++] = "Assign eq modes...";
            entries[pos++] = state_assign_eq_modes;
            mode_base = 1;
        } else if (vcu.eq.get_mode_count() == 0) {
            entries[pos++] = "Equalisation off";
            mode_base = 1;
        } else if (vcu.eq.get_mode_count() == 1) {
            mode = vcu.runtime.eq.get_mode() + 1;
            entries[pos++] = "Equalisation [off/on]";
            entries[pos++] = &mode;
            mode_base = 1;
        } else {
            mode = vcu.runtime.eq.get_mode() + 1;
            entries[pos] = alloca(vcu.eq.get_mode_count() *
                                  (VCU_EQMODE_NAME_SIZE + 3) + 20);            
            p = (char *)entries[pos++];
            
            strcpy(p, "Eq [off/");
            p += strlen(p);
            for (n = 0; n < vcu.eq.get_mode_count(); n++) {
                eqmode = vcu.eq.get_mode(n);
                sprintf(p, "%s/", eqmode->name);
                p += strlen(p);
            }
            p[-1] = ']';
            entries[pos++] = &mode;
            mode_base = 1;
        }
        
        for (n = 0; n < vcu.eq.get_mode_count(); n++) {
            eqmode = vcu.eq.get_mode(n);
            entries[pos] = alloca(VCU_EQMODE_NAME_SIZE + 20);
            sprintf((char *)entries[pos++], "Edit \"%s\"...", eqmode->name);
            entries[pos++] = state_eq_edit_mode;
        }
        entries[pos++] = "New mode...";
        entries[pos++] = state_eq_edit_mode;
        if (vcu.eq.get_mode_count() > 1) {
            entries[pos++] = "Reorder modes...";
            entries[pos++] = state_eq_reorder_modes;
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }
    }

    if (mstate->pos >= mode_base) {
        if (mstate->pos == mode_base + vcu.eq.get_mode_count()) {
            current_eq_mode = -1;
        } else {
            current_eq_mode = mstate->pos - mode_base;
        }
    }
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        if (vcu.channel.get_mode_count() < 2) {
            vcu.runtime.eq.set_mode(mode - 1);
        }
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }    
}

static void
state_eq_setup(int action)
{
    static struct textmenu_state *mstate;

    current_program_type = VCU_PROGRAM_TYPE_EQ;

    switch (action) {
    case ACTION_INIT: {
        void *entries[100];
        int pos;

        pos = 0;
        entries[pos++] = "Eq design...";
        entries[pos++] = state_eq_design;
        entries[pos++] = "Eq modes...";
        entries[pos++] = state_eq_modes;
        entries[pos++] = "Load custom...";
        entries[pos++] = state_load_program;
        entries[pos++] = "Del custom...";
        entries[pos++] = state_del_program;
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }
    
    textmenu_action(mstate, &action);
    
    switch (action) {
    case ACTION_EXIT:
        elog(LOGDEBUG, "calling update_hard_settings");
        UPDATE_HARD_SETTINGS_CALL;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	break;
    case ACTION_CHILD_POP:
        UPDATE_HARD_SETTINGS_CHILD_POP;
        return;
    }    
}

static void
state_convolver_setup_benchmark(int action)
{
    static struct textmenu_state *mstate;
    static int benchmark_pos;
    static int dlp_choice[VCU_VALID_RATES_COUNT];
    static bool_t can_choose = false;
    static int benchmark_load;
    static double benchmark_length;
    static bool_t has_run_benchmark;
    static bool_t test_low_latency = false;
    static bool_t powersave;

    int n, i, err, n_rates, rates[VCU_VALID_RATES_COUNT], limit;
    bool_t b;

    /* pre textmenu_action action handling */
    switch (action) {
    case ACTION_INIT: {
	void *entries[30 + 2 * VCU_VALID_RATES_COUNT];	
	int pos, n_alts, largest;
	char str[VCU_VALID_RATES_COUNT]
	    [20 + VCU_BENCHMARK_MAX_ALTERNATIVES * 20];
	char tmpstr[100];
	double delay[VCU_BENCHMARK_MAX_ALTERNATIVES];
	double maxrevlen[VCU_BENCHMARK_MAX_ALTERNATIVES];
        double range[2];

        has_run_benchmark = false;
	
	pos = 0;
	can_choose = false;
        powersave = vcu.ctl.get_powersave();
	vcu.signal.get_rates(rates, &n_rates);
        benchmark_load = (int)rint(vcu.iodelay.get_benchmark_max_load() *
                                   100.0);
        entries[pos++] = "Max load [%d,1,10,1,100]%";
        entries[pos++] = &benchmark_load;
        if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
            benchmark_length = vcu.iodelay.get_benchmark_max_length();
            entries[pos] = alloca(100);
            vcu.iodelay.get_benchmark_max_length_range(range);
            sprintf((char *)entries[pos++], "Max length [%%.1f,0.1,10,%f,%f]s",
                    range[0], range[1]);
            entries[pos++] = &benchmark_length;
        }
        entries[pos++] = "Low latency [off/on]";
        entries[pos++] = &test_low_latency;
	if (vcu.iodelay.need_to_run_benchmark() ||
	    !vcu.iodelay.has_run_benchmark() ||
            vcu.iodelay.manually_set())
	{
	    benchmark_pos = pos / 2;
	    entries[pos++] = "Run benchmark?ask";
	    if (!vcu.iodelay.need_to_run_benchmark() &&
                !vcu.iodelay.manually_set())
            {
		for (n = 0; n < n_rates; n++) {
		    vcu.iodelay.get_delay_length_pair(rates[n], delay,
                                                      maxrevlen, NULL);
		    if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
			sprintf(str[n], "%.1f %s",
                                (double)rates[n] / 1000,
				make_delaylenstr(tmpstr, maxrevlen[0],
                                                 delay[0]));
		    } else {
			sprintf(str[n], "%.1f %s", (double)rates[n] / 1000,
				make_delaylenstr(tmpstr, 0, delay[0]));
		    }
		    entries[pos++] = str[n];
		}
	    }
	} else {
	    for (n = 0; n < n_rates; n++) {
		vcu.iodelay.get_dlp_alternatives(rates[n], delay,
                                                 maxrevlen, &n_alts);
		sprintf(str[n], "%.1f [", (double)rates[n] / 1000);
		if (n_alts == 1) {
		    if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
			sprintf(str[n], "%.1f %s",
                                (double)rates[n] / 1000,
				make_delaylenstr(tmpstr, maxrevlen[0],
                                                 delay[0]));
		    } else {
			sprintf(str[n], "%.1f %s", (double)rates[n] / 1000,
				make_delaylenstr(tmpstr, 0, delay[0]));
		    }
		    entries[pos++] = str[n];
		} else {
		    largest = 0;
		    for (i = 0; i < n_alts; i++) {
			if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
                            make_delaylenstr(tmpstr, maxrevlen[i], delay[i]);
			} else {
                            make_delaylenstr(tmpstr, 0, delay[i]);
			}
			if (maxrevlen[i] > maxrevlen[largest]) {
			    largest = i;
			}
			strcat(str[n], tmpstr);
			if (i < n_alts - 1) {
			    strcat(str[n], "/");
			}
		    }
                    strcat(str[n], "]");
		    entries[pos++] = str[n];
		    vcu.iodelay.get_delay_length_pair(rates[n], NULL, NULL,
                                                      &dlp_choice[n]);
                    elog(LOGDEBUG, "dlp_choice[%d]: %d", n, dlp_choice[n]);
		    if (dlp_choice[n] == -1) {
			dlp_choice[n] = largest;
		    }
		    entries[pos++] = &dlp_choice[n];
		    can_choose = true;
		}
	    }
            if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
                benchmark_pos = n_rates + 3;
            } else {
                benchmark_pos = n_rates + 2;
            }
	    entries[pos++] = "Rerun benchmark?ask";
	}
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }
    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    /* post textmenu_action action handling */
    switch (action) {
    default:
	if (action != benchmark_pos) {
	    return;	
	}
	tw_clear();
        vcu.iodelay.set_benchmark_max_load((double)benchmark_load / 100.0);
        if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
            vcu.iodelay.set_benchmark_max_length(benchmark_length);
        }
	if (tw_ysize() < 4) {
	    tw_putstring(0, 0, "Benchmarking is now ");
	    tw_putstring(0, 1, "starting up...      ");
	} else {
	    tw_putstring(0, 0, "Benchmarking may    ");
	    tw_putstring(0, 1, "take a few minutes. ");
	    
	    tw_putstring(0, 3, "   Starting up...   ");
	}
	tw_refresh();
        has_run_benchmark = true;
	if ((err = vcu.iodelay.run_benchmark(test_low_latency,
                                             &limit)) != 0) {
	    tw_clear();
	    switch (err) {
	    case VCU_ERR_NOT_SUPPORTED_BY_HARDWARE:
		textmenu_message("Current settings are",
				 "too demanding       ");
		break;
	    default:
		textmenu_message("Benchmark run failed",
				 "system may be broken");
		break;
	    }
	    return;
	}
	tw_clear();
        switch (limit) {
        case VCU_BENCHMARK_LIMIT_CPU:
            textmenu_message("Benchmark complete  ",
                             "(CPU limited)       ");
            break;
        case VCU_BENCHMARK_LIMIT_MEMORY:
            textmenu_message("Benchmark complete  ",
                             "(Memory limited)    ");
            break;
        case VCU_BENCHMARK_LIMIT_HARD:
            textmenu_message("Benchmark complete  ",
                             "(Hard limit)        ");
            break;
        }
	return;
    case ACTION_EXIT:
	if (can_choose) {
            vcu.signal.get_rates(rates, &n_rates);
            for (n = 0; n < n_rates; n++) {
		vcu.iodelay.set_delay_length_pair(rates[n], dlp_choice[n]);
            }
	}	
        elog(LOGDEBUG, "calling update_hard_settings");
        UPDATE_HARD_SETTINGS_CALL;
        vcu.iodelay.set_benchmark_max_load((double)benchmark_load / 100.0);
        if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
            vcu.iodelay.set_benchmark_max_length(benchmark_length);
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_CHILD_POP:
        UPDATE_HARD_SETTINGS_CHILD_POP;
	textmenu_state_delete(mstate);
        b = has_run_benchmark;
	state_convolver_setup_benchmark(ACTION_INIT);
        has_run_benchmark = b;
	return;
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_convolver_setup_manually(int action)
{
    static struct textmenu_state *mstate;
    static double reverb_length[VCU_VALID_RATES_COUNT];
    static double iodelay[VCU_VALID_RATES_COUNT];
    static int apply_pos;

    int n, n_rates, rates[VCU_VALID_RATES_COUNT];
    
    switch (action) {
    case ACTION_INIT: {
	void *entries[30 + 4 * VCU_VALID_RATES_COUNT];
        double range[2];
        int pos;

        pos = 0;
	vcu.signal.get_rates(rates, &n_rates);
        vcu.iodelay.get_benchmark_max_length_range(range);
        for (n = 0; n < n_rates; n++) {
            vcu.iodelay.get_delay_length_pair(rates[n], &iodelay[n],
                                              &reverb_length[n], NULL);
            entries[pos] = alloca(100);
            sprintf((char *)entries[pos++],
                    "%.0f delay [%%.0f,1,100,0,10000]ms", rates[n] / 1000.0);
            iodelay[n] *= 1000.0;
            entries[pos++] = &iodelay[n];
            if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
                entries[pos] = alloca(100);
                sprintf((char *)entries[pos++],
                        "%.0f reverb [%%.1f,0.1,10,%f,%f]s", rates[n] / 1000.0,
                        range[0], range[1]);
                entries[pos++] = &reverb_length[n];
            }
        }
        entries[pos++] = "Apply the above?";
	entries[pos++] = NULL;
        apply_pos = n_rates;
        if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0) {
            apply_pos *= 2;
        }
	mstate = textmenu_state_new(NULL, entries);
	break;
    }
    default:
	break;
    }
    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action != apply_pos) {
            return;
        }
	vcu.signal.get_rates(rates, &n_rates);
        for (n = 0; n < n_rates; n++) {
            vcu.iodelay.set_delay_length_pair_manual(rates[n],
                                                     iodelay[n] / 1000.0,
                                                     reverb_length[n]);
        }
	textmenu_state_delete(mstate);
        state_convolver_setup_manually(ACTION_INIT);
        return;
    case ACTION_EXIT:
        UPDATE_HARD_SETTINGS_CALL;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_CHILD_POP:
        UPDATE_HARD_SETTINGS_CHILD_POP;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_convolver_setup(int action)
{
    static struct textmenu_state *mstate;
    static bool_t powersave;
    static int menupos = 0;

    switch (action) {
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        /* fall through */
    case ACTION_INIT: {
	void *entries[30];
        int pos;

	/* sanity check */
	if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}

        powersave = vcu.ctl.get_powersave();
	
        pos = 0;
        entries[pos++] = "Power save [off/on]";
        entries[pos++] = &powersave;
        if (vcu.iodelay.manually_set()) {
            entries[pos++] = "Benchmark (off)...";
        } else {
            entries[pos++] = "Benchmark (on)...";
        }
        entries[pos++] = state_convolver_setup_benchmark;
        if (vcu.iodelay.manually_set()) {
            entries[pos++] = "Manual (on)...";
        } else {
            entries[pos++] = "Manual (off)...";
        }
        entries[pos++] = state_convolver_setup_manually;
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
	break;
    }
    default:
	break;
    }
    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
        vcu.ctl.set_powersave(powersave);
        return;
    case ACTION_EXIT:
        menupos = 0;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    default:
	return;
    }
}

static void
state_user_interface_setup(int action)
{
    static struct textmenu_state *mstate;
    static int dimmer, volume_step;

    void *entries[40 + VCU_MAX_CHANNEL_GROUPS];
    const struct vcu_chgrp *chgrp;
    int pos, n;
    
    if (action == ACTION_INIT) {
        pos = 0;        
        mstate = NULL;

        dimmer = conf_user.dimmer - 1;
        entries[pos++] = "Screen dimmer [1/2/3/4]";
        entries[pos++] = &dimmer;

        entries[pos++] = "Autostart [off/on]";
        entries[pos++] = &conf_user.autostart;

        entries[pos++] = "Volume step [0.1/0.5/1.0/3.0/6.0]";
	if (conf_user.volume_step < 0.5) {
	    conf_user.volume_step = 0.1;
	    volume_step = 0;
	} else if (conf_user.volume_step < 1.0) {
	    conf_user.volume_step = 0.5;
	    volume_step = 1;
	} else if (conf_user.volume_step < 3.0) {
	    conf_user.volume_step = 1.0;
	    volume_step = 2;
	} else if (conf_user.volume_step < 6.0) {
	    conf_user.volume_step = 3.0;
	    volume_step = 3;
	} else {
	    conf_user.volume_step = 6.0;
	    volume_step = 4;
	}
	entries[pos++] = &volume_step;

        if (vcu.runtime.channel.get_mode() >= 0) {
            entries[pos++] = "Ch mode [locked/sel on/sel off]";
            entries[pos++] = &conf_user.channel_mode_option;
        }
        if (vcu.eq.get_mode_count() > 0) {
            entries[pos++] = "Eq mode [locked/sel on/sel off]";
            entries[pos++] = &conf_user.eq_mode_option;
        }

        entries[pos++] = "Master vol [off/on]";
        entries[pos++] = &conf_user.vol_master;
        
        for (n = 0; n < vcu.channel.get_group_count(); n++) {
            chgrp = vcu.channel.get_group(n);
            entries[pos] = alloca(40 + VCU_CHGRP_NAME_SIZE);
            sprintf((char *)entries[pos], "%s vol [off/on]", chgrp->name);
            ((char *)entries[pos])[0] = toupper(((char *)entries[pos])[0]);
            pos++;
            entries[pos++] = &conf_user.vol_groups[n];
        }
        entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        vcu.ctl.schedule_save_config();
        return;
    case 0:
        conf_user.dimmer = dimmer + 1;
	tw_brightness(conf_user.dimmer);
	vcu.ctl.schedule_save_config();
	return;
    case 2:
        switch (volume_step) {
        case 0: conf_user.volume_step = 0.1; break;
        case 1: conf_user.volume_step = 0.5; break;
        case 2: conf_user.volume_step = 1.0; break;
        case 3: conf_user.volume_step = 3.0; break;
        case 4: conf_user.volume_step = 6.0; break;
        }
        vcu.ctl.schedule_save_config();
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
    
}

static void
state_select_xtc_prg(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprgs;

    switch (action) {
    case ACTION_INIT: {
	int pos, idx, n, n_programs;
	void **entries;
	prgid_t id;
        char *hdr;

	fprgs = vcu.ambiopole.get_programs(&n_programs);
        
        if (n_programs == 0) {
	    mstate = NULL;
	    textmenu_message("There are no xtc    ",
			     "programs loaded     ");
	    return;
        }
        
	id = vcu.runtime.ambiopole.get_id(current_ambiopole);
	for (idx = 0; memcmp(&id, &fprgs[idx].id, sizeof(id)) != 0 &&
                 idx < n_programs; idx++);
        if (idx == n_programs) {
            vcu.runtime.ambiopole.set_id(current_ambiopole, fprgs[idx].id);
            idx = 0;
        }
        entries = alloca((10 + 2 * n_programs) * sizeof(void *));
	pos = 0;
	for (n = 0; n < n_programs; n++) {
            entries[pos] = alloca(40 + strlen(fprgs[n].longname));
            sprintf((char *)entries[pos++], "%s", fprgs[n].longname);
        }
        entries[pos++] = NULL;
        hdr = alloca(40 + strlen(fprgs[idx].longname));
        sprintf(hdr, "A-pole %d: %s", current_ambiopole + 1,
                fprgs[idx].longname);
	mstate = textmenu_state_new(hdr, entries);
        mstate->pos = idx;
	break;
    }
    default:
	break;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            break;
        }
	vcu.runtime.ambiopole.set_id(current_ambiopole, fprgs[action].id);
	textmenu_state_delete(mstate);
	state_select_xtc_prg(ACTION_INIT);
        break;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }    
}

static void
state_ambiopole_setup(int action)
{
    static struct textmenu_state *mstate;
    static bool_t *xtc_active, isinit = false;
    
    current_program_type = VCU_PROGRAM_TYPE_XTC;
    
    if (action == ACTION_INIT) {
	int n, pos, n_ambiopoles;
	void **entries;

        if (!isinit) {
            xtc_active =
                malloc(vcu.channel.get_max_count(VCU_OUT, VCU_CH_AMBIOPOLE) *
                       sizeof(bool_t));
            isinit = true;
        }
        
        n_ambiopoles = vcu.channel.get_count(VCU_OUT, VCU_CH_AMBIOPOLE) / 2;
        for (n = 0; n < n_ambiopoles; n++) {
            xtc_active[n] = vcu.ambiopole.get_xtc_active(n);
        }
        entries = alloca((5 + 4 * n_ambiopoles) * sizeof(void *));        
        pos = 0;
        entries[pos++] = "Load xtc prg...";
        entries[pos++] = state_load_program;
        entries[pos++] = "Del xtc prg...";
        entries[pos++] = state_del_program;
        for (n = 0; n < n_ambiopoles; n++) {
            entries[pos] = alloca(40);
            sprintf((char *)entries[pos++], "A-pole %d xtc [off/on]", n + 1);
            entries[pos++] = &xtc_active[n];
            entries[pos++] = "Select xtc prg...";
            entries[pos++] = state_select_xtc_prg;
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }

    current_ambiopole = (mstate->pos - 2) / 2;
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action > 2 && ((action - 2) % 2 == 0)) {
            vcu.ambiopole.set_xtc_active(current_ambiopole,
                                         xtc_active[current_ambiopole]);
        }
        break;
    case ACTION_EXIT: {
        int n, n_programs;

        vcu.ambiopole.get_programs(&n_programs);
        for (n = 0;
             n < vcu.channel.get_count(VCU_OUT, VCU_CH_AMBIOPOLE) / 2;
             n++)
        {
            if (xtc_active[n]) {
                if (n_programs == 0) {
                    textmenu_message("Cannot do xtc with- ",
                                     "out xtc programs    ");
                    return;
                }
            }
            vcu.ambiopole.set_xtc_active(n, xtc_active[n]);
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }
    }    
}

static void
state_reverb_defaults(int action)
{
    static struct textmenu_state *mstate;
    static double dry_volume, bformat_directivity;
    static int response_mode, decorrelate_mode;
    static bool_t mix_with_dry, force_delay_diff, force_decay;
    
    if (action == ACTION_INIT) {
        void *entries[100];
        double range[2];
        int pos;

        dry_volume = vcu.reverb.get_dry_volume();
        bformat_directivity = vcu.reverb.get_bformat_directivity();
        response_mode = vcu.reverb.get_response_mode() - 1;
        decorrelate_mode = vcu.reverb.get_decorrelate_mode() - 1;
        force_delay_diff = vcu.reverb.get_force_delay_diff();
        force_decay = vcu.reverb.get_force_decay();
        mix_with_dry = vcu.reverb.get_dry_active();
        pos = 0;
        entries[pos++] = "[Full response/Mute direct snd/"
            "Mute early resp/Stage adapted]";
        entries[pos++] = &response_mode;
        entries[pos++] = "[Decorrelate all/Decorr as needed/Decorrelate none]";
        entries[pos++] = &decorrelate_mode;
        entries[pos++] = "Force LRdelay[off/on]";
        entries[pos++] = &force_delay_diff;
        entries[pos++] = "Force decay [off/on]";
        entries[pos++] = &force_decay;
        entries[pos++] = "Mix with dry [off/on]";
        entries[pos++] = &mix_with_dry;
        entries[pos] = alloca(100);
        vcu.runtime.volume.get_range(range);
        sprintf((char *)entries[pos++], "Dry vol [%%+.1f,0.1,10,%f,%f]dB",
                range[0], range[1]);
        entries[pos++] = &dry_volume;
        entries[pos] = alloca(100);
        vcu.reverb.get_bformat_directivity_range(range);
        sprintf((char *)entries[pos++], "B-format dir [%%.1f,0.1,1,%f,%f]",
                range[0], range[1]);
        entries[pos++] = &bformat_directivity;
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
        vcu.reverb.set_dry_volume(dry_volume);
        vcu.reverb.set_bformat_directivity(bformat_directivity);
        vcu.reverb.set_response_mode(response_mode + 1);
        vcu.reverb.set_decorrelate_mode(decorrelate_mode + 1);
        vcu.reverb.set_force_delay_diff(force_delay_diff);
        vcu.reverb.set_force_decay(force_decay);
        vcu.reverb.set_dry_active(mix_with_dry);
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    }
}

static void
state_reverb_program_properties(int action)
{
    static char *lines[200];
    static char strings[200][100];
    static int n_lines, pos;
    
    if (action == ACTION_INIT) {
        double orig_length, new_length, orig_tail_volume, new_tail_volume;
        double orig_rt60, new_rt60;
        const struct filterprg *fprg;
        int n, i;

        fprg = tune_reverb_program_fprg;
        vcu.runtime.reverb.get_preprocess_properties(&orig_tail_volume,
                                                     &orig_length, &orig_rt60);
        vcu.runtime.reverb.get_postprocess_properties(&new_tail_volume,
                                                      &new_length, &new_rt60);
        n = 0;
        sprintf(strings[n++], "Orig length: %.1fs", orig_length);
        sprintf(strings[n++], "Curnt length: %.1fs", new_length);
        sprintf(strings[n++], "Orig RT60: %.1fs", orig_rt60);
        sprintf(strings[n++], "Curnt RT60: %.1fs", new_rt60);
        sprintf(strings[n++], "Orig tail: %.0fdB", orig_tail_volume);
        sprintf(strings[n++], "Curnt tail: %.0fdB", new_tail_volume);
        switch (fprg->type) {
        case FILTERPRG_TYPE_REVERB_TRADITIONAL:
            sprintf(strings[n++], "Type: traditional");
            break;
        case FILTERPRG_TYPE_REVERB_BFORMAT:
            sprintf(strings[n++], "Type: b-format");
            break;
        }
        switch (fprg->source_layout) {
        case VCU_REVERB_SOURCE_LAYOUT_CENTER:
            sprintf(strings[n++], "Source: center");
            break;
        case VCU_REVERB_SOURCE_LAYOUT_STEREO:
            sprintf(strings[n++], "Source: stereo");
            break;
        case VCU_REVERB_SOURCE_LAYOUT_LRC:
            sprintf(strings[n++], "Source: lrc");
            break;
        }
        if (fprg->n_receivers == 1) {
            sprintf(strings[n++], "Receiver: %.0f/%.0f",
                    fprg->recv_azimuth[0], fprg->recv_elevation[0]);
        } else if (fprg->n_receivers > 100) {
            sprintf(strings[n++], "Receivers: >100");
        } else {
            for (i = 0; i < fprg->n_receivers; i++) {
                sprintf(strings[n++], "Recv %d: %.0f/%.0f",
                        i + 1, fprg->recv_azimuth[i],
                        fprg->recv_elevation[i]);
            }
        }
        if (fprg->n_rates == 1) {
            sprintf(strings[n++], "Rate: %.1fkHz",
                    (double)fprg->rates[0] / 1000.0);
        } else {
            for (i = 0; i < fprg->n_rates; i++) {
                sprintf(strings[n++], "Rate %d: %.1fkHz",
                        i + 1, (double)fprg->rates[i] / 1000.0);
            }
        }
        sprintf(strings[n++], "Format: %s", bf_strsampleformat(fprg->format));
        n_lines = n;
	for (n = 0; n < n_lines; n++) {
	    lines[n] = strings[n];
	}
	lines[n] = NULL;
        pos = 0;
    }

    textmenu_showpage(&action, &pos, (const char **)lines, n_lines);
    
    switch (action) {
    case ACTION_EXIT:
	textgui_pop_state();
	break;
    default:
	break;
    }    
}

static void
state_tune_reverb_program(int action)
{
    static struct textmenu_state *mstate;
    static double dry_volume, bformat_directivity, predelay, volume;
    static int response_mode, decorrelate_mode;
    static int properties_pos;
    static bool_t mix_with_dry, enter_properties, force_delay_diff, force_decay;
    static double balance[3], scale_length;
    
    if (action == ACTION_INIT) {
        const struct filterprg *fprg;
        void *entries[100];
        double range[2];
        int pos;

        enter_properties = false;
        fprg = tune_reverb_program_fprg;
        dry_volume = vcu.runtime.reverb.get_dry_volume();
        bformat_directivity = vcu.runtime.reverb.get_bformat_directivity();
        response_mode = vcu.runtime.reverb.get_response_mode() - 1;
        decorrelate_mode = vcu.runtime.reverb.get_decorrelate_mode() - 1;
        force_delay_diff = vcu.runtime.reverb.get_force_delay_diff();
        force_decay = vcu.runtime.reverb.get_force_decay();
        mix_with_dry = vcu.runtime.reverb.get_dry_active();
        predelay = vcu.runtime.reverb.get_trim_delay();
        volume = vcu.runtime.reverb.get_trim_volume();
        scale_length = vcu.runtime.reverb.get_scale_length() * 100.0;
        pos = 0;
        vcu.runtime.reverb.get_scale_length_range(range);
        entries[pos] = alloca(100);
        sprintf((char *)entries[pos++], "Length [%%.0f,1,10,%f,%f]%%",
                range[0] * 100.0, range[1] * 100.0);
        entries[pos++] = &scale_length;
        entries[pos] = alloca(100);
        vcu.runtime.reverb.get_trim_delay_range(range);
        sprintf((char *)entries[pos++], "Predelay [%%.1f,0.1,10,%f,%f]ms",
                range[0] * 1000.0, range[1] * 1000.0);
        entries[pos++] = &predelay;
        entries[pos] = alloca(100);
        vcu.runtime.reverb.get_trim_volume_range(range);
        sprintf((char *)entries[pos++], "Volume [%%+.1f,0.1,10,%f,%f]dB",
                range[0], range[1]);
        entries[pos++] = &volume;
        entries[pos++] = "[Full response/Mute direct snd/Mute early resp/"
            "Stage adapted]";
        entries[pos++] = &response_mode;
        entries[pos++] = "[Decorrelate all/Decorr as needed/Decorrelate none]";
        entries[pos++] = &decorrelate_mode;
        if (vcu.channel.get_count(IN, VCU_CH_REVERB) > 1 &&
            fprg->source_layout != VCU_REVERB_SOURCE_LAYOUT_CENTER)
        {
            entries[pos++] = "Force LRdelay[off/on]";
            entries[pos++] = &force_delay_diff;
        }
        entries[pos++] = "Force decay [off/on]";
        entries[pos++] = &force_decay;
        entries[pos++] = "Mix with dry [off/on]";
        entries[pos++] = &mix_with_dry;
        entries[pos] = alloca(100);
        vcu.reverb.get_dry_volume_range(range);
        sprintf((char *)entries[pos++], "Dry vol [%%+.1f,0.1,10,%f,%f]dB",
                range[0], range[1]);
        entries[pos++] = &dry_volume;

        vcu.runtime.reverb.get_xyz_balance_range(range);
        vcu.runtime.reverb.get_xyz_balance(balance);
	entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Xbalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[0];
        entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Ybalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[1];
	entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Zbalance [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &balance[2];
                
        if (bformat_directivity >= 0.0) {
            entries[pos] = alloca(100);
            vcu.reverb.get_bformat_directivity_range(range);
            sprintf((char *)entries[pos++], "B-format dir [%%.1f,0.1,1,%f,%f]",
                    range[0], range[1]);
            entries[pos++] = &bformat_directivity;
        }        
        entries[pos++] = "Properties...";
        entries[pos++] = state_reverb_program_properties;
        entries[pos++] = NULL;
	mstate = textmenu_state_new(fprg->longname, entries);
        if (bformat_directivity >= 0.0) {
            properties_pos = 11;
        } else {
            properties_pos = 10;
        }
    }

    if (mstate->pos == properties_pos && action == KI_KEY_RIGHT) {
        /* want to enter "Properties", must update settings */
        enter_properties = true;
        state_tune_reverb_program(ACTION_EXIT);
        return;
    } else {    
        textmenu_action(mstate, &action);
    }

    switch (action) {
    case ACTION_EXIT:        
        vcu.runtime.reverb.set_scale_length(scale_length / 100.0);
        vcu.runtime.reverb.set_trim_volume(volume);
        vcu.runtime.reverb.set_trim_delay(predelay);
        vcu.runtime.reverb.set_response_mode(response_mode + 1);
        vcu.runtime.reverb.set_decorrelate_mode(decorrelate_mode + 1);
        vcu.runtime.reverb.set_force_delay_diff(force_delay_diff);
        vcu.runtime.reverb.set_force_decay(force_decay);
        vcu.runtime.reverb.set_dry_active(mix_with_dry);
        vcu.runtime.reverb.set_dry_volume(dry_volume);
        vcu.runtime.reverb.set_xyz_balance(balance);
        if (bformat_directivity >= 0) {
            vcu.runtime.reverb.set_bformat_directivity(bformat_directivity);
        }
        UPDATE_HARD_SETTINGS_CALL;
        if (enter_properties) {
            enter_properties = false;
            textgui_push_state(state_reverb_program_properties);
            return;
        }
	textmenu_state_delete(mstate);
	textgui_pop_state();
        break;
    case ACTION_CHILD_POP:
        enter_properties = false;
        UPDATE_HARD_SETTINGS_CHILD_POP;
	return;
    }    
}

static void
state_tune_reverb_program_list(int action)
{
    static struct textmenu_state *mstate;
    static const struct filterprg *fprg;
    static int menupos = 0;
    static int n_programs;
    
    if (action == ACTION_INIT) {
	void **entries;
	int n, pos;
	
        fprg = vcu.reverb.get_programs(&n_programs);	
	if (n_programs == 0) {
	    mstate = NULL;
            textmenu_message("There are no reverb ",
                             "programs loaded     ");
	    return;
	}
	
	pos = 0;
	entries = alloca((1 + n_programs) * sizeof(void *));
	for (n = 0; n < n_programs; n++) {
	    /* FIXME: menu command characters in longname can make menu do dumb
	       stuff */
	    entries[pos] = alloca(40 + strlen(fprg[n].longname));
	    sprintf((char *)entries[pos], "%s", fprg[n].longname);
	    pos++;
	}
	entries[pos++] = NULL;
        mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }

    if (mstate == NULL) {
	textgui_pop_state();
	return;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    default:
	if (action >= 0 && action < n_programs) {
            vcu.runtime.reverb.set_id(fprg[action].id);
            tune_reverb_program_fprg = &fprg[action];
            menupos = mstate->pos;
            textgui_push_state(state_tune_reverb_program);
	}
	return;
    case ACTION_EXIT:
        menupos = 0;
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    case ACTION_CHILD_POP:
        textmenu_state_delete(mstate);
        state_tune_reverb_program_list(ACTION_INIT);
        return;
    case ACTION_INIT:
    case ACTION_IGNORE:
	return;
    }
}

static void
state_system_info(int action)
{
    static char *lines[60];
    static char strings[60][100];
    static int n_lines, pos;
    
    if (action == ACTION_INIT) {
	int speed, mem, fmem, days, hours, cmem, n, ri;
        int rates[VCU_VALID_RATES_COUNT], n_rates;
        double alignment_delay, iodelay;

	vcu.sysinfo.get_uptime(&days, &hours);
	speed = vcu.sysinfo.get_cpu_megahertz();
	mem = vcu.sysinfo.get_ram_megabytes();
	fmem = vcu.sysinfo.get_filter_ram_megabytes();
        cmem = (int)(vcu.cache.get_capacity(NULL) / (1024 * 1024));
        vcu.signal.get_rates(rates, &n_rates);
        pos = 0;

	n = 0;
	sprintf(strings[n++], "Version: %s", vcu.sysinfo.get_version());
	sprintf(strings[n++], "Uptime: %ddays/%dh", days, hours);        
        for (ri = 0; ri < n_rates; ri++) {
            if (!vcu.iodelay.need_to_run_benchmark() ||
                vcu.iodelay.manually_set())
            {
                vcu.iodelay.get_delay_length_pair(rates[ri], &iodelay,
                                                  NULL, NULL);
                alignment_delay = vcu.iodelay.get_alignment_delay(rates[ri]);
                if (alignment_delay >= 0) {
                    sprintf(strings[n++], "%.1fkHz I/O-delay:",
                            (double)rates[ri] / 1000.0);
                    sprintf(strings[n++], "  static: %.1fms", iodelay * 1000.0);
                    sprintf(strings[n++], "   %.0f samples",
                            iodelay * (double)rates[ri]);
                    sprintf(strings[n++], "  align: %.1fms",
                            alignment_delay * 1000.0);
                    sprintf(strings[n++], "   %.0f samples",
                            alignment_delay * (double)rates[ri]);
                    sprintf(strings[n++], "  total: %.1fms",
                            (alignment_delay + iodelay) * 1000.0);
                    sprintf(strings[n++], "   %.0f samples",
                            (alignment_delay + iodelay) * (double)rates[ri]);
                }
            }
        }
	if (cmem < 0) {
	    sprintf(strings[n++], "Cache free: <1MB");
	} else if (cmem < 1024) {
	    sprintf(strings[n++], "Cache free: %dMB", (int)cmem);
	} else {
	    sprintf(strings[n++], "Cache free: %.1fGB", (double)cmem / 1024.0);
	}
	if (fmem < 1024) {
	    sprintf(strings[n++], "Filter mem: %dMB", fmem);
	} else {
	    sprintf(strings[n++], "Filter mem: %.1fGB", (double)fmem / 1024.0);
	}
	if (mem < 1024) {
	    sprintf(strings[n++], "Total mem: %dMB", mem);
	} else {
	    sprintf(strings[n++], "Total mem: %.1fGB", (double)mem / 1024.0);
	}
	sprintf(strings[n++], "CPU: %s", vcu.sysinfo.get_cpu_name());
	sprintf(strings[n++], "CPU count: %d", vcu.sysinfo.get_cpu_count());
	if (speed < 1000) {
	    sprintf(strings[n++], "CPU speed: %dMHz", speed);
	} else {
	    sprintf(strings[n++], "CPU speed: %.1fGHz",
		    (double)speed / 1000.0);
	}
	n_lines = n;
	for (n = 0; n < n_lines; n++) {
	    lines[n] = strings[n];
	}
	lines[n] = NULL;
    }

    textmenu_showpage(&action, &pos, (const char **)lines, n_lines);
    
    switch (action) {
    case ACTION_EXIT:
	textgui_pop_state();
	break;
    default:
	break;
    }
}

static int
scandir_select_tar(const struct dirent *dirent)
{
    int slen;
    
    slen = strlen(dirent->d_name);
    if (strstr(dirent->d_name, ".tar") == &dirent->d_name[slen - 4] ||
        strstr(dirent->d_name, ".tgz") == &dirent->d_name[slen - 4] ||
        strstr(dirent->d_name, ".tar.gz") == &dirent->d_name[slen - 7])
    {
	return 1;
    }
    return 0;
}

static void
state_update_software_sub(int action)
{
    static struct textmenu_state *mstate;
    static struct dirent **namelist = NULL;
    static int n_files, idx;
    char path[PATH_MAX], errstr[100];
    int err;
    
    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        textgui_pop_state();
        return;
    case ACTION_INIT: {
        struct dirent **temp_namelist;
	void **entries;
        int n, pos;
        
        idx = access_program_on_rom;
        if (!mount_device_helper(idx)) {
            mstate = NULL;
            return;
        }
        if (namelist != NULL) {
            while (n_files--) {
                free(namelist[n_files]);
            }
            free(namelist);
            namelist = NULL;
        }
        /* scandir seem to fail with namelist on heap... (strange) */
        n_files = scandir(conf_static.rom[idx].path, &temp_namelist,
                          scandir_select_tar, alphasort);
        namelist = temp_namelist;
        if (n_files == 0) {
            mstate = NULL;
            textmenu_message("The medium stores no",
                             "software updates    ");
            return;
        } else if (n_files < 0) {
            elog(LOGDEBUG, "scandir in \"%s\" failed: %s",
                 conf_static.rom[idx].path, strerror(errno));
            mstate = NULL;
            textmenu_message("Failed to list dir  ",
                             "on medium           ");
            return;
        }
        entries = alloca((30 + n_files) * sizeof(void *));
        pos = 0;
        for (n = 0; n < n_files; n++) {
            entries[pos] = alloca(strlen(namelist[n]->d_name) + 10);
            sprintf((char *)entries[pos++], "\"%s\"", namelist[n]->d_name);
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new("Install from:", entries);
    }    
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        elog(LOGDEBUG, "Install update");
        snprintf(path, PATH_MAX, "%s/%s", conf_static.rom[idx].path,
                 namelist[action]->d_name);
        path[PATH_MAX-1] = '\0';
        tw_clear();
        tw_putstring(0, 0, "Installing software ");
        tw_putstring(0, 1, "from archive...     ");
        tw_refresh();
        err = vcu.ctl.update_software(path);
        switch (err) {
        case 0:
            textmenu_message("Software successful-",
                             "ly installed        ");
            return;
        case VCU_ERR_INVALID_FORMAT:
            textmenu_message("File doesn't contain",
                             "software update     ");
            return;
        case VCU_ERR_READ_FAILED:
            textmenu_message("Could not read file ",
                             "                    ");
            return;
        case VCU_ERR_WRITE_FAILED:
            textmenu_message("Write failed, disk  ",
                             "full?               ");
            return;
        default:
            sprintf(errstr, "Unexpeced error %d", err);
            textmenu_message(errstr,
                             "                    ");
            return;
        }
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_update_software(int action)
{
    static struct textmenu_state *mstate;
    int n, pos;

    switch (action) {
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[30 + ROM_COUNT];

        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            snprintf(rom[n], 40, "From %s...", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
            entries[pos++] = state_update_software_sub;
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new("Update software:", entries);
    }
    }

    access_program_on_rom = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_load_config_sub(int action)
{
    static struct textmenu_state *mstate;
    static const struct vcu_config *configs;
    static bool_t load_user_config, load_revprg_config, load_static_config;
    static bool_t really_load, show_message;
    static int idx;
    
    switch (action) {
    case ACTION_CHILD_POP:
        if (show_message) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
        break;
    case ACTION_INIT: {
	void **entries;
        int n, n_configs, pos;

        show_message = false;
        idx = access_program_on_rom;
        if (idx < 0) {
            configs = vcu.ctl.scan_config_dir(vcu.cache.get_path(), &n_configs);
        } else {
            if (!mount_device_helper(idx)) {
                mstate = NULL;
                return;
            }
            configs = vcu.ctl.scan_config_dir(conf_static.rom[idx].path,
                                              &n_configs);
        }
        if (configs == NULL || n_configs == 0) {
            mstate = NULL;
            if (access_program_on_rom >= 0) {
                textmenu_message("The medium stores no",
                                 "configurations      ");
            } else {
                textmenu_message("There are no cached ",
                                 "configurations      ");
            }
            show_message = true;
            return;
        }
        entries = alloca((30 + n_configs) * sizeof(void *));
        pos = 0;
        vcu.ctl.get_load_config_settings(&load_user_config,
                                         &load_revprg_config,
                                         &load_static_config);
        entries[pos++] = "Load user [off/on]";
        entries[pos++] = &load_user_config;
        entries[pos++] = "Load reverb [off/on]";
        entries[pos++] = &load_revprg_config;
        entries[pos++] = "Load static [off/on]";
        entries[pos++] = &load_static_config;
        for (n = 0; n < n_configs; n++) {
            entries[pos] = alloca(VCU_CONFIG_NAME_SIZE + 10);
            sprintf((char *)entries[pos++], "Load \"%s\"?", configs[n].name);
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }    
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 0:
    case 1:
    case 2:
        vcu.ctl.set_load_config_settings(load_user_config,
                                         load_revprg_config,
                                         load_static_config);
        return;
    default:
        if (action < 3) {
            return;
        }
        really_load = true;
        idx = action - 3;
        if (load_static_config) {
            textmenu_yesno(&really_load,
                           "Really load static  ",
                           "config?             ");
            return;
        }
        /* fall through */
    case ACTION_CHILD_POP:
        if (!really_load) {
            return;
        }
        elog(LOGDEBUG, "Loading config");
        tw_clear();
        tw_putstring(0, 0, "Loading config...   ");
        tw_refresh();
        vcu.ctl.load_config(&configs[idx]);
        /* if successful, this should not be reached */
        tw_clear();
        textmenu_message("Could not load conf-",
                         "iguration           ");
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_load_config(int action)
{
    static struct textmenu_state *mstate;
    int n, pos;

    switch (action) {
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[30 + ROM_COUNT];

        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            snprintf(rom[n], 40, "From %s...", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
            entries[pos++] = state_load_config_sub;
        }
        entries[pos++] = "From cache...";
	entries[pos++] = state_load_config_sub;
	entries[pos++] = NULL;
	mstate = textmenu_state_new("Load config:", entries);
    }
    }

    if (mstate->pos < conf_static.n_roms) {
        access_program_on_rom = mstate->pos;
    } else {
        access_program_on_rom = -1;
    }
    
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_save_config_sub(int action)
{
    static struct textmenu_state *mstate;
    static const struct vcu_config *configs;
    static char name[20];
    static int n_configs, idx;
    static int menupos = 0;
    char errstr[100];
    int err;
    
    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        state_save_config_sub(ACTION_INIT);
        return;
    case ACTION_INIT: {
	void **entries;
        int n, pos;
        
        idx = access_program_on_rom;
        if (idx < 0) {
            configs = vcu.ctl.scan_config_dir(vcu.cache.get_path(), &n_configs);
        } else {
            if (!mount_device_helper(idx)) {
                mstate = NULL;
                return;
            }
            configs = vcu.ctl.scan_config_dir(conf_static.rom[idx].path,
                                              &n_configs);
        }
        if (configs == NULL) {
            n_configs = 0;
        }
        entries = alloca((30 + n_configs) * sizeof(void *));
        pos = 0;
        /* FIXME: ugly constant handling 12 below... */
        entries[pos++] = "New [%s,12]";
        name[0] = '\0';
        entries[pos++] = name;
        for (n = 0; n < n_configs; n++) {
            entries[pos] = alloca(VCU_CONFIG_NAME_SIZE + 40);
            sprintf((char *)entries[pos++], "Replace \"%s\"?ask",
                    configs[n].name);
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
        mstate->pos = menupos;
    }    
    }

    menupos = mstate->pos;
    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        if (action == 0) {
            tw_clear();
            tw_putstring(0, 0, "Saving config...    ");
            tw_refresh();
            err = vcu.ctl.save_config((idx < 0) ? vcu.cache.get_path() :
                                      conf_static.rom[idx].path, name,
                                      false);
        } else {
            elog(LOGDEBUG, "Overwriting config \"%s\"", configs[action-1].name);
            tw_clear();
            tw_putstring(0, 0, "Saving config...    ");
            tw_refresh();
            remove(configs[action-1].path);
            err = vcu.ctl.save_config(configs[action-1].path,
                                      configs[action-1].name,
                                      true);
        }
        tw_clear();
        switch (err) {
        case 0:
            textmenu_message("Config was success- ",
                             "fully saved         ");
            return;
        case VCU_ERR_INVALID_VALUE:
            textmenu_message("Invalid device conf-",
                             "iguration           ");
            return;
        case VCU_ERR_OBJECT_EXISTS:
            textmenu_message("Too many saved conf-",
                             "igs                 ");
            return;
        case VCU_ERR_WRITE_FAILED:
            textmenu_message("Write failed, disk  ",
                             "full?               ");
            return;
        default:
            sprintf(errstr, "Unexpeced error %d", err);
            textmenu_message(errstr,
                             "                    ");
            return;
        }
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
        menupos = 0;
	return;
    }
}
    
static void
state_save_config(int action)
{
    static struct textmenu_state *mstate;
    static int rom_idx[ROM_COUNT];
    static bool_t do_pop;
    int n, pos;

    switch (action) {
    case ACTION_CHILD_POP:
        if (do_pop) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
        break;
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[30 + 2 * ROM_COUNT];

        for (n = 0; n < conf_static.n_roms; n++) {
            if (conf_static.rom[n].write) {
                break;
            }
        }        
        if (n == conf_static.n_roms) {
	    mstate = NULL;
	    textmenu_message("There is no place to",
			     "save config to      ");
            do_pop = true;
	    return;
        }
        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            if (!conf_static.rom[n].write) {
                continue;
            }
            rom_idx[pos/2] = n;
            snprintf(rom[n], 40, "To %s...", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
            entries[pos++] = state_save_config_sub;
        }
        rom_idx[pos/2] = -1;
        entries[pos++] = "To cache...";
	entries[pos++] = state_save_config_sub;
	entries[pos++] = NULL;
	mstate = textmenu_state_new("Save config:", entries);
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
        return;
    }
    
    access_program_on_rom = rom_idx[mstate->pos];
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_del_config_sub(int action)
{
    static struct textmenu_state *mstate;
    static const struct vcu_config *configs;
    static bool_t do_pop;
    static int idx;
    
    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        if (do_pop) {
            textgui_pop_state();
        } else {
            state_del_config_sub(ACTION_INIT);
        }
        return;
    case ACTION_INIT: {
	void **entries;
        int n, n_configs, pos;
        
        do_pop = true;
        idx = access_program_on_rom;
        if (idx < 0) {
            configs = vcu.ctl.scan_config_dir(vcu.cache.get_path(), &n_configs);
        } else {
            if (!mount_device_helper(idx)) {
                mstate = NULL;
                return;
            }
            configs = vcu.ctl.scan_config_dir(conf_static.rom[idx].path,
                                              &n_configs);
        }
        if (configs == NULL || n_configs == 0) {
            mstate = NULL;
            if (access_program_on_rom >= 0) {
                textmenu_message("The medium stores no",
                                 "configurations      ");
            } else {
                textmenu_message("There are no cached ",
                                 "configurations      ");
            }
            return;
        }
        entries = alloca((30 + n_configs) * sizeof(void *));
        pos = 0;
        for (n = 0; n < n_configs; n++) {
            entries[pos] = alloca(VCU_CONFIG_NAME_SIZE + 10);
            sprintf((char *)entries[pos++], "Delete \"%s\"?ask",
                    configs[n].name);
        }
        entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }    
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        elog(LOGDEBUG, "Removing config \"%s\"/\"%s\"", configs[action].name,
             configs[action].path);
        tw_clear();
        tw_printf(0, 0, "Deleting config...");
        tw_refresh();
        if (remove(configs[action].path) == -1) {
            tw_clear();
            textmenu_message("Config could not be ",
                             "deleted             ");
        } else {
            tw_clear();
            textmenu_message("Config was success- ",
                             "fully deleted       ");
        }
        do_pop = false;
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_del_config(int action)
{
    static struct textmenu_state *mstate;
    static int rom_idx[ROM_COUNT];
    static bool_t do_pop;
    int n, pos;

    switch (action) {
    case ACTION_CHILD_POP:
        if (do_pop) {
            textmenu_state_delete(mstate);
            textgui_pop_state();
            return;
        }
        break;
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[30 + 2 * ROM_COUNT];

        do_pop = false;
        for (n = 0; n < conf_static.n_roms; n++) {
            if (conf_static.rom[n].write) {
                break;
            }
        }        
        if (n == conf_static.n_roms) {
	    mstate = NULL;
	    textmenu_message("There is no place to",
			     "del config from     ");
            do_pop = true;
	    return;
        }
        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            if (!conf_static.rom[n].write) {
                continue;
            }
            rom_idx[pos/2] = n;
            snprintf(rom[n], 40, "From %s...", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
            entries[pos++] = state_del_config_sub;
        }
        rom_idx[pos/2] = -1;
        entries[pos++] = "From cache...";
	entries[pos++] = state_del_config_sub;
	entries[pos++] = NULL;
	mstate = textmenu_state_new("Delete config:", entries);
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
        return;
    }
    
    access_program_on_rom = rom_idx[mstate->pos];
    textmenu_action(mstate, &action);

    switch (action) {
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_save_logs(int action)
{
    static struct textmenu_state *mstate;
    char errstr[100];
    int n, pos, err;

    switch (action) {
    case ACTION_CHILD_POP:
	textmenu_state_delete(mstate);
        textgui_pop_state();
        return;
    case ACTION_INIT: {
	char rom[ROM_COUNT][40];
	void *entries[5 + ROM_COUNT];

        for (n = 0; n < conf_static.n_roms; n++) {
            if (conf_static.rom[n].write) {
                break;
            }
        }        
        if (n == conf_static.n_roms) {
	    mstate = NULL;
	    textmenu_message("There is no place to",
			     "save logs to        ");
	    return;
        }
        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            if (!conf_static.rom[n].write) {
                continue;
            }
            snprintf(rom[n], 40, "Save to %s?", conf_static.rom[n].name);
            rom[n][39] = '\0';
            entries[pos++] = rom[n];
        }
	entries[pos++] = NULL;
	mstate = textmenu_state_new(NULL, entries);
    }
    }

    if (mstate == NULL) {
	textgui_pop_state();
        return;
    }

    textmenu_action(mstate, &action);

    switch (action) {
    default:
        if (action < 0) {
            return;
        }
        pos = 0;
        for (n = 0; n < conf_static.n_roms; n++) {
            if (!conf_static.rom[n].write) {
                continue;
            }
            if (action == pos) {
                break;
            }
            pos++;
        }
        if (n == conf_static.n_roms) {
            return;
        }
        if (!mount_device_helper(n)) {
            return;
        }
        elog(LOGDEBUG, "Saving logs");
        err = vcu.ctl.save_logs(conf_static.rom[n].path);
        if (conf_static.rom[n].mount) {
            umount(conf_static.rom[n].path);
            mount_path = NULL;
        }
        switch (err) {
        case 0:
            textmenu_message("Logs were success-  ",
                             "fully saved         ");
            return;
        case VCU_ERR_INVALID_VALUE:
            textmenu_message("Invalid device conf-",
                             "iguration           ");
            return;
        case VCU_ERR_READ_FAILED:
            textmenu_message("Could not read logs ",
                             "                    ");
            return;
        case VCU_ERR_WRITE_FAILED:
            textmenu_message("Could not write log,",
                             "disk full?          ");
            return;
        default:
            sprintf(errstr, "Unexpeced error %d", err);
            textmenu_message(errstr,
                             "                    ");
            return;
        }
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	return;
    }
}

static void
state_reset_config(int action)
{
    static struct textmenu_state *mstate;
    static bool_t reset_revprg;
    static bool_t reset_user;

    if (action == ACTION_INIT) {

        reset_revprg = true;
        reset_user = true;
	
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
		"Reset user [off/on]", &reset_user,
		"Reset reverb [off/on]", &reset_revprg,
                "Reset config?ask",
		NULL
	    });
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 2:
        vcu.ctl.restart(reset_revprg, reset_user);
        return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	break;
    default:
	return;
    }
}

static void
state_maintenance(int action)
{
    static struct textmenu_state *mstate;
    static bool_t debug;

    if (action == ACTION_INIT) {

	debug = vcu.ctl.get_debug();
	
	mstate = textmenu_state_new(
	    NULL,
	    (void *[]){
                "Load config...", state_load_config,
                "Save config...", state_save_config,
                "Del config...", state_del_config,
                "Reset config...", state_reset_config,
                "Flush cache junk?ask",
		"Debug mode [off/on]", &debug,
                "Flush logs?ask",
                "Save logs...", state_save_logs,
                "Update software...", state_update_software,
                "Restart AlmusVCU?ask",
		"Reboot machine?ask",
		"Shutdown machine?ask",
		NULL
	    });
    }

    textmenu_action(mstate, &action);

    switch (action) {
    case 4:
        vcu.cache.flush_orphan_resources();
        return;
    case 5:
	vcu.ctl.set_debug(debug);
	return;
    case 6:
        vcu.ctl.flush_logs();
        return;
    case 9:
        vcu.ctl.restart(false, false);
        return;
    case 10:
	tw_clear();
	tw_putstring(0, 0, "The machine is prep-");
	tw_putstring(0, 1, "aring for reboot.   ");
	tw_refresh();
        system("shutdown -r now");
        vcu.ctl.exit(EXIT_SUCCESS);
	return;
    case 11:
	tw_clear();
	tw_putstring(0, 0, "The machine is prep-");
	tw_putstring(0, 1, "aring for shutdown. ");
	tw_refresh();
        system("shutdown -h now");
        vcu.ctl.exit(EXIT_SUCCESS);
	return;
    case ACTION_EXIT:
	textmenu_state_delete(mstate);
	textgui_pop_state();
	break;
    default:
	return;
    }
}

static void
state_setup(int action)
{
    static struct textmenu_state *mstate;
    static bool_t first_call = true;
    static time_t start_time;
    static int dimmer;
    static bool_t ask_ignore = false;
    static bool_t master_mute;
    static double master_vol;
    static int restart = 0;
    bool_t start_convolver = false;
    int err, n, start_sample_rate;
    prgid_t id, zeroid;
    char s[40];

    current_program_type = VCU_PROGRAM_TYPE_REVERB;
    
    switch (action) {
    case ACTION_CHILD_POP:
        master_vol = vcu.runtime.volume.get_master();
        master_mute = vcu.runtime.volume.get_master_temp_mute();
        if (mount_path != NULL) {
            umount(mount_path);
            mount_path = NULL;
        }
        break;
    case ACTION_INIT: {
        void *entries[100];
        double range[2];
        int pos;
        
        dimmer = conf_user.dimmer - 1;
        vcu.runtime.volume.get_range(range);
        master_vol = vcu.runtime.volume.get_master();
        master_mute = vcu.runtime.volume.get_master_temp_mute();

        pos = 0;
        entries[pos++] = "Start convolver?";
        entries[pos] = alloca(80);
	sprintf((char *)entries[pos++], "Master [%%+.1f,0.1,10,%f,%f]dB",
		range[0], range[1]);
	entries[pos++] = &master_vol;
        entries[pos++] = "Master mute [off/on]";
        entries[pos++] = &master_mute;
        entries[pos++] = "Load reverb prg...";
        entries[pos++] = state_load_program;
        entries[pos++] = "Del reverb prg...";
        entries[pos++] = state_del_program;
        entries[pos++] = "Tune reverb prg...";
        entries[pos++] = state_tune_reverb_program_list;
        entries[pos++] = "Signal setup...";
        entries[pos++] = state_signal_setup;
        entries[pos++] = "Channel setup...";
        entries[pos++] = state_channel_setup;
        entries[pos++] = "Eq setup...";
        entries[pos++] = state_eq_setup;
        entries[pos++] = "Ambiopole setup...";
        entries[pos++] = state_ambiopole_setup;
        entries[pos++] = "Reverb defaults...";
        entries[pos++] = state_reverb_defaults;
        entries[pos++] = "Convolver setup...";
        entries[pos++] = state_convolver_setup;
        entries[pos++] = "User intf setup...";
        entries[pos++] = state_user_interface_setup;
        entries[pos++] = "System info...";
        entries[pos++] = state_system_info;
        entries[pos++] = "Maintenance...";
        entries[pos++] = state_maintenance;
        entries[pos++] = NULL;

	mstate = textmenu_state_new(NULL, entries);
	if (first_call) {
	    first_call = false;
	    if (conf_user.autostart) {
		start_convolver = true;
	    }
	}
    }
    }

    if (start_convolver) {
	action = 0;
    } else if (convolver_last_error != 0) {
        n = convolver_last_error;
        convolver_last_error = 0;
	switch (n) {
	case VCU_ERR_CONVOLVER_FAILED_UNKNOWN:
	    textmenu_message("Unknown convolver   ",
			     "failure!            ");
            return;
	case VCU_ERR_CONVOLVER_FAILED_UNDERFLOW:
	    textmenu_message("Convolver got under-",
			     "flowed              ");
            return;
	case VCU_ERR_CONVOLVER_FAILED_NO_MEMORY:
	    textmenu_message("Convolver ran out of",
			     "memory              ");
            return;
	case VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX:
	    textmenu_message("Too complex config  ",
			     "for convolver       ");
            return;
	case VCU_ERR_CONVOLVER_FAILED_SIGNAL:
            if (restart == 3) {
                restart = 0;
                textmenu_message("Unsupported or unst-",
                                 "able signal         ");
                return;
            } else {
                if (time(NULL) < start_time + 5) {
                    restart += 1;
                } else {
                    restart = 1;
                }
                action = 0;
            }
	    break;
	}
    } else if (ask_ignore) {
        switch (action) {
        case KI_KEY_LEFT:
            start_convolver = true;
            ask_ignore = false;
            action = 0;
            break;
        case KI_KEY_RIGHT:
            ask_ignore = false;
            mstate->pos = 11;
            action = ACTION_REFRESH;
            textmenu_action(mstate, &action);
            return;
        default:
            return;
        }
    } else {
        restart = 0;
	textmenu_action(mstate, &action);
    }
    
    switch (action) {
    case 0:
	if (!have_channels()) {
	    textmenu_message("There are no confi- ",
			     "gured channels      ");
	    return;
	}
	if (vcu.iodelay.need_to_run_benchmark() &&
            !vcu.iodelay.manually_set() && !start_convolver)
        {
            tw_clear();
            tw_putstring(0, 0, "Need to benchmark   ");
            tw_printf(0, 1, " Ignore?  %cyes %cno", TWC_LEFT, TWC_RIGHT);
            tw_refresh();
            ask_ignore = true;
	    return;
	}
	if (vcu.iodelay.has_truncated_filters() && !start_convolver) {
            tw_clear();
            tw_putstring(0, 0, "Filters truncated   ");
            tw_printf(0, 1, " Ignore?  %cyes %cno", TWC_LEFT, TWC_RIGHT);
            tw_refresh();
            ask_ignore = true;
	    return;
	}
	memset(&zeroid, 0, sizeof(zeroid));
	id = vcu.runtime.reverb.get_id();
	if (vcu.channel.get_count(VCU_OUT, VCU_CH_REVERB) > 0 &&
	    memcmp(&id, &zeroid, sizeof(id)) == 0)
	{
	    mstate->pos = 3;
	    textmenu_message("There are no reverb ",
			     "programs loaded     ");	    
	    return;
	}
        for (n = 0;
             n < vcu.channel.get_count(VCU_OUT, VCU_CH_AMBIOPOLE) >> 1;
             n++)
        {
            if (vcu.ambiopole.get_xtc_active(n)) {
                id = vcu.runtime.ambiopole.get_id(n);
                if (memcmp(&id, &zeroid, sizeof(id)) == 0) {
                    mstate->pos = 9;
                    textmenu_message("Ambiopole needs an  ",
                                     "xtc program         ");	    
                    return;
                }
            }
        }
        
        if (conf_user.file.use && conf_user.file.path != NULL) {
            for (n = 0; n < conf_static.n_roms; n++) {
                if (strstr(conf_user.file.path, conf_static.rom[n].path) ==
                    conf_user.file.path)
                {
                    if (conf_static.rom[n].mount) {
                        if (!mount_device_helper(n)) {
                            return;
                        }
                    }
                    break;
                }
            }
        }
        elog(LOGDEBUG, "calling update_hard_settings");
	vcu.ctl.update_hard_settings();
        tw_clear();
        if (restart != 0) {
            tw_putstring(0, 0, "Input signal changed");
            tw_putstring(0, 1, "properties (restart)");
        } else {
            tw_putstring(0, 0, "Convolver engine is ");
            tw_putstring(0, 1, "starting...         ");
        }
        tw_refresh();
	runtime_state.peak = VCU_PEAK_NEG_INFINITE;
        if (conf_user.file.use && conf_user.file.path != NULL) {
            start_sample_rate = 0;
            err = vcu.runtime.start_file_input(peak_meter_callback,
                                               convolver_terminated,
                                               conf_user.file.sample_format,
                                               conf_user.file.channel_count,
                                               conf_user.file.loop,
                                               conf_user.file.path);
        } else {
            start_sample_rate = vcu.runtime.signal.get_input_sample_rate();
            err = vcu.runtime.start(peak_meter_callback,
                                    convolver_terminated);
        }
	switch (err) {
	case 0:
	    elog(LOGDEBUG, "successfully started convolver");
            start_time = time(NULL);
	    textgui_push_state(state_runtime_main_screen);
	    return;
	case VCU_ERR_NO_SIGNAL:
            if (conf_user.file.use) {
                textmenu_message("Could not read the  ",
                                 "input file          ");
            } else {
                textmenu_message("There is no input   ",
                                 "signal              ");
            }
	    return;
	case VCU_ERR_UNSUPPORTED_SAMPLE_RATE:
            sprintf(s, "rate (%dkHz)", start_sample_rate / 1000);
	    textmenu_message("Unsupported sample  ", s);
	    return;
	case VCU_ERR_CONVOLVER_FAILED_UNKNOWN:
	    textmenu_message("Failed! Static conf-",
			     "iguration error?    ");
	    return;
	case VCU_ERR_CONVOLVER_FAILED_UNDERFLOW:
	    textmenu_message("Convolver got under-",
			     "flowed              ");
	    return;
	case VCU_ERR_CONVOLVER_FAILED_NO_MEMORY:
	    textmenu_message("Convolver ran out of",
			     "memory              ");
	    return;
	case VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX:
	    textmenu_message("Too complex config  ",
			     "for convolver       ");
            return;
	case VCU_ERR_CONVOLVER_FAILED_SIGNAL:
	    /* very unlikely case(?) */
	    textmenu_message("Convolver could not ",
			     "handle signal       ");
	    return;
	default:
	    textmenu_message("Unexpected error    ",
			     "probably a bug      ");
	    elog(LOGDEBUG, "start returned: %d", err);
	    return;
	}
	break;
    case 1:
        vcu.runtime.volume.set_master(master_vol);
        break;
    case 2:
        vcu.runtime.volume.set_master_temp_mute(master_mute);
        break;
    default:
	return;
    }
}

const char *
textgui_straction(int action)
{
    static char s[50];
    
    switch (action) {
    case KI_KEY_UP: return "UP";
    case KI_KEY_DOWN: return "DOWN";
    case KI_KEY_LEFT: return "LEFT";
    case KI_KEY_RIGHT: return "RIGHT";
    case ACTION_REFRESH: return "REFRESH";
    case ACTION_IGNORE: return "IGNORE";
    case ACTION_INIT: return "INIT";
    case ACTION_EXIT: return "EXIT";
    case ACTION_CHILD_POP: return "CHILD_POP";
    default:
        sprintf(s, "UNKNOWN(%d)", action);
        return s;
    }
}

void
textgui_push_state(state_func_t state)
{
    elog(LOGDEBUG, "push state %p %d", state, stackpos + 1);
    
    stackpos++;
    if (stackpos == STATE_STACK_SIZE) {
	elog(LOGERROR, "state stack overflow");
        exit(EXIT_FAILURE);
    }
    state_stack[stackpos] = state;
    state(ACTION_INIT);
}

void
textgui_pop_state(void)
{
    elog(LOGDEBUG, "pop state");
    
    stackpos--;
    if (stackpos == -1) {
	elog(LOGERROR, "state stack underflow");
        exit(EXIT_FAILURE);
    }
    state_stack[stackpos](ACTION_CHILD_POP);
}

static void
vcu_callback(int fd)
{
    struct timeval tv, tv1;
    int action;

    if ((action = keyinput_getkey(fd, vcu.runtime.isruntime()))
        == ACTION_IGNORE)
    {
	return;
    }
    if (action == KI_KEY_FAILURE) {
        elog(LOGDEBUG, "keyinput on fd %d failed, deactivating it", fd);
        close(fd);
        vcu.ctl.unregister_read_fd(fd);
        return;
    }
    vcu.ctl.unregister_timer(screensaver_timer_id);
    tv.tv_usec = 0;
    tv.tv_sec = SCREENSAVER_TIMER;
    screensaver_timer_id = vcu.ctl.register_timer(NULL, screensaver, &tv);
    if (screensaver_screen != NULL) {
        tw_restore_screen(screensaver_screen);
        tw_delete_screen(screensaver_screen);
        screensaver_screen = NULL;
        tw_refresh();
        return;
    }
    if (vcu.runtime.isruntime()) {
        action = handle_runtime_key(action);
    } else if (action >= KI_KEY_OPTIONAL_BASE) {
        return;
    }
    elog(LOGDEBUG, "action: %s, state_stack[%d]: %p",
         textgui_straction(action), stackpos, state_stack[stackpos]);
    gettimeofday(&tv, NULL);
    state_stack[stackpos](action);
    gettimeofday(&tv1, NULL);
    timersub(&tv1, &tv, &tv);
    if (tv.tv_sec > 0 || tv.tv_usec > 200000) {
        elog(LOGDEBUG, "flushing keyinput queue");
        keyinput_flush_queue(-1);
    }
}

const char *
vcu_name(void)
{
    return "textgui";
}

#define GET_TOKEN(token, error)                                                \
    if (get_config_token(&lval) != token) {                                    \
        elog(LOGERROR, "Parse error: " error );                                \
        return false;                                                          \
    }

bool_t
vcu_loaduserconf(int (*get_config_token)(union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int tok, n;
    
    if (get_config_token == NULL) {
	set_conf_user_defaults();
	return true;
    }

    while ((tok = get_config_token(&lval)) > 0) {
        if (tok == VCU_LEXVAL_STRING && strcmp(lval.string, "file") == 0) {
            GET_TOKEN(VCU_LEX_LBRACE, "expected left brace ({)");
            while ((tok = get_config_token(&lval)) > 0) {
                if (tok == VCU_LEX_RBRACE) {
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                    break;
                }
                if (tok != VCU_LEXVAL_FIELD) {
                    elog(LOGERROR, "Parse error: expected field.");
                    return false;            
                }
                if (strcmp(lval.field, "use") == 0) {
                    GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean for "
                              "file / use");
                    conf_user.file.use = lval.boolean;
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                } else if (strcmp(lval.field, "loop") == 0) {
                    GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean for "
                              "file / loop");
                    conf_user.file.loop = lval.boolean;
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                } else if (strcmp(lval.field, "sample_format") == 0) {
                    GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                              "file / sample_format");
                    conf_user.file.sample_format = BF_SAMPLE_FORMAT_S16_LE;
                    for (n = BF_SAMPLE_FORMAT_MIN;
                         n <= BF_SAMPLE_FORMAT_MAX;
                         n++)
                    {
                        if (strcmp(lval.string, bf_strsampleformat(n)) == 0) {
                            conf_user.file.sample_format = n;
                            break;
                        }
                    }
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                } else if (strcmp(lval.field, "channel_count") == 0) {
                    GET_TOKEN(VCU_LEXVAL_REAL, "expected real for "
                              "file / channel_count");
                    conf_user.file.channel_count = (int)lval.real;
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                } else if (strcmp(lval.field, "path") == 0) {
                    GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                              "file / path");
                    free(conf_user.file.path);
                    conf_user.file.path = strdup(lval.string);
                    GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
                } else {
                    /* skip to VCU_LEX_EOS */
                    while ((tok = get_config_token(&lval)) > 0 &&
                           tok != VCU_LEX_EOS);
                }
            }
            continue;
        }
        if (tok != VCU_LEXVAL_FIELD) {
            elog(LOGERROR, "Parse error: expected field.");
            return false;            
        }
        if (strcmp(lval.field, "autostart") == 0) {
            GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean");
            conf_user.autostart = lval.boolean;
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "channel_mode_option") == 0) {
            GET_TOKEN(VCU_LEXVAL_STRING, "expected string");
            if (strcmp(lval.string, "locked") == 0) {
                conf_user.channel_mode_option = SELECT_OPTION_LOCKED;
            } else if (strcmp(lval.string, "select_on") == 0) {
                conf_user.channel_mode_option = SELECT_OPTION_SEL_ON;
            } else if (strcmp(lval.string, "select_off") == 0) {
                conf_user.channel_mode_option = SELECT_OPTION_SEL_OFF;
            } else {
                conf_user.channel_mode_option = SELECT_OPTION_LOCKED;
            }
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "eq_mode_option") == 0) {
            GET_TOKEN(VCU_LEXVAL_STRING, "expected string");
            if (strcmp(lval.string, "locked") == 0) {
                conf_user.eq_mode_option = SELECT_OPTION_LOCKED;
            } else if (strcmp(lval.string, "select_on") == 0) {
                conf_user.eq_mode_option = SELECT_OPTION_SEL_ON;
            } else if (strcmp(lval.string, "select_off") == 0) {
                conf_user.eq_mode_option = SELECT_OPTION_SEL_OFF;
            } else {
                conf_user.eq_mode_option = SELECT_OPTION_LOCKED;
            }
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "use_master_vol") == 0) {
            GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean");
            conf_user.vol_master = lval.boolean;
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "use_groups_vol") == 0) {
            GET_TOKEN(VCU_LEX_LBRACE, "expected '{'");
            for (n = 0, tok = VCU_LEX_COMMA;
                 n < VCU_MAX_CHANNEL_GROUPS && tok == VCU_LEX_COMMA;
                 n++)
            {
                GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean");
                conf_user.vol_groups[n] = lval.boolean;
                tok = get_config_token(&lval);
            }
            if (tok != VCU_LEX_RBRACE) {
                elog(LOGERROR, "Parse error: expected '}'");
                return false;
            }
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "volume_step") == 0) {
            GET_TOKEN(VCU_LEXVAL_REAL, "expected real value");
            conf_user.volume_step = lval.real;
            if (conf_user.volume_step < 0.1 ||
                conf_user.volume_step > 6.0)
            {
                conf_user.volume_step = 3.0;
            }
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else if (strcmp(lval.field, "dimmer") == 0) {
            GET_TOKEN(VCU_LEXVAL_REAL, "expected real value");
            conf_user.dimmer = (int)lval.real;
            if (conf_user.dimmer < 1 || conf_user.dimmer > 4) {
                conf_user.dimmer = 1;
            }
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        } else {
            /* skip to VCU_LEX_EOS */
            while ((tok = get_config_token(&lval)) > 0 &&
                   tok != VCU_LEX_EOS);
        }
    }
    tw_brightness(conf_user.dimmer);
    return true;
}

void
vcu_saveuserconf(void (*put_config_token)(int tok,
                                          const union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int n;
    
    lval.field = "autostart";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.boolean = conf_user.autostart;
    put_config_token(VCU_LEXVAL_BOOLEAN, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "channel_mode_option";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    switch (conf_user.channel_mode_option) {
    default:
    case SELECT_OPTION_LOCKED:
        lval.string = "locked";
        break;
    case SELECT_OPTION_SEL_ON:
        lval.string = "select_on";
        break;
    case SELECT_OPTION_SEL_OFF:
        lval.string = "select_off";
        break;
    }
    put_config_token(VCU_LEXVAL_STRING, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "eq_mode_option";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    switch (conf_user.eq_mode_option) {
    default:
    case SELECT_OPTION_LOCKED:
        lval.string = "locked";
        break;
    case SELECT_OPTION_SEL_ON:
        lval.string = "select_on";
        break;
    case SELECT_OPTION_SEL_OFF:
        lval.string = "select_off";
        break;
    }
    put_config_token(VCU_LEXVAL_STRING, &lval);
    put_config_token(VCU_LEX_EOS, NULL);
    
    lval.field = "use_master_vol";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.boolean = conf_user.vol_master;
    put_config_token(VCU_LEXVAL_BOOLEAN, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "use_groups_vol";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    put_config_token(VCU_LEX_LBRACE, NULL);
    for (n = 0; n < VCU_MAX_CHANNEL_GROUPS; n++) {
        lval.boolean = conf_user.vol_groups[n];
        put_config_token(VCU_LEXVAL_BOOLEAN, &lval);
        if (n < VCU_MAX_CHANNEL_GROUPS - 1) {
            put_config_token(VCU_LEX_COMMA, NULL);
        }
    }
    put_config_token(VCU_LEX_RBRACE, NULL);
    put_config_token(VCU_LEX_EOS, NULL);
    
    lval.field = "dimmer";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.real = (double)conf_user.dimmer;
    put_config_token(VCU_LEXVAL_REAL, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "volume_step";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.real = conf_user.volume_step;
    put_config_token(VCU_LEXVAL_REAL, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.string = "file";
    put_config_token(VCU_LEXVAL_STRING, &lval);
    put_config_token(VCU_LEX_LBRACE, NULL);

    lval.field = "use";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.boolean = conf_user.file.use;
    put_config_token(VCU_LEXVAL_BOOLEAN, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "loop";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.boolean = conf_user.file.loop;
    put_config_token(VCU_LEXVAL_BOOLEAN, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "sample_format";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.string =
        alloca(strlen(bf_strsampleformat(conf_user.file.sample_format)) + 1);
    strcpy(lval.string, bf_strsampleformat(conf_user.file.sample_format));
    put_config_token(VCU_LEXVAL_STRING, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "channel_count";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.real = (double)conf_user.file.channel_count;
    put_config_token(VCU_LEXVAL_REAL, &lval);
    put_config_token(VCU_LEX_EOS, NULL);

    lval.field = "path";
    put_config_token(VCU_LEXVAL_FIELD, &lval);
    lval.string = conf_user.file.path;
    if (lval.string == NULL) {
        lval.string = "";
    }
    put_config_token(VCU_LEXVAL_STRING, &lval);
    put_config_token(VCU_LEX_EOS, NULL);
    
    put_config_token(VCU_LEX_RBRACE, NULL);
    put_config_token(VCU_LEX_EOS, NULL);    
}

bool_t
vcu_preinit(int (*get_config_token)(union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int tok, idx, n;
    char *s;

    memset(&conf_static, 0, sizeof(conf_static));
    while ((tok = get_config_token(&lval)) > 0) {
        if (tok != VCU_LEXVAL_FIELD) {
	    elog(LOGERROR, "Parse error: expected field.");
	    return false;            
        }
        if (strstr(lval.field, "rom") == lval.field) {
            if ((s = strchr(lval.field, '_')) == NULL) {
                elog(LOGERROR, "Parse error: \"%s\" field unknown", lval.field);
                return false;
            }
            if (lval.field[3] == '_') {
                idx = 1;
            } else {
                idx = atoi(&lval.field[3]);
            }
            if (idx < 1 || idx > ROM_COUNT) {
                elog(LOGERROR, "Parse error: invalid rom index.");
                return false;
            }
            idx--;
            if (strcmp(s, "_path") == 0) {
                GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "rom_path field");
                conf_static.rom[idx].path = strdup(lval.string);
                if (conf_static.rom[idx].path
                    [strlen(conf_static.rom[idx].path)-1] == '/')
                {
                    conf_static.rom[idx].path
                        [strlen(conf_static.rom[idx].path)-1] = '\0';
                }
                GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            } else if (strcmp(s, "_device") == 0) {
                GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "rom_device field");
                conf_static.rom[idx].device = strdup(lval.string);
                GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            } else if (strcmp(s, "_filesys") == 0) {
                GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "rom_filesys field");
                conf_static.rom[idx].filesys = strdup(lval.string);
		GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            } else if (strcmp(s, "_name") == 0) {
                GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "rom_name field");
                conf_static.rom[idx].name = strdup(lval.string);
                GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            } else if (strcmp(s, "_mount") == 0) {
                GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean for "
			  "rom_mount");
                conf_static.rom[idx].mount = lval.boolean;
                GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            } else if (strcmp(s, "_write") == 0) {
                GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean for "
			  "rom_write");
                conf_static.rom[idx].write = lval.boolean;
                GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            }
        } else if (strcmp(lval.field, "display_size") == 0) {
            GET_TOKEN(VCU_LEX_LBRACE, "expected left brace ({)");
            GET_TOKEN(VCU_LEXVAL_REAL, "expected integer for "
                      "display_size");
            conf_static.xsize = (int)lval.real;
            if (conf_static.xsize < MIN_XSIZE) {
                elog(LOGERROR, "Display xsize is too small, must be at "
			 "least %d.", MIN_XSIZE);
            }
            GET_TOKEN(VCU_LEX_COMMA, "expected comma (;)");
            GET_TOKEN(VCU_LEXVAL_REAL, "expected integer for "
                      "display_size");
            conf_static.ysize = (int)lval.real;
            if (conf_static.ysize < MIN_YSIZE) {
                elog(LOGERROR, "Display ysize is too small, must be at "
                     "least %d.", MIN_YSIZE);
            }
            GET_TOKEN(VCU_LEX_RBRACE, "expected right brace (})");
            GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
        }
    }

    for (n = 0; n < ROM_COUNT; n++) {        
        if (conf_static.rom[n].name == NULL &&
            conf_static.rom[n].path == NULL)
        {
            break;
        }
        if (conf_static.rom[n].mount &&
            (conf_static.rom[n].device == NULL ||
             conf_static.rom[n].filesys == NULL))
        {
            elog(LOGERROR, "Parse error: missing device or filesys for "
                 "mountable rom.");
            return false;
        }
    }
    conf_static.n_roms = n;
    
    return true;
}

#undef GET_TOKEN

void
vcu_finish(void)
{
    if (mount_path != NULL) {
        umount(mount_path);
        mount_path = NULL;
    }
}

bool_t
vcu_init(const struct vcu *_vcu)
{
    const struct ui_module *modules;
    struct textwin_output outputs[100];    
    struct keyinput inputs[100];
    struct twki twki;
    int n_modules, n, n_outputs = 0, n_inputs = 0;
    struct timeval tv;
    fd_set rfds;

    memcpy(&vcu, _vcu, sizeof(struct vcu));
    elog_init(vcu.ctl.get_log_file_path());
    modules = vcu.ctl.get_modules(&n_modules);
    for (n = 0; n < n_modules; n++) {
	if (strstr(modules[n].name, "twki_") == modules[n].name) {
	    if (n_outputs == 100) {
		elog(LOGDEBUG, "Too many outputs.");
		return false;
	    }
	    if (n_inputs == 100) {
		elog(LOGDEBUG, "Too many inputs.");
		return false;
	    }
	    if (modules[n].shared_data == NULL) {
		elog(LOGERROR, "Module \"%s\" does not share its functions.",
                     modules[n].name);
		return false;
	    }
            memcpy(&twki, modules[n].shared_data, sizeof(struct twki));
	    memcpy(&outputs[n_outputs], &twki.tw, sizeof(twki.tw));
	    memcpy(&inputs[n_inputs], &twki.ki, sizeof(twki.ki));
	    n_inputs++;
	    n_outputs++;
	} else if (strstr(modules[n].name, "tw_") == modules[n].name) {
	    if (n_outputs == 100) {
		elog(LOGDEBUG, "Too many outputs.");
		return false;
	    }
	    if (modules[n].shared_data == NULL) {
		elog(LOGERROR, "Module \"%s\" does not share its functions.",
                     modules[n].name);
		return false;
	    }
	    memcpy(&outputs[n_outputs], modules[n].shared_data,
		   sizeof(struct textwin_output));
	    n_outputs++;
	} else if (strstr(modules[n].name, "ki_") == modules[n].name) {
	    if (n_inputs == 100) {
		elog(LOGDEBUG, "Too many inputs.");
		return false;
	    }
	    if (modules[n].shared_data == NULL) {
		elog(LOGERROR, "Module \"%s\" does not share its functions.",
		     modules[n].name);
		return false;
	    }
	    memcpy(&inputs[n_inputs], modules[n].shared_data,
		   sizeof(struct keyinput));
	    n_inputs++;
	}
    }
    if (n_outputs == 0) {
	elog(LOGDEBUG, "There are no outputs to draw the UI on.");
	return false;
    }
    if (n_inputs == 0) {
	elog(LOGDEBUG, "There are no inputs to get user input from.");
	return false;
    }
    
    if (!tw_init(outputs, n_outputs, conf_static.xsize, conf_static.ysize,
		 vcu.ctl.register_timer, vcu.ctl.unregister_timer))
    {
	elog(LOGDEBUG, "Could not initialise outputs.");
	return false;
    }
    if (!keyinput_init(inputs, n_inputs)) {
	elog(LOGDEBUG, "Could not initialise inputs.");
	return false;
    }
    
    FD_ZERO(&rfds);
    keyinput_setfds(&rfds);
    n = -1;
    while ((n = bit_find((uint32_t *)((void *)&rfds), n + 1,
                         FD_SETSIZE - 1)) != -1)
    {
	vcu.ctl.register_read_fd(n, vcu_callback);
    }

    tw_clear();
    tw_refresh();
    srandom(0);
    
    tv.tv_usec = 0;
    tv.tv_sec = SCREENSAVER_TIMER;
    screensaver_timer_id = vcu.ctl.register_timer(NULL, screensaver, &tv);

    last_sysinfo.callback_active = false;
    
    return true;
}
